
/***************************************************************************
 *  filename  : LDAPSynchConnection.cpp
 *  authors   : Salvatore Monforte <salvatore.monforte@ct.infn.it>
 *  copyright : (C) 2001, 2002 by INFN
 ***************************************************************************/

/**
 * @file LDAPSynchConnection.cpp
 * @brief This files contains LDAP Synchronous connections implementation.
 * @author Salvatore Monforte salvatore.monforte@ct.infn.it
 * @author comments by Marco Pappalardo marco.pappalardo@ct.infn.it and Salvatore Monforte
 */
// $Id:

/** Includes header definitions. */
#include "edg/workload/common/ldif2classad/LDAPQuery.h"
#include "edg/workload/common/ldif2classad/LDAPSynchConnection.h"

#ifdef WITH_LDAP_EXCEPTIONS
#include "edg/workload/common/ldif2classad/exceptions.h"
#include "edg/workload/common/utilities/edgstrstream.h"
#endif 

namespace edg {
namespace workload {
namespace common {
namespace ldif2classad {

/**
 * Constructor.
 * @param b the base DN of the entry at which to start operations.
 * @param hn the host name of the LDAP server.
 * @param p the port of the LDAP server.
 * @param t the timeout used during binding to an LDAP server and searching its content.
 * @param retry the maximum errors during open() or execute() before giving up
 */
LDAPSynchConnection::LDAPSynchConnection(const std::string& base_dn, const std::string& source_name, int source_port, long tv_sec)
{
  this -> handle     = NULL;
  this -> base_dn     = base_dn;
  this -> source_name = source_name;
  this -> source_port = source_port;
  this -> max_retries = 3;
  this -> failures = 0;
  this -> saved_error = false;
  timeout.tv_sec = tv_sec;
  timeout.tv_usec = 0;
}

/**
 * Destructor.
 */
LDAPSynchConnection::~LDAPSynchConnection()
{
  close();
}

/**
 * Private function to incriment the failure
 * count and wait a short time, to allow
 * whatever condition that caused the last
 * failure to pass
 */
void LDAPSynchConnection::l_note_failure()
{
  failures++;
  if (failures <= max_retries) {
    sleep(5);
  }
}

/**
 * Opens a connection to the Server. 
 * This function overrides virtual LDAP Connection one.
 * @return true on success, false otherwise.
 */
bool LDAPSynchConnection::open()
{
  std::string connection_info;
  bool could_not_init,got_error_code,result;
  int rc;

  failures = 0;
  saved_error = false;

  result = l_open(rc,got_error_code,could_not_init);
 
  if (!result) {
    if (got_error_code) {
      saved_error = true;
      ldap_last_error = rc;
    }
#ifdef WITH_LDAP_EXCEPTIONS
    std::string connection_info;
    utilities::oedgstrstream source;
    source << "contact=" << source_name << ":" << source_port << ", dn=" << base_dn;      
    connection_info = source.str();

    if (got_error_code) {
      std::string error_msg( ldap_err2string(ldap_last_error) );
      throw ConnectionException(connection_info, std::string("ldap_simple_bind_s()"), error_msg); 
    } else {
      std::string error_msg("Unknown ldap init error");
      throw ConnectionException(connection_info, std::string("ldap_init()"), error_msg);
    }
#endif
  }

  return result;
}

/**
 * Private function that actualy does most
 * of the work of openening a connection
 * @param int: the LDAP error code of the first error, if available
 * @param bool: indicates whether the LDAP error code was filled in or not
 * @param bool: indicates whether the first error was a filure of ldap_init()
 * returns bool: indicates whether the open eventually succeeds.
 */
bool LDAPSynchConnection::l_open(int& ldap_first_error,bool& got_error_code,bool& could_not_init)
{
  bool result = false;
  int rc;

  close();

  got_error_code = false;
  could_not_init = false;

  LDAP* h_ldap = NULL;
  while(!result && failures <= max_retries) {

    if (h_ldap != NULL) {
      ldap_unbind(h_ldap);
      h_ldap=NULL;
    }

    if ((h_ldap = ldap_init( const_cast<char*>(source_name.c_str()), source_port)) != NULL) {
    
      ldap_set_option(h_ldap, LDAP_OPT_NETWORK_TIMEOUT, &timeout);
    
      if( (rc=ldap_simple_bind_s(h_ldap,0,0)) == LDAP_SUCCESS )  {      
        handle = h_ldap;
        result  = true;
      } else if (!got_error_code && !could_not_init) {
        got_error_code = true;
	ldap_first_error = rc;
      }
    } else if (!got_error_code && !could_not_init) {
      could_not_init = true;
    }

    if (!result) l_note_failure();
  }

  if (result) return true;

  // connection failed

  if (h_ldap != NULL)
    ldap_unbind(h_ldap);

  return false;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// Search
//
generic_search_result_t* LDAPSynchConnection::execute( LDAPQuery* query ) 
{
  std::auto_ptr<generic_search_result_t> search_result;
  bool result = false;
  bool got_any_error = false;
  bool got_error_code = false;
  int ldap_first_error=0,rc;

  int    scope  = query -> scope();
  char*  filter = (char*)query -> filter().c_str();

#ifdef WITH_LDAP_EXCEPTIONS
  bool got_exception = false;
  bool local_exception = false;
  int e_code = 0;
  std::string e_source,e_method,e_message;
  std::string source_info;
  {
    utilities::oedgstrstream source;	    
    source << "contact=" << source_name << ":" << source_port << ", dn=" << base_dn  
           << "filter='" << filter << "'";
    source_info = source.str();
  }
#endif

  if (! is_established() ) return NULL;

  saved_error = false;
  failures = 0;

  char** topics = make_topics( query -> topics() );
    
  while(1) {
    LDAPMessage *ldresult = NULL;

    while(!result && is_established() && failures <= max_retries) {
      rc = ldap_search_st( handle,
      (char*)base_dn.c_str(), 
      scope,
      filter,
      topics,
      false,
      &timeout, 
      &ldresult );

      if ( rc == LDAP_SUCCESS ) {
        result = true;
      } else {
        if (!got_any_error) {
          got_error_code = true;
	  got_any_error = true;
	  ldap_first_error = rc;

#ifdef WITH_LDAP_EXCEPTIONS
	  got_exception = true;
	  local_exception = true;
	  std::string error_msg( ldap_err2string(ldap_first_error) );

	  e_code = LDAP_QUERY_ERROR;
	  e_source = source_info;
	  e_method.assign("ldap_search_st");
	  e_message = error_msg;
#endif
        }
	l_note_failure();

	{ int l_rc; bool l_b1, l_b2; l_open(l_rc,l_b1,l_b2); }

      }
    }

    if (!result) break;

    // query was made OK

#ifdef WITH_LDAP_EXCEPTIONS
    try {
#endif
    search_result.reset(new ldap_search_result_t( ldresult, handle ));
#ifdef WITH_LDAP_EXCEPTIONS
    } catch(Exception &exc) {
      if (!got_any_error) {
        got_any_error = true;
	got_exception = true;
	local_exception = false;
	e_code = exc.getCode();
	e_message = exc.what();
	result = false;
      }
    }
#endif
    if (search_result.get() == 0 || !search_result->good()) result=false;
    if (!result && search_result.get() != 0) {
      search_result.reset();
    }

    if (result) break;

    // get query results failed

    if ( ldresult != NULL)
      ldap_msgfree(ldresult);

    l_note_failure();

    if (!got_any_error) {
      got_any_error = true;

#ifdef WITH_LDAP_EXCEPTIONS
      got_exception = true;
      local_exception = true;
      std::string str("Unspecified error creating results");

      e_code = LDAP_QUERY_ERROR;
      e_source = source_info;
      e_method.assign("ldap_search_result_t");
      e_message = str;
#endif
    }

    { int l_rc; bool l_b1, l_b2; l_open(l_rc,l_b1,l_b2); }

  } // while(1)

  delete [] topics;

  if (!result) {
    if (got_error_code) {
      saved_error = true;
      ldap_last_error = ldap_first_error;
    }

#ifdef WITH_LDAP_EXCEPTIONS
    if (got_exception) {
      if (!local_exception) {
        e_source = source_info;
	e_method.assign("during LDAPSynchConnection::execute");
      }
      switch(e_code) {
        case LDAP_CONNECTION_ERROR:
          throw ConnectionException(e_source,e_method,e_message);
	  break;
	case LDAP_QUERY_ERROR:
          throw QueryException(e_source,e_method,e_message);
          break;
	case LDAP_UNDEFINED_VALUE_ERROR:
	  throw UndefinedValueException(e_source,e_method,e_message);
	  break;
	default:
	  break;
      }
    }
#endif
  }

  return search_result.release();
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// Close
//
bool LDAPSynchConnection::close()
{
  bool return_status = ( is_established() && 
			 ldap_unbind( handle ) == LDAP_SUCCESS);
   
  handle = NULL;
  
  return return_status; 
}


/**
 * Topics creator.
 * Creates a vector of pointer to topics to us AP Connection 
 * for record satisfing a filter, having a specified scope.
 */
char** LDAPSynchConnection::make_topics(const std::vector<std::string>& topic)
{
	std::vector<std::string>::const_iterator it;
  
  int i = 0;
  char** t = new char*[ topic.size()+1 ];
  
  for( it = topic.begin(); it != topic.end(); it++) {
    
    t[i++] = const_cast<char*>( (*it).c_str() );
  }
  t[i]=NULL;
  return (t);
}

/**
 * Shows wheter tconnection is established or not.
 * @return true if connection is established, false otherwise.
 */
bool LDAPSynchConnection::is_established() const 
{  
  return handle != NULL;
}

} // namespace ldif2classad 
} // namespace common
} // namespace workload
} // namespace edg 
