// File: convert.cpp
// Author: Francesco Giacomini <Francesco.Giacomini@cnaf.infn.it>
// Copyright (c) 2002 EU DataGrid.
// For license conditions see http://www.eu-datagrid.org/license.html

// $Id: convert.cpp,v 1.9 2003/05/30 16:49:51 fpacini Exp $

#include "edg/workload/common/requestad/convert.h"
#include <numeric>
#include <algorithm>
#include <set>
#include <iostream>
#include <boost/tuple/tuple.hpp>
#include <boost/lexical_cast.hpp>
#include <classad_distribution.h>
//#include "edg/workload/common/requestad/DAGAd.h"
#include "edg/workload/common/utilities/classad_utils.h"

using namespace std;
using namespace classad;

namespace edg {
namespace workload {
namespace common {
namespace requestad {

namespace {
// condor submit file keywords
const char* c_condor_keys_begin[] = {
  "allow_startup_script",
  "append_files",
  "arguments",
  "buffer_block_size",
  "buffer_files",
  "buffer_size",
  "cluster",
  "compress_files",
  "copy_to_spool",
  "coresize",
  "dagman_job_id",
  "dag_node_name",
  "environment",
  "error",
  "executable",
  "exit_requirements",
  "fetch_files",
  "file_remaps",
  "getenv",
  "globus_resubmit",
  "globusrsl",
  "globusscheduler",
  "hold",
  "image_size",
  "initialdir",
  "input",
  "jar_files",
  "kill_sig",
  "local_files",
  "log",
  "log_xml",
  "machine_count",
  "nice_user",
  "notification",
  "notify_user",
  "on_exit_hold",
  "on_exit_remove",
  "output",
  "parallel_script_shadow",
  "parallel_script_starter",
  "periodic_hold",
  "periodic_release",
  "periodic_remove",
  "preferences",
  "priority",
  "process",
  "rank",
  "remote_initialdir",
  "remove_kill_sig",
  "rendezvousdir",
  "requirements",
  "rootdir",
  "submit_event_notes",
  "submit_user_event_notes",
  "transfer_error",
  "transfer_executable",
  "transfer_files",
  "transfer_input",
  "transfer_input_files",
  "transfer_output",
  "transfer_output_files",
  "universe",
  "x509userproxy"
};

const char** c_condor_keys_end = c_condor_keys_begin +
sizeof(c_condor_keys_begin) / sizeof(c_condor_keys_begin[0]);

const set<string> condor_keys(c_condor_keys_begin, c_condor_keys_end);

// ostream*
// InsertAttributeInSubmitFile_i(ostream* osp, const string& key,
//                               Literal* const literal)
// {
//   ostream& os = *osp;
//   Value value;
//   literal->GetValue(value);

//   switch (value.GetType()) {
//   case Value::BOOLEAN_VALUE:
//     bool b;
//     value.IsBooleanValue(b);
//     os << key << " = " << (b ? "True" : "False") << "\n";
//     break;
//   case Value::INTEGER_VALUE:
//     int i;
//     value.IsIntegerValue(i);
//     os << key << " = " << i << "\n";
//     break;
//   case Value::REAL_VALUE:
//     double d;
//     value.IsRealValue(d);
//     os << key << " = " << d << "\n";
//     break;
//   case Value::STRING_VALUE: {
//     string s;
//     value.IsStringValue(s);
//     os << key << " = " << s << "\n";
//     break;}
//   default:
//     break;
//   }

//   return osp;
// }

bool
eq_char_nocase(char c1, char c2)
{
  return tolower(c1) == tolower(c2);
}

struct eq_string_nocase: binary_function<string, string, bool>
{
  bool operator()(const string& s1, const string& s2) const
  {
    if (s1.size() < s2.size()) {
      return equal(s1.begin(), s1.end(), s2.begin(), eq_char_nocase);
    } else {
      return equal(s2.begin(), s2.end(), s1.begin(), eq_char_nocase);
    }
  }
};

std::string stringize_literal(Literal const* literal, bool quote_strings)
{
  std::string result;

  Value value;
  literal->GetValue(value);

  switch (value.GetType()) {
  case Value::BOOLEAN_VALUE:
    bool b;
    value.IsBooleanValue(b);
    result = (b ? "True" : "False");
    break;
  case Value::INTEGER_VALUE:
    int i;
    value.IsIntegerValue(i);
    result = boost::lexical_cast<std::string>(i);
    break;
  case Value::REAL_VALUE:
    double d;
    value.IsRealValue(d);
    result = boost::lexical_cast<std::string>(d);
    break;
  case Value::STRING_VALUE: {
    std::string s;
    value.IsStringValue(s);
    if (quote_strings) {
      result = "\"" + s + "\"";
    } else {
      result = s;
    }
    break;}
  default:
    break;
  }

  return result;
}


struct InsertAttributeInSubmitFile:
  binary_function<ostream*, AttrList::value_type, ostream*>
{
  ostream* operator()(ostream* osp, const classad::AttrList::value_type& key_value)
  {
    switch (key_value.second->GetKind()) {

    case ExprTree::LITERAL_NODE: {

      std::string key;
      bool quote_strings = false;

      if (eq_string_nocase()(key_value.first, string("error_"))) {
          key = "error";
      } else if (count_if(condor_keys.begin(), condor_keys.end(),
                          bind1st(eq_string_nocase(), key_value.first)) > 0) {
        key = key_value.first;
      } else {
        key = "+" + key_value.first;
        quote_strings = true;           // quote user-defined attribute values
      }

      assert(! key.empty());

      *osp << key << " = "
           << stringize_literal(dynamic_cast<Literal*>(key_value.second), quote_strings)
           << "\n";
//       InsertAttributeInSubmitFile_i(osp, key,
//                                     dynamic_cast<Literal*>(key_value.second));
      break;
    }

    default:
      // nothing for the moment
      break;
      
    }
  
    return osp;
  }

};

struct InsertAttributeInDAGFile:
  binary_function<ostream*, AttrList::value_type, ostream*>
{
  ostream* operator()(ostream* osp, const classad::AttrList::value_type& key_value)
  {
    if (eq_string_nocase()(key_value.first, string("type"))) return osp;

    switch (key_value.second->GetKind()) {

    case ExprTree::CLASSAD_NODE: { // job description
      const ClassAd* job_desc_ad =
        dynamic_cast<const ClassAd*>(key_value.second);
      string jdl_file;
      assert(job_desc_ad->EvaluateAttrString(string("classad_file"), jdl_file));
      assert(! jdl_file.empty());
      *osp << "JOB " << key_value.first << " " << jdl_file << "\n";
      break;
    }

    default:
      break;
    }

    return osp;
  }

};

}

std::ostream&
to_submit_stream(std::ostream& os, const ClassAd& ad)
{
  std::string type;
  assert(ad.EvaluateAttrString("type", type));
  assert(! type.empty());

  if (type == "job") {
    std::accumulate(ad.begin(), ad.end(), &os, InsertAttributeInSubmitFile());
    os << "Queue 1" << endl;
//   } else if (type == "dag") {
//     DAGAd dag_ad(ad);
//     to_submit_stream(os, dag_ad);
  }

  return os;
}

// struct InsertJobInClassAd: binary_function<ClassAd*, DAGAd::job_value_type, ClassAd*>
// {
//   ClassAd* operator()(ClassAd* ad, const DAGAd::job_value_type& job)
//   {
//     assert(ad->Insert(job.first, new ClassAd(*job.second)));

//     return ad;
//   }
// };

// struct InsertDependencyInClassAd:
//   binary_function<vector<ExprTree*>*, DAGAd::dependency_value_type, vector<ExprTree*>*>
// {
//   vector<ExprTree*>* operator()(vector<ExprTree*>* dependencies, const DAGAd::dependency_value_type& dependency)
//   {
//     // transform the first end of a dependency in a classad Literal
//     Value dependency_first_value;
//     dependency_first_value.SetStringValue(dependency.first);
//     ExprTree* dependency_first_literal = Literal::MakeLiteral(dependency_first_value);
    
//     // transform the second end of a dependency in a classad Literal
//     Value dependency_second_value;
//     dependency_second_value.SetStringValue(dependency.second);
//     ExprTree* dependency_second_literal = Literal::MakeLiteral(dependency_second_value);

//     // build a two-value list from the two literals above
//     vector<ExprTree*> dependency_list;
//     dependency_list.push_back(dependency_first_literal);
//     dependency_list.push_back(dependency_second_literal);

//     dependencies->push_back(ExprList::MakeExprList(dependency_list));
    
//     return dependencies;
//   }
// };

// void
// to_classad(ClassAd& class_ad, const DAGAd& dag_ad)
// {
//   class_ad.InsertAttr(string("type"), string("dag"));

//   accumulate(dag_ad.jobs().first, dag_ad.jobs().second, &class_ad, InsertJobInClassAd());
//   vector<ExprTree*>* dependencies = accumulate(dag_ad.dependencies().first,
//                                                dag_ad.dependencies().second,
//                                                new vector<ExprTree*>,
//                                                InsertDependencyInClassAd());

//   class_ad.Insert("dependencies", ExprList::MakeExprList(*dependencies));
// }

// struct InsertDependencyInDAGAd:
//   binary_function<DAGAd*, ExprTree*, DAGAd*>
// {
//   DAGAd* operator()(DAGAd* dag_ad_p, const ExprTree* value)
//   {
//     // value should be a list of two string elements

//     assert(value->GetKind() == ExprTree::EXPR_LIST_NODE);
//     const ExprList* dependency = dynamic_cast<const ExprList*>(value);

//     string first_node_name;
//     string second_node_name;

//     // extract the name of the first node
//     ExprList::const_iterator node_it = dependency->begin();
//     assert(node_it != dependency->end());
//     assert((*node_it)->GetKind() == ExprTree::LITERAL_NODE);
//     const Literal* lit = dynamic_cast<const Literal*>(*node_it);
//     Value v;
//     lit->GetValue(v);
//     bool first_node_is_string = v.IsStringValue(first_node_name);
//     assert(first_node_is_string);

//     // extract the name of the second node
//     ++node_it;
//     assert(node_it != dependency->end());
//     assert((*node_it)->GetKind() == ExprTree::LITERAL_NODE);
//     lit = dynamic_cast<const Literal*>(*node_it);
//     lit->GetValue(v);
//     bool second_name_is_string = v.IsStringValue(second_node_name);
//     assert(second_name_is_string);

//     // there is no third element in the list
//     ++node_it;
//     assert(node_it == dependency->end());

//     assert(dag_ad_p->add_dependency(first_node_name, second_node_name));

//     return dag_ad_p;
//   }

// };

// struct InsertJobInDAGAd:
//   binary_function<DAGAd*, AttrList::value_type, DAGAd*>
// {
//   DAGAd* operator()(DAGAd* dag_ad_p, const AttrList::value_type& key_value)
//   {
//     DAGAd& dag_ad = *dag_ad_p;
//     const string key = key_value.first;
//     const ExprTree* value = key_value.second;

//     switch (key_value.second->GetKind()) {

//     case ExprTree::CLASSAD_NODE: { // job description

//       const ClassAd* job_desc = dynamic_cast<const ClassAd*>(value);

//       ClassAd* ad = new ClassAd(*job_desc);
//       assert(dag_ad.add_job(key, ad));

//       break;
//     }

//     default:
//       // ignore the rest
//       break;
//     }

//     return dag_ad_p;
//   }
// };

// void
// to_dagad(DAGAd& dag_ad, const ClassAd& class_ad)
// {
//   // insert first job descriptions...
//   accumulate(class_ad.begin(), class_ad.end(),
//              &dag_ad, InsertJobInDAGAd());

//   // ... and then the dependencies
//   const ExprTree* dep_tree = class_ad.Lookup(string("dependencies"));
//   if (dep_tree != 0 && dep_tree->GetKind() == ExprTree::EXPR_LIST_NODE) {
//     const ExprList* dep_list = dynamic_cast<const ExprList*>(dep_tree);
//     accumulate(dep_list->begin(), dep_list->end(),
//                &dag_ad, InsertDependencyInDAGAd());
//   }
// }

// namespace {

// struct InsertJobInSubmitFile:
//   binary_function<std::ostream*, DAGAd::node_value_type, std::ostream*>
// {
//   std::ostream* operator()(std::ostream* osp, const DAGAd::node_value_type& node) const
//   {
//     std::ostream& os = *osp;

//     std::string node_name = node.first;
//     ClassAd const& node_info = dynamic_cast<ClassAd const&>(*node.second);

//     std::string file = utilities::evaluate_attribute(node_info, "file");
//     os << "JOB " << node_name << " " << file << "\n";

//     std::string pre_script;
//     if (node_info.EvaluateAttrString("pre_script", pre_script)) {
//       os << "SCRIPT PRE " << node_name << " " << pre_script << "\n";
//     }

//     string post_script;
//     if (node_info.EvaluateAttrString("post_script", post_script)) {
//       os << "SCRIPT POST " << node_name << " " << post_script << "\n";
//     }

//     int retry;
//     if (node_info.EvaluateAttrInt("retry", retry)) {
//       os << "RETRY " << node_name << " " << retry << "\n";
//     }

//     // TODO: manage pre/post script arguments

//     return osp;
//   }
// };

// }

// struct InsertDependencyInSubmitFile:
//   binary_function<ostream*, DAGAd::dependency_value_type, ostream*>
// {
//   ostream* operator()(ostream* osp,
//                       const DAGAd::dependency_value_type& dependency) const
//   {
//     ostream& os = *osp;

//     os << "PARENT " << dependency.first
//        << " CHILD " << dependency.second
//        << "\n";

//     return osp;
//   }
// };

// std::ostream&
// to_submit_stream(std::ostream& os, const DAGAd& ad)
// {
//   DAGAd::node_iterator node_b;
//   DAGAd::node_iterator node_e;
//   boost::tie(node_b, node_e) = ad.nodes();
//   std::accumulate(node_b, node_e, &os, InsertJobInSubmitFile());

// //   accumulate(ad.dependencies().first, ad.dependencies().second, &os,
// //              InsertDependencyInSubmitFile());

//   return os;
// }

}}}} // edg::workload::common::requestad
