/***************************************************************************
 *  filename  : LDAPConnection.h
 *  authors   : Salvatore Monforte <salvatore.monforte@ct.infn.it>
 *  copyright : (C) 2001 by INFN
 ***************************************************************************/

/**
 * @file LDAPConnection.h
 * @brief This file contains definition for LDAP Connections.
 * @author Salvatore Monforte salvatore.monforte@ct.infn.it
 * @author comments by Marco Pappalardo marco.pappalardo@ct.infn.it
 */

// $Id:

#ifndef _LDAPCONNECTION_
#define _LDAPCONNECTION_

#include <string>

#include "lber.h"	/**< LDAP header files. */
#include "ldap.h"       /**< LDAP header files. */

using namespace std;

/**
 * Defines an LDAP Connection.
 * This class models a basic standard connection to an LDAP Server.
 * It will be the common superclass for both asynchronous- and sinchrnonous-
 * connection classes.
 * @author Salvatore Monforte salvatore.monforte@ct.infn.it
 * @auhor comments by Marco Pappalardo marco.pappalardo@ct.infn.it
 * @see LDAPSynchConnection.
 */
class LDAPConnection
{
 public:
  
  /**
   * Opens a connection to the Server. 
   * This function must be overridden by all classes extending basic LDAP Connection.
   * @return false.
   */
  virtual bool Open() = 0;
  /**
   * Closes a connection to the Server. 
   * This function must be overridden by all classes extending basic LDAP Connection.
   * @return false.
   */  
  virtual bool Close() = 0;
  /**
   * Searches a string in a message.    
   * This function must be overridden by all classes extending basic LDAP Connection.
   * @return false.
   */  
  virtual bool Search(const std::string&, int, char**, LDAPMessage**) = 0;
  /**
   * Returns the handle of the connection.
   * @return the handle.
   */
  LDAP* GetHandle() const{ return handle; }
  /**
   * Shows wheter tconnection is established or not.
   * If handle is a not null pointer, the connection certainly does.
   * @return true if connection is established, false otherwise.
   */
  bool IsEstablished() { return handle != NULL; }
 
 protected:

  /**
   * Constructor.
   * @param b the base DN string.
   * @param hn the host name.
   * @parma p the port.
   */
  LDAPConnection(const std::string& b, const std::string& hn, const int& p) :
    baseDN( b ), hostname( hn ), port( p ), handle( NULL ) {}
  /** 
   * Destructor.
   */
  virtual ~LDAPConnection() {}
 protected:

  /** The base DN. */
  std::string baseDN;
  /** The host name. */
  std::string hostname;
  /** The port. */
  int port;
  /** The connection handle.*/
  LDAP *handle;
};

#endif

/*
  Local Variables:
  mode: c++
  End:
*/




