/***************************************************************************
 *  filename  : LDAPForwardIterator.cpp
 *  authors   : Salvatore Monforte <salvatore.monforte@ct.infn.it>
 *  copyright : (C) 2001 by INFN
 ***************************************************************************/

// $Id:


/**
 * @file LDAPForwardIterator.cpp
 * @brief Contains implementations for LDAP Forward Iterator.
 * @author Salvatore Monforte salvatore.monforte@ct.infn.it
 * @author comments by Marco Pappalardo marco.pappalardo@ct.infn.it  and Salvatore Monforte
 */

/** This class header file. */
#include "LDAPForwardIterator.h"
/** The LDIF header file. */
#include "LDIFObject.h"
/** The LDAP Query header file. */
#include "LDAPQuery.h"
/** The LDAP Connection header file. */
#include "LDAPConnection.h"

/**
 * Constructor
 * @param q a reference to an LDAPQuery result object.
 */
LDAPForwardIterator::LDAPForwardIterator(LDAPQuery* q) 
{
  query = q;
  ldresult = NULL;
  ld_current_result = NULL;
  object = NULL;
}

/** 
 * Returns the DN attribute for LDAP.
 */ 
const string& LDAPForwardIterator::ldap_dn()
{
  LDAP* ld = query -> connection-> GetHandle();
  char *ptr = ldap_get_dn(ld, ldresult);
  dn = string(ptr);
  ber_memfree(ptr);
  return dn;
}

/**
 * Rewinds the LDAP forward iterator.
 * After use of this method, the forward iterator is addressed 
 * to the first query result record.
 * @return true if first query entry is reached, false otherwise.
 */
bool LDAPForwardIterator::rewind()
{
  bool result;
  
  if(ldresult) {
    ldap_msgfree(ldresult);
    ldresult = NULL;
  }
	
  if(object) { 
	
    delete object; 
    object = NULL;
  } 

  if( (result = query -> connection->
       Search(query -> filter, 
	      query -> scope, 
	      query -> topic, 
	      &ldresult)) ) {
    
    LDAP* ld = query -> connection-> GetHandle();
  
    result = ((ld_current_result = ldap_first_entry(ld,ldresult)) != NULL);
  }
  return result;
}

/**
 * Moves to next record.
 * Redefines '++' operator in order to provide a simple way to
 * move forward inside the LDAP Query.
 * @return if next entry exists, false otherwise.
 */
bool LDAPForwardIterator::operator++()
{
  if(object) {
 
    delete object;
    object = NULL;
  }  
  
  LDAP* ld = query -> connection-> GetHandle();
  ld_current_result = ldap_next_entry(ld,ld_current_result);
  return ld_current_result != NULL;
}

const LDIFObject& LDAPForwardIterator::operator*()
{
  LDAP* ld = query -> connection-> GetHandle();
  BerElement *ber;

  if(object) {
 
    delete object;
    object = NULL;
  }  
  object = new LDIFObject();

  for(char *attribute = ldap_first_attribute(ld, ld_current_result, &ber); 
      attribute; attribute = ldap_next_attribute(ld, ld_current_result, ber) ) {

    char **values;
    values = ldap_get_values(ld, ld_current_result, attribute);
    
    for(int i=0; values[i]!=NULL; i++)
    
      object -> add( (string)attribute, (string)values[i] );
    
    ber_memfree(attribute);
    ldap_value_free(values);
  }
  
  if( ber != NULL ) {
    ber_free( ber, 0 );
  }

  return *object;
}

/**
 * Destructor.
 */
LDAPForwardIterator::~LDAPForwardIterator()
{
  if(object) delete object;
  if(ldresult) ldap_msgfree(ldresult);
}

