/***************************************************************************
 *  filename  : LDAPSynchConnection.cpp
 *  authors   : Salvatore Monforte <salvatore.monforte@ct.infn.it>
 *  copyright : (C) 2001 by INFN
 ***************************************************************************/

/**
 * @file LDAPSynchConnection.cpp
 * @brief This files contains LDAP Synchronous connections implementation.
 * @author Salvatore Monforte salvatore.monforte@ct.infn.it
 * @author comments by Marco Pappalardo marco.pappalardo@ct.infn.it and Salvatore Monforte
 */
// $Id:

/** Includes header definitions. */
#include "LDAPSynchConnection.h"

/**
 * Constructor.
 * @param b the base DN of the entry at which to start operations.
 * @param hn the host name of the LDAP server.
 * @param p the port of the LDAP server.
 * @param t the timeout used during binding to an LDAP server and searching its content.
 */
LDAPSynchConnection::LDAPSynchConnection(const string& b, const string& hn, int p, long t) :
  LDAPConnection(b, hn, p)
{
  handle = NULL;
  timeout.tv_sec = t;
  timeout.tv_usec = 0;
}

/**
 * Destructor.
 */
LDAPSynchConnection::~LDAPSynchConnection()
{
  Close();
}


/**
 * Opens a connection to the Server. 
 * This function overrides virtual LDAP Connection one.
 * @return true on success, false otherwise.
 */
bool LDAPSynchConnection::Open()
{
  bool result = false;
  
  Close();

  if ( (handle = ldap_init( const_cast<char*>(hostname.c_str()), port) ) ) {

        ldap_set_option( handle, LDAP_OPT_NETWORK_TIMEOUT, &timeout);

        if( (ldap_simple_bind_s(handle,0,0) == LDAP_SUCCESS) ) {

                result = true;
        }
  }
  else {

    if( handle ) {
      
      ldap_unbind(handle);
      handle = NULL;
    }
  }
  return result;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// Search
//
bool LDAPSynchConnection::Search(const string& filter, int scope, char** topic, LDAPMessage** ldresult)
{
  bool result = false;

  if( IsEstablished() ) {
	
        if( *ldresult ) { 
 		
		ldap_msgfree(*ldresult);
		*ldresult = NULL;       
        }	
	
        result = ( ldap_search_st( handle,  
			const_cast<char*>(baseDN.c_str()), 
			scope,
			const_cast<char*>(filter.c_str()),
			topic,
			false,
			&timeout, 
			ldresult) == LDAP_SUCCESS );

  }

  return result;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// Close
//
bool LDAPSynchConnection::Close()
{
  bool return_status = (IsEstablished() && ldap_unbind(handle) == LDAP_SUCCESS);
   
  handle = NULL;

  return return_status; 
}
