/***************************************************************************
 *  filename  : LDIFObject.h 
 *  authors   : Salvatore Monforte <salvatore.monforte@ct.infn.it>
 *  copyright : (C) 2001 by INFN
 ***************************************************************************/

// $Id:

/**
 * @file LDIFObject.cpp
 * This file contains LDIF Object implementation.
 * @author Salvatore Monforte salvatore.monforte@ct.infn.it
 * @author comments by Marco Pappalardo marco.pappalardo@ct.infn.it and Salvatore Monforte
 */

#ifndef _LDIFOBJECT_
#define _LDIFOBJECT_

#include <iostream>
#include <map>
#include <string>
#include <sstream>
#include <vector>


#define LEFT  1
#define RIGHT 2

using namespace std;

/** Defines an LDIFValue type. */
typedef vector<string> LDIFValue;
/** Defines an LDIFAttribute type. */
typedef map<string,LDIFValue> LDIFAttributes;


/**
 * An LDIF Object.
 * @author Salvatore Monforte salvatore.monforte@ct.infn.it
 * @author comments by Marco Pappalardo marco.pappalardo@ct.infn.it and Salvatore Monforte
 */
class LDIFObject
{
 public:

  /**
   * Empty Constructor.
   * It does nothing.
   */
  LDIFObject();

  /**
   * Constructor.
   * Creates a new LDIF Object reusing attributes of an existing one.
   * @param o teh existing LDIF Object.
   */
  LDIFObject(const LDIFObject&);
  /** 
   * Redefines '=' operator between LDIF Objects.
   * It establishes that two LDIF object are the same if
   * their attributes are the same.
   * @param o the LDIF Object to compare.
   */  
  LDIFObject& operator= (const LDIFObject& o );
  
  /**
   * Adds an attribute.
   * It pushes back the val attribute in the att item in attributes map.
   * @param att the attribute item.
   * @param the value to insert.
   */
  void add(const string&, const string&);


  /**
   * Evaluates an attribute.
   * If the string name exists in attribtues map, then the front value is put in value string.
   * @param name the attribute map item.
   * @param value the destination string.
   * @return true for sucessful evaluation, false otherwise.
   */
  bool EvaluateAttribute(const string&, vector<string>&) const;

  /**
   * Evaluates an attribute set.
   * If the string name exists in attribtues map, then all values are put in value vector.
   * @param name the attribute map item.
   * @param value the destination vector.
   * @return true for sucessful evaluation, false otherwise.
   */  
  bool EvaluateAttribute(const string&, string&) const;	

 private:
  void ParseValue( const std::string&, stringstream& ) const;
  void ParseMultiValue( const LDIFValue&, stringstream& ) const; 
/**
 * Writes the LDIFObject to a stream.
 * Serializes LDIF Object info into an output stream.
 * @param s the output stream.
 * @param o the LDIF Object.
 * @return the written ostream.
 */
  friend ostream& operator << (ostream&, const LDIFObject&); 

  /** Attributes vector. */
 mutable LDIFAttributes attributes;
};

/**
 * Writes the LDIFObject to a stream.
 * Serializes LDIF Object info into an output stream.
 * @param s the output stream.
 * @param o the LDIF Object.
 * @return the written ostream.
 */
extern ostream& operator<<        (ostream&, const LDIFObject&);


#endif

/*
  Local Variables:
  mode: c++
  End:
*/

