// DGAS (DataGrid Accounting System) 
// Programm for setting suggested resource prices in the PA database.
// A manual pricing scheme has to be specified in the PA's configuration file
// in order to use this suggested resource price.
// 
// $Id: pa_ResAdmin.cpp,v 1.2 2003/12/02 13:50:06 guarise Exp $
// -------------------------------------------------------------------------
// Copyright (c) 2001-2002, The DataGrid project, INFN, 
// All rights reserved. See LICENSE file for details.
// -------------------------------------------------------------------------
// Author: Rosario Piro <piro@to.infn.it>
 /***************************************************************************
 * Code borrowed from: pa_engine_test.cpp dgas_paResItem.cpp
 *  authors   : Andrea Guarise <andrea.guarise@to.infn.it>
 *  copyright : Copyright (c) 2001-2002, The DataGrid project, INFN
 ***************************************************************************/
//
//

// ---------------------------------------------------------------------------
//  Includes
// ---------------------------------------------------------------------------
#include <string>
#include <vector>
#include <iostream>
#include <getopt.h>
#include <cstring>
#include "edg/workload/dgas/common/dgas_config.h"
#include "edg/workload/dgas/pa_api/pa_db_default.h"
#include "edg/workload/dgas/services/priceAlgorithms/edg-wl-dgas-paPriceType.h"
#include "edg/workload/dgas/pa_api/dgas_paResource.h"
#include <db.h>

using namespace std;

#define PRA_VERSION "0.1.0"

// error codes
#define USAGE_ERR       1
#define RES_PUT_ERR     2
#define RES_REPLACE_ERR 3
#define CONFIGFILE_ERR  4

#define UNSET -42 // just a dummy value
#define CONF_PATH "/opt/edg/etc/edg-wl-dgas-pa.conf" // edg default pa conf file

#define  OPTION_STRING "hc:r:fp:m:"

string confFName = "";      // name of the PA configuration file
string resID = "";          // grid ID of resource. maybe also "all".
bool force = false;         // force setting of prices
bool needsHelp = false;     // display help screen
price_t resPrice = UNSET;
int minTTL = UNSET;
vector<string> info_v;

const char * pa_sql_server;
const char * pa_sql_user;
const char * pa_sql_password;
const char * pa_sql_dbname;

const char * pa_def_price_ttl;


void help()
{
  cerr << "DGAS pa_resource_admin:" << endl << "Version " << PRA_VERSION <<endl
       << "Modifies resource records in the PA database. If no record for the "
       << "resource" << endl << "with <resID> exists, a new one will be created. The suggested price"
       << " will be used" << endl << "only if a manual pricing scheme is"
       << " specified in the PA's configuration file!" << endl;
  cerr << "Author: Rosario M. Piro <piro@to.infn.it>" << endl << endl;
  cerr << "Usage:" << endl;
  cerr << "pa_resource_admin <OPTIONS>" << endl;
  cerr << "Where options are:" << endl;
  cerr << "-h  --help             :this help" << endl;
  cerr << "-c  --conf <filename>  :PA configuration file (optional)" << endl;
  cerr << "-r  --resource <resID> :grid ID of resource. Instead of a specific "
       << "ID, 'all'" << endl << "                        may be specified, "
       << "this requires the --force option" << endl;
  cerr << "-f  --force            :force the modification (mandatory for -r all)" << endl;
  cerr << "-p  --price <value>    :resource price to be set (optional if -m specified)" << endl;
  cerr << "-m  --minTTL <value>   :time-to-live in seconds for resource price"
       << endl << "                        values below " << MIN_PRICE_TTL
       << " will be set to " << MIN_PRICE_TTL << endl
       << "                        (optional if -p specified)" << endl;
  cerr << endl;

}

int options ( int argc, char **argv )
{
  int option_char;
  int option_index = 0;
  static struct option long_options[] =
  {
    {"conf",1,0,'c'},
    {"force",0,0,'f'},
    {"help",0,0,'h'},
    {"resource",1,0,'r'},
    {"price",1,0,'p'},
    {"minTTL",1,0,'m'},
    {0,0,0,0}
  };
  while (( option_char = getopt_long( argc, argv, OPTION_STRING,
				      long_options, &option_index)) != EOF)
    switch (option_char)
      {
        case 'c': confFName = optarg;
	          break;
        case 'f': force = true;
                  break;
        case 'h': needsHelp = true;
	          break;
        case 'r': resID = optarg;
	          break;
        case 'p': resPrice = c_strToPrice(optarg);
	          break;
        case 'm': minTTL = atoi(optarg);
		  break;
        default : break;
      }
  if (optind < argc)
    {
      while (optind < argc)
	info_v.push_back(argv[optind++]);
    }
  return 0;
}


int main (int argc, char *argv[])
{
  confFName =  CONF_PATH;  // CONF_PATH passed by compiler option -D
  confFName += "/pa.conf"; // default PA configuration file

  // get command line options
  options(argc, argv);

  // display help screen if necessary
  if ( argc < 2 )
    needsHelp = true; 
  if ( needsHelp )
    {
      help();
      return 0;
    }

  // check that resID has been specified!
  if (resID == "") {
    cerr << "No resource ID specified, nothing to be done ..." << endl;

    return USAGE_ERR;
  }

  // check if there is something to do ...
  if (minTTL == UNSET && resPrice == UNSET) {
    cerr << "At least one of the options --price and --minTTL has to be "
	 << "specified." << endl;

    return USAGE_ERR;
  }


  // get info from configuration file
  map <string,string> confMap;
  if (dgas_conf_read ( confFName, &confMap ) == 0) { // config file parsed
    pa_sql_server = (confMap["pa_sql_server"]).c_str();
    pa_sql_user = (confMap["pa_sql_user"]).c_str();
    pa_sql_password = (confMap["pa_sql_password"]).c_str();
    pa_sql_dbname = (confMap["pa_sql_dbname"]).c_str();
    pa_def_price_ttl = (confMap["def_price_minTTL"]).c_str();
  }
  else if (dgas_conf_read ( confFName, &confMap ) == 1) {
    #ifdef DEBUG
    cerr << "Error: PA configuration file '" << confFName << "' doesn't exist."
	 << endl;
    #endif

    return CONFIGFILE_ERR;
  }
  else {
    #ifdef DEBUG
    cerr << "Error while parsing PA configuration file '" << confFName << "'."
	 << endl;
    #endif

    return CONFIGFILE_ERR;
  }



  if (resID == "all") { // all resources!

    if (!force) {    
      cerr << "To modify ALL resource records, please add the command line"
	   << " option '--force'." << endl;

      return USAGE_ERR;
    }

    // open the database
    db paDb(pa_sql_server,
	    pa_sql_user,
	    pa_sql_password,
	    pa_sql_dbname );

    if ( paDb.errNo == 0 ) // successfully opened
      {
	//insert 
        #ifdef DEBUG
	cerr << "DEBUG: pa_resource_admin: database opened." << endl;
	#endif

	// building database query string
	string queryStr;
	queryStr = "UPDATE res_index SET ";

	if (minTTL != UNSET) { // set minTTL of all resources

	  // make sure we don't set a minTTL below the minimum specified in
	  // pa_db_default.h
	  if (minTTL < MIN_PRICE_TTL)
	    minTTL = MIN_PRICE_TTL;

	  queryStr += "d_time=";
	  queryStr += int2string(minTTL);

          #ifdef DEBUG
	  cerr << "       setting minTTL of all resources to "
	       << minTTL << endl;
	  #endif
	}

	if (resPrice != UNSET) { // set price of all resources

	  if (minTTL != UNSET) queryStr += ","; // seperate the two fields
	  queryStr += "suggested_price=";
	  queryStr += int2string(resPrice);

	  #ifdef DEBUG
	  cerr << "       setting suggested prices of all resources to "
	       << resPrice << endl;
          #endif
	}

	#ifdef DEBUG
	cerr << "DEBUG: pa_resource_admin: the query string is" << endl
	     << "       '" << queryStr << "'" << endl;
	#endif

	// query the DB
	paDb.query(queryStr);
	if ( paDb.errNo != 0 )
	  {
            #ifdef DEBUG
	    cerr << "DEBUG: " << __FILE__ << ":" << __LINE__ << " : ";
	    cerr << "error while modifying the table:" << endl;
	    cerr << paDb.errMsg << endl;
	    #endif 
	    return paDb.errNo;
	    //error in query
	  }

	// should we verify the changes by counting the records that contain
	// the new values (should equal the total number of records)?

      }
    else // DB could not be opened!
      {
	//Error connecting to the database, reporting the error
	#ifdef DEBUG
	cerr << "DEBUG:" << __FILE__ << ":" << __LINE__ << " : ";
	cerr << "Error connecting to the database, got error: " << endl;
	cerr << paDb.errMsg << endl;
	#endif
	return paDb.errNo;
      }

  } // if (resID == "all") ...

  else { // single resource record to be added or modified!

    // at least one of --price and --minTTL was specified, get the other
    // value!

    if (minTTL == UNSET) { // take the current value from the DB or from the
                           // config file if the DB record doesn't exist yet. 
      paResource resBuffer;
      if ( resBuffer.get(resID) == 0 )
	minTTL = resBuffer.defMinTTL; // from DB
      else
	minTTL = atoi(pa_def_price_ttl); // from config file
    }

    if (resPrice == UNSET) { // take the current value from the DB or set it
                             // to 0 if the DB record doesn't exist yet. 
      paResource resBuffer;
      if ( resBuffer.get(resID) == 0 )
	resPrice = resBuffer.suggestedPrice; // from DB
      else
	resPrice = 0; // default
    }


    #ifdef DEBUG
    cerr << "DEBUG: pa_resource_admin with the following options:" 
	 << endl << "       -r " << resID << " -m " << minTTL << " -p " 
	 << resPrice << endl << "       -c " << confFName;
    if (force)
      cerr << " -f";
    cerr << endl;
    #endif

    paResItem resItem(resID);
    // get the right record, if it's already in the DB!
    if ( resItem.get(resID) == 0 ) {
      resItem.defMinTTL = minTTL;
      resItem.suggestedPrice = resPrice;

      if ( resItem.replace() != 0 ) {
	#ifdef DEBUG
	cerr << "DEBUG: pa_resource_admin: could not modify the existing resource record!" << endl;
        #endif
	return RES_REPLACE_ERR;
      }
    }
    else { // new record ...
      #ifdef DEBUG
      cerr << "DEBUG: pa_resource_admin: adding new resource record!" << endl;
      #endif      
      resItem.defMinTTL = minTTL;
      resItem.suggestedPrice = resPrice;

      if ( resItem.put() != 0) {
        #ifdef DEBUG
	cerr << "DEBUG: pa_resource_admin: could not add the new resource record!" << endl;
        #endif
	return RES_PUT_ERR;
      }
    }

  }

  return 0; // no error

}
