// File: JobControllerReal.cpp
// Author: Francesco Giacomini <Francesco.Giacomini@cnaf.infn.it>
//         Rosario Peluso <Rosario.Peluso@pd.infn.it>
// Copyright (c) 2001 EU DataGrid.
// For license conditions see http://www.eu-datagrid.org/license.html

// $Id: JobControllerReal.cpp,v 1.62 2005/09/06 16:02:20 dhsmith Exp $

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>

#include <memory>
#include <iostream>
#include <fstream>
#ifdef HAVE_STRINGSTREAM
#include <sstream>
#else
#include <strstream>
#endif

#include <boost/regex.hpp>
#include <boost/filesystem/path.hpp>

#include <user_log.c++.h>

#include "edg/workload/common/configuration/Configuration.h"
#include "edg/workload/common/configuration/JCConfiguration.h"
#include "edg/workload/common/configuration/LMConfiguration.h"
#include "edg/workload/common/configuration/NSConfiguration.h"
#include "edg/workload/common/utilities/filecontainer.h"
#include "edg/workload/common/jobid/JobId.h"
#include "edg/workload/common/jobid/manipulation.h"
#include "edg/workload/common/requestad/convert.h"
#include "edg/workload/common/requestad/PrivateAdManipulation.h"
#include "edg/workload/common/requestad/JobAdManipulation.h"
#include "edg/workload/common/logger/manipulators.h"
#include "edg/workload/jobcontrol/common/IdContainer.h"
#include "edg/workload/jobcontrol/common/RamContainer.h"
#include "edg/workload/jobcontrol/common/JobFilePurger.h"
#include "edg/workload/jobcontrol/common/ProxyUnregistrar.h"

#include "JobControllerReal.h"
#include "JobControllerExceptions.h"
#include "CondorG.h"
#include "SubmitAd.h"
#include "SubmitAdExceptions.h"

using namespace std;
USING_COMMON_NAMESPACE;

JOBCONTROL_NAMESPACE_BEGIN {

namespace controller {

namespace {

GenericEvent *createGenericEvent( jccommon::generic_event_t ev )
{
  int              evn = (int) ev;
  GenericEvent    *event = (GenericEvent *) instantiateEvent( ULOG_GENERIC );
  string           info( "JC: " );

  info.append( boost::lexical_cast<string>(evn) );
  info.append( " - " );
  info.append( jccommon::generic_events_string[evn] );

  strncpy( event->info, info.c_str(), (sizeof(event->info) - 1) );

  return event;
}

} // Anonymous namespace

// init static member var
int JobControllerReal::jcr_n_fork=0;
std::map<pid_t, int> JobControllerReal::jcr_read_descriptor;

bool JobControllerReal::cancelJob( const string &condorid, bool force, string &info, bool recursive )
{
  bool                     ret = true;
  int                      result;
  string                   parameters;
  logger::StatePusher      pusher( clog, "cancelJob(...)" );
  pid_t                    pid;

  clog << logger::setlevel( logger::debug ) << "Condor id of job was: " << condorid << endl;

  if( force ) {
    clog << logger::setlevel( logger::info ) << "Forcing job removal." << endl;

    parameters.assign( "-f " );
    parameters.append( condorid );
  }
  else parameters.assign( condorid );

  if (!recursive) {
    pid = fork();
    if (pid < 0) {
      // could not fork a  process to do the removal
      char *err = strerror(errno);
      string errstr(err);

      clog << logger::setlevel( logger::severe )
	   << "Error cancelling job." << endl
	   << "could not fork a process to do the cancellation" << endl
	   << logger::setmultiline( true, "CE-> " ) << "Given reason\n" << errstr << endl;
      clog << logger::setmultiline( false );

      return false;
    } else if (pid > 0) {
      ++jcr_n_fork;
      return true;
    }
  }

  // we are a child process. Possibly a recursive call from an existing child
  try {
    result = CondorG::instance()->set_command( CondorG::remove, parameters )->execute( info );
    if( result ) {
      clog << logger::setlevel( logger::severe ) << "Job cancellation (child) refused." << endl
    	   << "Condor ID = " << condorid << endl
	   << logger::setmultiline( true, "CE-> " ) << "Given reason\n" << info << endl;
      clog << logger::setmultiline( false );

      ret = ( force ? false : cancelJob(condorid, true, info, true) );
    }
  } catch(...) {
    if (recursive) throw;
  }

  if (!recursive)
    exit(0);

  return ret;
}

void JobControllerReal::checkChildren( void )
{
  pid_t pid;
  if (jcr_n_fork>0) {
    int status;
    do {
      pid = waitpid(-1, &status, (jcr_n_fork < jcr_max_children) ? WNOHANG : 0);
      if (pid>0) {
        if (jcr_read_descriptor.find(pid) != jcr_read_descriptor.end()) {
          char buf[256];
          ssize_t buf_len = sizeof(buf);
          ssize_t res,len=0;
          bool eof = false;
          do {
            res = read(jcr_read_descriptor[pid], &buf[len], buf_len-len);
            if (res > 0) len += res;
            if (res == 0) eof = true;
          } while( (res==-1 && errno==EINTR) || (res>0 && len<buf_len) );
          if (len>0 && eof) {
            char jobidstr[256],condoridstr[256];
            buf[(len < buf_len) ? len : (buf_len-1)] = '\0';
            sscanf(buf,"%s %s",jobidstr,condoridstr);
            std::string my_jobid(jobidstr);
            std::string condorid(condoridstr);
            if( this->jcr_repository->inserted() >= this->jcr_threshold ) this->readRepository();
            this->jcr_repository->insert( my_jobid, condorid );
            clog << logger::setlevel( logger::verylow )
                 << "Added condor id " << condorid << " for job id " << my_jobid << endl;
          }
          close(jcr_read_descriptor[pid]);
          jcr_read_descriptor.erase(pid);
        } else {
          clog << logger::setlevel( logger::verylow )
               << "Child " << pid << " finished" << endl;
        }
        --jcr_n_fork;
      }
    } while(pid>0 || (pid<0 && errno == EINTR));
  }
}

void JobControllerReal::readRepository( void )
{
  const configuration::LMConfiguration   *lmconfig = configuration::Configuration::instance()->lm();
  string                                  repname( lmconfig->id_repository_name() );
  auto_ptr<jccommon::IdContainer>         repository;
  boost::filesystem::path                 repfile( lmconfig->monitor_internal_dir(), boost::filesystem::system_specific );
  logger::StatePusher                     pusher( clog, "JobControllerReal::readRepository()" );

  repfile <<= repname;

  try {
    clog << logger::setlevel( logger::medium )
	 << "Reading repository from LogMonitor file: " << repfile.file_path() << endl;

    repository.reset( new jccommon::IdContainer(repfile.file_path().c_str()) );
    this->jcr_repository->copy( *repository );
  }
  catch( utilities::FileContainerError &err ) {
    clog << logger::setlevel( logger::null ) << "File container error: " << err.string_error() << endl;

    throw CannotCreate( err.string_error() );
  }

  return;
}

JobControllerReal::JobControllerReal( edg_wll_Context *cont ) : jcr_threshold( 0 ), jcr_repository(), jcr_logger( cont )
{
  const configuration::LMConfiguration   *lmconfig = configuration::Configuration::instance()->lm();
  const configuration::JCConfiguration   *jcconfig = configuration::Configuration::instance()->jc();
  string                                  repname( lmconfig->id_repository_name() );
  auto_ptr<jccommon::IdContainer>         repository;
  boost::filesystem::path                 repfile( lmconfig->monitor_internal_dir(), boost::filesystem::system_specific );
  logger::StatePusher                     pusher( clog, "JobControllerReal::JobControllerReal()" );

  repfile <<= repname;

  jcr_max_children = jcconfig->max_concurrent_condor_commands();
  if (jcr_max_children < 1)
    jcr_max_children = 1;

  try {
    repository.reset( new jccommon::IdContainer(repfile.file_path().c_str()) );
    this->jcr_repository.reset( new jccommon::RamContainer(*repository) );
  }
  catch( utilities::FileContainerError &err ) {
    clog << logger::setlevel( logger::null ) << "File container error: " << err.string_error() << endl;

    throw CannotCreate( err.string_error() );
  }

  this->jcr_threshold = jcconfig->container_refresh_threshold( jcr_s_threshold );
  if( this->jcr_threshold < jcr_s_threshold ) this->jcr_threshold = jcr_s_threshold;

  clog << logger::setlevel( logger::ugly ) << "Controller created..." << endl;
}

JobControllerReal::~JobControllerReal( void )
{}

int JobControllerReal::submit( const classad::ClassAd *pad )
try {
  int                              result, numberId = -1;
  string                           rsl, parameters, info, condorid, seqcode;
  SubmitAd                         ad( pad );
  logger::StatePusher              pusher( clog, "JobControllerReal::submit(...)" );
  ofstream                         ofs;
#ifdef HAVE_STRINGSTREAM
  ostringstream                    oss;
#else
  ostrstream                       oss;
#endif

  boost::match_results<string::const_iterator>    pieces;
  static boost::regex                             expr( "^.*[0-9]+ job\\(s\\) submitted to cluster ([0-9]+)\\.[0-9]*.*$" );

  this->jcr_logger.condor_submit_start_event( ad.log_file() );

  seqcode.assign( this->jcr_logger.sequence_code() );
  ad.set_sequence_code( seqcode );

  if( ad.good() ) {
    clog << logger::setlevel( logger::verylow ) << "Submitting job \"" << ad.job_id() << "\"" << endl;

    requestad::to_submit_stream( oss, ad );

#ifndef HAVE_STRINGSTREAM
    if( oss.str() == NULL ) {
      clog << logger::setlevel( logger::verylow ) << "Submit file is empty, aborting job..." << endl;
      this->jcr_logger.job_abort_cannot_write_submit_file_event( ad.log_file(), ad.submit_file(), "Submit file empty" );

      jccommon::ProxyUnregistrar( ad.job_id() ).unregister();
      jccommon::JobFilePurger( ad.job_id() ).do_purge( true );

      return 0;
    }
#endif

    rsl.assign( oss.str() );

#ifndef HAVE_STRINGSTREAM
    oss.freeze( false );
#else
    if( rsl.length() == 0 ) {
      clog << logger::setlevel( logger::verylow ) << "Submit file is empty, aborting job..." << endl;
      this->jcr_logger.job_abort_cannot_write_submit_file_event( ad.log_file(), ad.submit_file(), "Submit file empty" );

      jccommon::ProxyUnregistrar( ad.job_id() ).unregister();
      jccommon::JobFilePurger( ad.job_id() ).do_purge( true );

      return 0;
    }
#endif

    ofs.open( ad.submit_file().c_str() );

    if( ofs.good() ) {
      ofs << rsl << endl;
      ofs.close();
      clog << logger::setlevel( logger::medium ) << "Submit file created..." << endl;

      ofs.open( ad.classad_file().c_str() );
      if( ofs.good() ) {
	ofs << ad.classad() << endl;
	ofs.close();
	clog << logger::setlevel( logger::medium ) << "Classad file created..." << endl;

	parameters.assign( "-d " ); parameters.append( ad.submit_file() ); parameters.append( " 2>&1" );

        checkChildren();

        int pipe_ends[2];
        pid_t pid = -1;

        if (!pipe(pipe_ends)) pid = fork();
        if (pid < 0) {
          // could not fork a submitting process
          char *err = strerror(errno);
          string errstr(err);

          close(pipe_ends[0]);
          close(pipe_ends[1]);

	  clog << logger::setlevel( logger::severe )
	       << "Error submitting job." << endl
	       << "could not make pipe or fork a submitting process" << endl
	       << logger::setmultiline( true, "CE-> " ) << "Given reason\n" << errstr << endl;
	  clog << logger::setmultiline( false );
	
	  this->jcr_logger.condor_submit_failed_event( rsl, errstr, ad.log_file() );

	  jccommon::ProxyUnregistrar( ad.job_id() ).unregister();
	  jccommon::JobFilePurger( ad.job_id() ).do_purge( true );

        } else if (pid > 0) {
          ++jcr_n_fork;
          close(pipe_ends[1]);
          jcr_read_descriptor[pid] = pipe_ends[0];
          // forked child will do the submit. We return a _fake id_
          numberId = 1;
        } else {
          // child: do the submission
          close(pipe_ends[0]);

          try {
    	    result = CondorG::instance()->set_command( CondorG::submit, parameters )->execute( info );

	    if( result || !boost::regex_match(info, pieces, expr) ) {
	      // The condor command has failed... Do the right thing

	      clog << logger::setlevel( logger::severe )
	           << "Error submitting job." << endl
	           << "condor_submit return code = " << result << endl
	           << logger::setmultiline( true, "CE-> " ) << "Given reason\n" << info << endl;
	      clog << logger::setmultiline( false );
	
	      this->jcr_logger.condor_submit_failed_event( rsl, info, ad.log_file() );

	      jccommon::ProxyUnregistrar( ad.job_id() ).unregister();
	      jccommon::JobFilePurger( ad.job_id() ).do_purge( true );
	    }
	    else {
	      // The condor command worked fine... Do the right thing

	      condorid.assign( pieces[1].first, pieces[1].second );

	      clog << logger::setlevel( logger::verylow )
	           << "Job submitted to Condor cluster: " << condorid << endl;

	      this->jcr_logger.condor_submit_ok_event( rsl, condorid, ad.log_file() );

              char buf[256];
              ssize_t buf_len = sizeof(buf);
              int res;
              res = snprintf(buf, buf_len, "%s %s", ad.job_id().c_str(), condorid.c_str());
              if (res>0 && res<buf_len) {
                (void)signal(SIGPIPE, SIG_IGN);
                (void)write(pipe_ends[1], buf, res);
              }
	    }
          } catch (...) {
            // ignore
          }
          close(pipe_ends[1]);
          exit(0);
        } // end of child
      }
      else {
	clog << logger::setlevel( logger::severe ) << "Cannot open classad file for writing." << endl
	     << "File: \"" << ad.submit_file() << "\"" << endl;

	this->jcr_logger.job_abort_cannot_write_submit_file_event( ad.log_file(), ad.classad_file(), "Cannot open file" );

	jccommon::ProxyUnregistrar( ad.job_id() ).unregister();
	jccommon::JobFilePurger( ad.job_id() ).do_purge( true );

        throw CannotExecute( "Cannot open classad file." );
      }
    }
    else {
      clog << logger::setlevel( logger::severe ) << "Cannot open condor submit file for writing." << endl
	   << "File: \"" << ad.submit_file() << "\"" << endl;

      this->jcr_logger.job_abort_cannot_write_submit_file_event( ad.log_file(), ad.submit_file(), "Cannot open file" );

      jccommon::ProxyUnregistrar( ad.job_id() ).unregister();
      jccommon::JobFilePurger( ad.job_id() ).do_purge( true );

      throw CannotExecute( "Cannot open condor submit file." );
    }
  }
  else { // SubmitAd not good...
    bool      good;
    string    id( requestad::get_edg_jobid(*pad, good) );

    clog << logger::setlevel( logger::severe ) << "Received classad is invalid." << endl
	 << "Reason: \"" << ad.reason() << "\"" << endl;

    this->jcr_logger.job_abort_classad_invalid_event( ad.log_file(), ad.reason() );

    jccommon::ProxyUnregistrar( id ).unregister();
    jccommon::JobFilePurger( id ).do_purge( true );
  }

  return numberId;
}
catch( utilities::FileContainerError &error ) {
  clog << logger::setlevel( logger::null ) << "Error during handling of internal FileList." << endl;

  throw CannotExecute( error.string_error() );
}
catch( SubmitAdException &error ) { throw CannotExecute( error.error() ); }

bool JobControllerReal::cancel( const jobid::JobId &id, const char *givenid, const char *logfile, bool force )
{
  bool                  good = false;
  string                sid( id.toString() ), condorid, info;
  logger::StatePusher   pusher( clog, "JobControllerReal::cancel(...)" );

  clog << logger::setlevel( logger::info )
       << "Asked to remove job: " << id.toString() << ( (givenid) ? ( " with condor id " + std::string(givenid)) : "" ) << endl;

  // check for any submitting child processes, perhaps they will update jcr_repository
  checkChildren();

  if (givenid) {
    condorid.assign(givenid);
  } else {
    condorid.assign( this->jcr_repository->condor_id(sid) );
    if (!condorid.empty()) {
      clog << logger::setlevel( logger::info )
           << "Determined that the corresponding condor id is: " << condorid << endl;
    }
  }

  if( condorid.size() != 0 ) {
    if( (good = cancelJob(condorid, force, info)) ) { // The condor command worked fine
      if( logfile ) {
	auto_ptr<GenericEvent>     event( createGenericEvent(jccommon::cancelled_event) );

	UserLog   logger( "owner", logfile, boost::lexical_cast<int>(condorid), 0, 0 );
	logger.writeEvent( event.get() );
      }

      clog << logger::setlevel( logger::info ) << "Job " << sid << " successfully marked for removal." << endl;
      this->jcr_repository->remove_by_condor_id( condorid );
    }
    else if( logfile ) {
      auto_ptr<GenericEvent>     event( createGenericEvent(jccommon::cannot_cancel_event) );

      UserLog    logger( "owner", logfile, boost::lexical_cast<int>(condorid), 0, 0 );
      logger.writeEvent( event.get() );

      this->jcr_logger.job_cancel_refused_event( info );
    }
  }
  else
    clog << logger::setlevel( logger::info ) << "Job " << sid << " unknown to the system." << endl;

  return good;
}

bool JobControllerReal::cancel( int condorid, bool force )
{
  bool                  good;
  string                sid( boost::lexical_cast<string>(condorid) ), info;
  logger::StatePusher   pusher( clog, "JobControllerReal::cancel(...)" );

  clog << logger::setlevel( logger::info )
       << "Asked to remove job: " << sid << " (by condor ID)." << endl;

  checkChildren();

  if( (good = cancelJob(sid, force, info)) )
    clog << logger::setlevel( logger::info ) << "Job " << sid << " successfully marked for removal." << endl;

  return good;
}

size_t JobControllerReal::queue_size( void )
{
  return 0;
}

}; // namespace controller

} JOBCONTROL_NAMESPACE_END;
