#include <iostream>
#include <fstream>
#include <memory>

#include <classad_distribution.h>

#include <boost/lexical_cast.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/exception.hpp>

#include "edg/workload/common/logger/manipulators.h"
#include "edg/workload/common/configuration/Configuration.h"
#include "edg/workload/common/configuration/LMConfiguration.h"
#include "edg/workload/common/configuration/JCConfiguration.h"
#include "edg/workload/common/requestad/JobAdManipulation.h"
#include "edg/workload/common/requestad/PrivateAdManipulation.h"
#include "edg/workload/common/requestad/ManipulationExceptions.h"
#include "edg/workload/common/jobid/JobId.h"
#include "edg/workload/common/jobid/manipulation.h"
#include "edg/workload/common/utilities/boost_fs_add.h"
#include "edg/workload/jobcontrol/jobcontrol_namespace.h"
#include "edg/workload/jobcontrol/common/files.h"

#include "SubmitAd.h"
#include "SubmitAdExceptions.h"

USING_COMMON_NAMESPACE;
using namespace std;

JOBCONTROL_NAMESPACE_BEGIN {

namespace controller {

void SubmitAd::loadStatus( void )
try {
  const configuration::LMConfiguration   *config = configuration::Configuration::instance()->lm();
  string                      err;
  ifstream                    ifs;
  boost::filesystem::path     status( config->monitor_internal_dir(), boost::filesystem::system_specific );

  status <<= "controller.status";

  if( boost::filesystem::exists(status) ) {
    ifs.open( status.file_path().c_str() );

    if( ifs.good() ) {
      ifs >> this->sa_lastEpoch >> this->sa_jobperlog;

      ifs.close();
    }
    else
      throw CannotOpenStatusFile( status.file_path() );
  }
  else {
    this->sa_lastEpoch = time( NULL );
    this->saveStatus();
  }

  return;
}
catch( boost::filesystem::filesystem_error &err ) {
  throw FileSystemError( err.what() );
}

void SubmitAd::saveStatus( void )
{
  const configuration::LMConfiguration   *config = configuration::Configuration::instance()->lm();
  string                      err;
  ofstream                    ofs;
  boost::filesystem::path     status( config->monitor_internal_dir(), boost::filesystem::system_specific );

  status <<= "controller.status";

  ofs.open( status.file_path().c_str() );
  if( ofs.good() ) {
    ofs << this->sa_lastEpoch << ' ' << this->sa_jobperlog << endl;

    ofs.close();
  }
  else
    throw CannotOpenStatusFile( status.file_path(), 1 );

  return;
}

void SubmitAd::createFromAd( const classad::ClassAd *pad )
{
  const configuration::LMConfiguration    *lmconfig = configuration::Configuration::instance()->lm();
  int                                      maxjobs = lmconfig->jobs_per_condor_log();
  time_t                                   epoch;
  char                                    *dirType;
  jobid::JobId                             edgId;
  auto_ptr<jccommon::Files>                files;
  logger::StatePusher                      pusher( clog, "SubmitAd::createFromAd(...)" );

  if( this->sa_ad.get() == NULL )
    this->sa_ad.reset( pad->Copy() );

  this->loadStatus();

  this->sa_jobid.assign( requestad::get_edg_jobid(*this->sa_ad, this->sa_good) );
  if( !this->sa_good ) this->sa_reason.assign( "Cannot extract \"dg_jobid\" from given classad." );
  else {
    edgId.fromString( this->sa_jobid );
    files.reset( new jccommon::Files(edgId) );

    try {
      dirType = "job directory";
      clog << logger::setlevel( logger::high ) << "Creating job directory path." << endl;
      boost::filesystem::create_parents( files->output_directory() );

      dirType = "submit file";
      if( !boost::filesystem::exists(files->submit_file().branch()) ) {
	clog << logger::setlevel( logger::info ) << "Path for submit file doesn't exists, creating..." << endl
	     << logger::setlevel( logger::debug ) << "Path = " << files->submit_file().branch().file_path() << endl;

	boost::filesystem::create_parents( files->submit_file().branch() );
      }
      this->sa_submitfile.assign( files->submit_file().file_path() );

      dirType = "classad file";
      if( !boost::filesystem::exists(files->classad_file().branch()) ) {
	clog << logger::setlevel( logger::info ) << "Path for classad file doesn't exist, creating..." << endl
	     << logger::setlevel( logger::debug ) << "Path = " << files->classad_file().branch().file_path() << endl;

	boost::filesystem::create_parents( files->classad_file().branch() );
      }
      this->sa_classadfile.assign( files->classad_file().file_path() );
    }
    catch( boost::filesystem::filesystem_error &err ) {
      clog << logger::setlevel( logger::null )
	   << "Failed to create " << dirType << " path \"" << files->output_directory().file_path() << "\"." << endl
	   << "Reason: " << err.what() << endl;

      throw CannotCreateDirectory( files->output_directory().file_path(), dirType, err.what() );
    }

    // Create CondorG log file name
    epoch = this->sa_lastEpoch;
    if( this->sa_jobperlog >= maxjobs ) {
      this->sa_last = true;

      this->sa_lastEpoch = time( NULL );
      this->sa_jobperlog = 0;

      clog << logger::setlevel( logger::medium )
	   << "Maximum number of jobs per log file reached." << endl
	   << "Next job will be submitted to the file: CondorG." << this->sa_lastEpoch << ".log." << endl;
    }
    else this->sa_jobperlog += 1;

    this->saveStatus();

    this->sa_logfile.assign( files->log_file(epoch).file_path() );

    try {
      requestad::set_log( *this->sa_ad, this->sa_logfile );
      requestad::set_condor_submit_file( *this->sa_ad, this->sa_submitfile );
    }
    catch( requestad::CannotGetAttribute &par ) {
      this->sa_reason.assign( "Cannot extract parameter \"" );
      this->sa_reason.append( par.parameter() ); this->sa_reason.append( "\" from given classad." );

      this->sa_good = false;
    }
    catch( requestad::CannotSetAttribute &par ) {
      this->sa_reason.assign( "Cannot set parameter \"" );
      this->sa_reason.append( par.parameter() ); this->sa_reason.append( "\" into given classad." );

      this->sa_good = false;
    }
    catch( requestad::CannotRemoveAttribute &par ) {
      this->sa_reason.assign( "Cannot remove parameter \"" );
      this->sa_reason.append( par.parameter() ); this->sa_reason.append( "\" from classad." );

      this->sa_good = false;
    }
  }

  return;
}

SubmitAd::SubmitAd( const classad::ClassAd *pad ) : sa_good( true ), sa_last( false ), sa_jobperlog( 0 ), sa_lastEpoch( 0 ),
						    sa_ad( pad ? pad->Copy() : NULL ),
						    sa_jobid(), sa_submitfile(), sa_submitad(), sa_reason(), sa_seqcode(),
						    sa_classadfile(), sa_logfile()
{
  if( pad ) this->createFromAd( pad );
}

SubmitAd::~SubmitAd( void ) {}

SubmitAd &SubmitAd::set_sequence_code( const string &code )
{
  string     seqcode, notes;

  this->sa_seqcode.assign( code );

  if( this->sa_good ) {
    if( this->sa_seqcode.size() == 0 )
      seqcode.assign( requestad::get_lb_sequence_code(*this->sa_ad) );
    else
      seqcode.assign( this->sa_seqcode );

    notes.assign( "(" ); notes.append( this->sa_jobid ); notes.append( ") (" );
    notes.append( seqcode ); notes.append( ") (" );
    notes.append( boost::lexical_cast<string>(this->sa_last) ); notes.append( 1, ')' );

    requestad::set_lb_sequence_code( *this->sa_ad, seqcode );
    requestad::set_submit_event_notes( *this->sa_ad, notes );

    seqcode.insert( seqcode.begin(), '\'' ); seqcode.append( 1, '\'' );
    requestad::set_arguments( *this->sa_ad, seqcode ); // Must pass the seqcode as first argument of the JobWrapper...
  }

  return *this;
}

}; // Namespace controller

} JOBCONTROL_NAMESPACE_END;
