#include <cstdio>
#include <ctime>

#include <iostream>
#include <fstream>
#include <list>

#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/exception.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/regex.hpp>

#include <user_log.c++.h>
#include <globus_gram_protocol_constants.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "edg/workload/common/configuration/Configuration.h"
#include "edg/workload/common/configuration/JCConfiguration.h"
#include "edg/workload/common/configuration/LMConfiguration.h"
#include "edg/workload/common/configuration/WMConfiguration.h"
#include "edg/workload/common/configuration/NSConfiguration.h"
#include "edg/workload/common/utilities/boost_fs_add.h"
#include "edg/workload/common/utilities/streamdescriptor.h"
#include "edg/workload/common/logger/manipulators.h"
#include "edg/workload/common/jobid/JobId.h"
#include "edg/workload/common/jobid/manipulation.h"
#include "edg/workload/common/utilities/FileList.h"
#include "edg/workload/jobcontrol/controller/JobController.h"
#include "edg/workload/jobcontrol/common/IdContainer.h"
#include "edg/workload/jobcontrol/common/files.h"
#include "edg/workload/jobcontrol/common/constants.h"
#include "edg/workload/jobcontrol/common/EventLogger.h"
#include "edg/workload/jobcontrol/common/JobFilePurger.h"
#include "edg/workload/jobcontrol/common/ProxyUnregistrar.h"

#include "JobWrapperOutputParser.h"
#include "CondorMonitor.h"
#include "AbortedContainer.h"
#include "exceptions.h"

#define create_path( string ) (boost::filesystem::path((string), boost::filesystem::system_specific))

using namespace std;
USING_COMMON_NAMESPACE;

JOBCONTROL_NAMESPACE_BEGIN {

namespace logmonitor {

namespace {

typedef  JobWrapperOutputParser   JWOP;

boost::filesystem::path createDotFile( const boost::filesystem::path &file, const string &ext )
{
  string                    name( file.leaf() );
  boost::filesystem::path   dotfile( file.branch() );

  name.insert( name.begin(), '.' ); name.append( ext );

  return( dotfile << name );
}

}; // Unnamed namespace closure

string  CondorMonitor::cm_s_recycleDirectory;

MonitorData::MonitorData( AbortedContainer *aborted, 
			  jccommon::IdContainer *container,
			  jccommon::EventLogger *logger ) : md_logger( logger ), md_container( container ), md_aborted( aborted )
{}

void CondorMonitor::readRSL( const string &edgId, string &full, string &globus )
{
  int                       c;
  char                     *where;
  streamsize                nbytes;
  jobid::JobId              jobid( edgId );
  jccommon::Files           files( jobid );
  ifstream                  rsl;
  char                      buffer[BUFSIZ];
  logger::StatePusher       pusher( clog, "CondorMonitor::readRSL(...)" );

  clog << logger::setlevel( logger::high ) << "Reading condor submit file of job " << edgId << endl;

  rsl.open( files.submit_file().file_path().c_str() );

  if( rsl.good() ) {
    full.erase();

    while( !rsl.eof() ) {
      nbytes = rsl.getline( buffer, BUFSIZ ).gcount();
      if( nbytes ) {
	full.append( buffer, nbytes - 1 );
	full.append( 1, '\n' );

	if( (where = strstr(buffer, "GlobusRSL = ")) != NULL )
	  globus.assign( buffer + 12 );
      }
      else if( (c = rsl.get()) != EOF ) full.append( 1, (char) c );
    }

    rsl.close();
  }
  else full.assign( "RSL not available anymore" );

  return;
}

void CondorMonitor::openFile( void )
{
  int                           position;
  FILE                         *fp;
  string                        err;
  boost::filesystem::path       file( this->cm_filename, boost::filesystem::system_specific );
  logger::StatePusher           pusher( clog, "CondorMonitor::openFile()" );

  if( !boost::filesystem::exists(file) ) {
    utilities::create_file( this->cm_filename.c_str() );
    this->cm_sizefile.reset( new SizeFile(this->cm_filename.c_str(), true) );

    clog << logger::setlevel( logger::info ) << "Created log position file." << endl;
  }
  else
    this->cm_sizefile.reset( new SizeFile(this->cm_filename.c_str()) );

  if( !this->cm_sizefile->good() )
    throw CannotOpenFile( this->cm_sizefile->filename(), errno );

  this->cm_logfile->initialize( this->cm_filename.c_str() );
  if( this->cm_logfile->getfd() == -1 ) {
    clog << logger::setlevel( logger::severe ) << "Cannot open Condor log file \"" << this->cm_filename << "\"." << endl;

    throw CannotOpenFile( this->cm_filename );
  }

  clog << logger::setlevel( logger::info ) << "Condor log file parser initialized." << endl;

  if( this->cm_sizefile->size_field().position() != 0 ) {
    clog << logger::setlevel( logger::debug ) << "Old (known) status for this file:" << endl
	 << "Log position = " << this->cm_sizefile->size_field().position() << ", "
	 << this->cm_sizefile->size_field().pending() << " job(s) running." << endl
	 << "Last job " << ( this->cm_sizefile->size_field().last() ? "yet" : "not" ) << " submitted." << endl;

    fp = this->cm_logfile->getfp();

    position = fseek( fp, this->cm_sizefile->size_field().position(), SEEK_SET );

    if( position == -1 ) {
      err.assign( "Cannot seek file \"" ); err.append( this->cm_filename );
      err.append( "\" to old position " ); 
      err.append( boost::lexical_cast<string>(this->cm_sizefile->size_field().position()) );
      err.append( ", reason: " ); err.append( strerror(errno) ); err.append( "." );

      clog << logger::setlevel( logger::fatal ) << "Cannot seek file \"" << this->cm_filename
  	   << "\" to old position " << this->cm_sizefile->size_field().position() << endl
  	   << "Due to: \"" << strerror(errno) << "\"" << endl;

      throw FileSystemError( err );
    }
  }

  return;
}

void CondorMonitor::writeCurrentPosition( void )
{
  long int               currpos;
  FILE                  *fp;
  string                 err;
  logger::StatePusher    pusher( clog, "CondorMonitor::writeCurrentPosition()" );

  if( this->cm_sizefile->good() ) {
    fp = this->cm_logfile->getfp();
    currpos = ftell( fp );

    if( currpos == -1 ) {
      clog << logger::setlevel( logger::null ) << "Error during ftell operation on log file \""
	   << this->cm_filename << "\"." << endl;

      err.assign( "Cannot ftell log file \"" ); err.append( this->cm_filename ); err.append( "\"." );
      throw FileSystemError( err );
    }
    else {
      this->cm_sizefile->update_position( currpos );

      if( !this->cm_sizefile->good() ) {
        err.assign( "Error while writing on position file \"" ); err.append( this->cm_sizefile->filename() ); err.append( "\"." );

        clog << logger::setlevel( logger::fatal ) << err << endl;

	throw FileSystemError( err );
      }
    }
  }
  else {
    err.assign( "Size file object contained a previous error. Giving up." );
    clog << logger::setlevel( logger::fatal ) << err << endl;

    throw CannotExecute( err );
  }

  return;
}

void CondorMonitor::doRecycle( void )
{
  boost::filesystem::path    recycleDir( cm_s_recycleDirectory, boost::filesystem::system_specific );
  boost::filesystem::path    oldfile( this->cm_filename, boost::filesystem::system_specific );
  boost::filesystem::path    dotfile( this->cm_sizefile->filename(), boost::filesystem::system_specific ), newfile;
  logger::StatePusher        pusher( clog, "CondorMonitor::doRecycle()" );

  newfile = recycleDir << oldfile.leaf();

  boost::filesystem::copy_file( oldfile, newfile );
  clog << logger::setlevel( logger::medium ) << "Old log file (" << this->cm_filename
       << ") copied to recycle file \"" << newfile.file_path() << "\"" << endl;

  boost::filesystem::remove( oldfile );
  clog << logger::setlevel( logger::high ) << "Successfully removed old file." << endl;

  boost::filesystem::remove( dotfile );
  clog << "Successfully removed size file." << endl;

  return;
}

void CondorMonitor::resubmitJob( int laststatus, const std::string &edgId, const std::string &sequence )
{
  const configuration::WMConfiguration       *config = configuration::Configuration::instance()->wm();
  utilities::FileList<classad::ClassAd>       flist( config->input() );
  utilities::FileListDescriptorMutex          flmutex( flist );
  classad::ClassAd                            ad, empty;
  logger::StatePusher                         pusher( clog, "CondorMonitor::resubmitJob(...)" );
  int                                         retcode;
  string                                      errors;
  JobWrapperOutputParser                      parser( edgId );
    
  clog << logger::setlevel( logger::ugly ) << "Last known status = " << laststatus << endl;

  switch( laststatus ) {
  case GLOBUS_GRAM_PROTOCOL_ERROR_USER_PROXY_EXPIRED:
    clog << logger::setlevel( logger::low ) << "Job has an expiring proxy." << endl
	 << logger::setlevel( logger::ugly ) << "Must not resubmit, but abort." << endl
	 << "EDG id = " << edgId << endl;

    this->cm_logger->abort_on_error_event( string("Job proxy is expired.") );

    jccommon::ProxyUnregistrar( edgId ).unregister();

    break;
    // see LCG2 bug 3987
  case GLOBUS_GRAM_PROTOCOL_ERROR_STAGE_OUT_FAILED:
    if( parser.parse_file(retcode, errors) == JWOP::good ) {
      // Job terminated successfully
      jccommon::IdContainer           &container = *this->cm_container;
      jccommon::IdContainer::iterator  position  = container.position_by_edg_id( edgId );
      
      clog << logger::setlevel( logger::verylow ) << "Real return code: " << retcode << endl;
      
	this->cm_logger->reset_user_proxy( position->proxy_file() ).reset_context( position->edg_id(), position->sequence_code() );
	this->cm_logger->terminated_event( retcode, errors );
	this->updateSequenceCode( position, ULOG_JOB_TERMINATED );
	jccommon::ProxyUnregistrar( position->edg_id() ).unregister();
	
	break;
    } else // resubmit it --> default label
      clog << logger::setlevel( logger::verylow ) << "The job is not terminated successfully. Reason: " << errors << endl;
  default:
    clog << logger::setlevel( logger::low ) << "Resubmitting job to WM." << endl
	 << logger::setlevel( logger::ugly )
	 << "EDG id = " << edgId << endl;

    ad.InsertAttr( "version", string("1.0.0") );
    ad.InsertAttr( "command", string("jobresubmit") );

    empty.InsertAttr( "id", edgId );
    empty.InsertAttr( "lb_sequence_code", sequence );

    ad.Insert( "arguments", empty.Copy() );

    this->cm_logger->job_resubmitting_event();
    this->cm_logger->job_wm_enqueued_start_event( config->input(), ad );
    try {
      utilities::FileListLock   lock( flmutex );
      flist.push_back( ad );

      this->cm_logger->job_wm_enqueued_ok_event( config->input(), ad );
    }
    catch( utilities::FileContainerError &error ) {
      this->cm_logger->job_wm_enqueued_failed_event( config->input(), ad, error.string_error() );

      throw CannotExecute( error.string_error() );
    }

    break;
  }

  return;
}

void CondorMonitor::updateSequenceCode( jccommon::IdContainer::iterator &position, int status, int last )
{
  string         seqcode( this->cm_logger->sequence_code() );

  this->cm_container->update_pointer( position, seqcode, status, last );

  return;
}

void CondorMonitor::processEvent( ULogEvent *event, bool removeTimer )
try {
  jccommon::IdContainer                 &container = *this->cm_container;
  string                                 edgid, last;
  string                                 condorid( boost::lexical_cast<string>(event->cluster) );
  jccommon::IdContainer::iterator        position;
  logger::StatePusher                    pusher( clog, "CondorMonitor::processEvent(...)" );
  const string     notsub( "Job seems to be not submitted by EDG." ), edgideq( "EDG id = " );
  const string     joberror( "Job got an error while in the CondorG queue." );
		
  static boost::regex  expr( "^\\((.*)\\) \\((.*)\\) \\(([01])\\)$" );
  static boost::regex  genexpr( "^JC: ([0-9]+) - (.*)$" );
  static boost::regex  lmexpr( "^LM: ([0-9]+) - (.*)$" );

  if( event->proc != 0 ) {
    condorid.append( 1, '.' );
    condorid.append( boost::lexical_cast<string>(event->proc) );
  }

  switch( event->eventNumber ) {
  default: // We have got an unhandled event
    clog << logger::setlevel( logger::null ) << "Got unhandled event " << event->eventNumber << '.' << endl
	 << "Meaning: \"" << ULogEventNumberNames[event->eventNumber] << "\"." << endl;

    this->cm_logger->unhandled_event( ULogEventNumberNames[event->eventNumber] );

    break;
  case ULOG_SUBMIT: {
    SubmitEvent     *subev = dynamic_cast<SubmitEvent *>( event );
    string           notes( subev->submitEventLogNotes ? subev->submitEventLogNotes : "" ), seqcode, rsl, globrsl;

    boost::match_results<string::const_iterator>    pieces;

    clog << logger::setlevel( logger::ugly ) << "Got job submit event." << endl
	 << logger::setlevel( logger::verylow ) << "Submitted job "
	 << condorid << " coming from host: " << subev->submitHost << endl;

    if( boost::regex_match(notes, pieces, expr) ) { // The event notes are in the right format.
      edgid.assign( pieces[1].first, pieces[1].second );
      seqcode.assign( pieces[2].first, pieces[2].second );
      last.assign( pieces[3].first, pieces[3].second );

      this->cm_sizefile->set_last( boost::lexical_cast<bool>(last) ).increment_pending();

      clog << logger::setlevel( logger::ugly ) << edgideq << edgid << endl
	   << "Sequence code = " << seqcode << endl;

      if( container.insert(edgid, condorid, seqcode, event->eventNumber) ) {
	clog << logger::setlevel( logger::null ) << "Error while inserting EDGid<->Condorid correspondance." << endl
	     << "Ignoring cluster " << condorid << endl;

//  	throw CannotExecute( "Error while inserting EDGid<->Condorid correspondance." );
      }
      else {
	position = container.last_inserted();

	this->readRSL( edgid, rsl, globrsl );

	this->cm_logger->reset_user_proxy( position->proxy_file() ).reset_context( edgid, seqcode );
	this->cm_logger->condor_submit_event( condorid, globrsl );
	this->updateSequenceCode( position, event->eventNumber );
      }
    }
    else
      clog << logger::setlevel( logger::null ) << "Cluster " << condorid << " does not seem an EDG job." << endl
	   << logger::setlevel( logger::low ) << "Event notes = \"" << notes << "\"." << endl;

    break;
  }
  case ULOG_GLOBUS_SUBMIT: {
    GlobusSubmitEvent     *gseevent = dynamic_cast<GlobusSubmitEvent *>( event );
    string                 ce( gseevent->rmContact ), rsl, globrsl;

    clog << logger::setlevel( logger::verylow ) << "Got job submit to globus event." << endl
	 << "For cluster " << condorid << endl
	 << "Contacts " << ce << ", " << gseevent->jmContact << '.' << endl;

    position = container.position_by_condor_id( condorid );

    if( position == container->end() )
      clog << logger::setlevel( logger::null ) << notsub << endl;
    else {
      clog << logger::setlevel( logger::verylow ) << edgideq << position->edg_id() << endl;

      this->readRSL( position->edg_id(), rsl, globrsl );

      this->cm_logger->reset_user_proxy( position->proxy_file() ).reset_context( position->edg_id(), position->sequence_code() );
      this->cm_logger->globus_submit_event( ce, globrsl, this->cm_filename );
      this->updateSequenceCode( position, event->eventNumber );
    }

    break;
  }
  case ULOG_EXECUTE: {
    ExecuteEvent      *eevent = dynamic_cast<ExecuteEvent *>( event );

    clog << logger::setlevel( logger::verylow ) << "Got job executing event." << endl
	 << "For cluster " << condorid << " at host " << eevent->executeHost << endl;

    position = container.position_by_condor_id( condorid );

    if( position == container->end() )
      clog << logger::setlevel( logger::null ) << notsub << endl;
    else {
      clog << logger::setlevel( logger::verylow ) << edgideq << position->edg_id() << endl;

      this->cm_logger->reset_user_proxy( position->proxy_file() ).reset_context( position->edg_id(), position->sequence_code() );
      this->cm_logger->execute_event( eevent->executeHost );
      this->updateSequenceCode( position, event->eventNumber );

      /*
	We are lacking sending of notification mail...
	Where should I get the information about the user email address ??
      */
    }

    break;
  }
  case ULOG_JOB_TERMINATED: {
    int                  retcode, stat;
    JobTerminatedEvent  *termevent = dynamic_cast<JobTerminatedEvent *>( event );
    string               errors;

    clog << logger::setlevel( logger::verylow ) << "Got job terminated event." << endl
	 << "For cluster " << condorid << "; fake return value " << termevent->returnValue << endl;

    position = container.position_by_condor_id( condorid );

    if( position == container->end() )
      clog << logger::setlevel( logger::null ) << notsub << endl;
    else {
      JobWrapperOutputParser   parser( position->edg_id() );

      if( removeTimer )
	  this->cm_timer->remove_all_timeouts( boost::lexical_cast<int>(condorid) ); // Remove any installed timeout for this job...

      clog << logger::setlevel( logger::verylow ) << edgideq << position->edg_id() << endl;

      this->cm_sizefile->decrement_pending();

      if( this->cm_aborted->search(condorid) ) { // Previously aborting job...
	this->cm_aborted->remove( condorid );

	this->cm_logger->reset_user_proxy( position->proxy_file() ).reset_context( position->edg_id(), position->sequence_code() );
	this->cm_logger->failed_on_error_event( joberror );

	jccommon::JobFilePurger( position->edg_id() ).do_purge();
	this->resubmitJob( position->last_status(), position->edg_id(), position->sequence_code() );
	this->updateSequenceCode( position, event->eventNumber );
      }
//        else if( (stat = this->parseJobWrapperOutput(position->edg_id(), retcode, errors)) == JWErrors::good ) {
      else if( (stat = parser.parse_file(retcode, errors)) == JWOP::good ) {
	// Job terminated successfully
	clog << logger::setlevel( logger::verylow ) << "Real return code: " << retcode << endl;

	this->cm_logger->reset_user_proxy( position->proxy_file() ).reset_context( position->edg_id(), position->sequence_code() );
	this->cm_logger->terminated_event( retcode, errors );
	this->updateSequenceCode( position, event->eventNumber );

	jccommon::ProxyUnregistrar( position->edg_id() ).unregister();
	jccommon::JobFilePurger( position->edg_id() ).do_purge();
      }
      else {
	// Job got error in the jobwrapper
	clog << logger::setlevel( logger::verylow ) << "Last job terminated (" << condorid 
	     << ") aborted." << endl
	     << "Reason: " << errors << endl;

	this->cm_logger->reset_user_proxy( position->proxy_file() ).reset_context( position->edg_id(), position->sequence_code() );
	this->cm_logger->failed_on_error_event( errors );

	jccommon::JobFilePurger    purger( position->edg_id() );
	switch( stat ) {
	case JWOP::resubmit:
	  purger.do_purge();
	  this->resubmitJob( jccommon::undefined_status, position->edg_id(), position->sequence_code() );

	  break;
	case JWOP::abort:
	  this->cm_logger->abort_on_error_event( string("Failure while executing job wrapper") );
	
	  jccommon::ProxyUnregistrar( position->edg_id() ).unregister();
	  purger.do_purge( true );

	  break;
	default:
	  break;
	}

	this->updateSequenceCode( position, event->eventNumber );
      }

      if( container.remove(position) ) {
	clog << logger::setlevel( logger::null )
	     << "Error removing CondorId<->EDGid correspondance." << endl;

	throw CannotExecute( "Error removing CondorId<->EDGid correspondance." );
      }
    }

    break;
  }
  case ULOG_JOB_ABORTED: {
    clog << logger::setlevel( logger::verylow ) << "Got job aborted event." << endl
	 << "For cluster " << condorid << endl;

    position = container.position_by_condor_id( condorid );

    if( position == container->end() )
      clog << logger::setlevel( logger::null ) << notsub << endl;
    else {
      clog << logger::setlevel( logger::verylow ) << edgideq << position->edg_id() << endl;

      if( removeTimer )
	  this->cm_timer->remove_all_timeouts( boost::lexical_cast<int>(condorid) ); // Remove any installed timeout for this job...

      this->cm_sizefile->decrement_pending();

      if( this->cm_aborted->search(condorid) ) { // Job got an error...
	this->cm_aborted->remove( condorid );

	this->cm_logger->reset_user_proxy( position->proxy_file() ).reset_context( position->edg_id(), position->sequence_code() );
	this->cm_logger->aborted_by_system_event( joberror );

	jccommon::JobFilePurger( position->edg_id() ).do_purge();
	this->resubmitJob( position->last_status(), position->edg_id(), position->sequence_code() );
	this->updateSequenceCode( position, event->eventNumber );
      }
      else { // Job successfully aborted
	this->cm_logger->reset_user_proxy( position->proxy_file() ).reset_context( position->edg_id(), position->sequence_code() );
	this->cm_logger->aborted_by_user_event();

	jccommon::ProxyUnregistrar( position->edg_id() ).unregister();
	jccommon::JobFilePurger( position->edg_id() ).do_purge( true );

	this->updateSequenceCode( position, event->eventNumber );
      }

      if( container.remove(position) ) {
	clog << logger::setlevel( logger::null )
	     << "Error removing CondorId<->EDGid correspondance." << endl;

	throw CannotExecute( "Error removing CondorId<->EDGid correspondance." );
      }
    }

    break;
  }
  case ULOG_GLOBUS_SUBMIT_FAILED: {
    GlobusSubmitFailedEvent    *gsfevent = dynamic_cast<GlobusSubmitFailedEvent *>( event );
    string                      rsl, globrsl;
    controller::JobController   controller( *this->cm_logger );

    clog << logger::setlevel( logger::verylow ) << "Got globus submit failed event." << endl
	 << "For cluster " << condorid << ", reason: " << gsfevent->reason << endl;

    position = container.position_by_condor_id( condorid );

    if( position == container->end() )
      clog << logger::setlevel( logger::null ) << notsub << endl;
    else {
      clog << logger::setlevel( logger::verylow ) << edgideq << position->edg_id() << endl;

      if( this->cm_aborted->insert(condorid) ) {
	clog << logger::setlevel( logger::null ) << "Failed insertion of aborting job in aborted list." << endl;

	throw CannotExecute( "Failed insertion of aborting job." );
      }

      this->readRSL( position->edg_id(), rsl, globrsl );

      this->cm_logger->reset_user_proxy( position->proxy_file() ).reset_context( position->edg_id(), position->sequence_code() );
      this->cm_logger->globus_submit_failed_event( rsl, gsfevent->reason, this->cm_filename );

      clog << logger::setlevel( logger::verylow ) << "Forwarding remove request to JC." << endl;

      controller.cancel( position->edg_id(), condorid.c_str(), this->cm_filename.c_str() );
      this->updateSequenceCode( position, event->eventNumber );
    }

    break;
  }
/*
  case ULOG_GLOBUS_RESOURCE_DOWN: {
    GlobusResourceDownEvent    *grdevent = dynamic_cast<GlobusResourceDownEvent *>( event );
    string                      ce( grdevent->rmContact );
    controller::JobController   controller( *this->cm_logger );

    clog << logger::setlevel( logger::verylow ) << "Got a Globus resource down event." << endl
	 << "For cluster: " << condorid << endl
	 << "On ce: " << ce << endl;

    position = container.position_by_condor_id( condorid );

    if( position == container->end() )
      clog << logger::setlevel( logger::null ) << notsub << endl;
    else {
      clog << logger::setlevel( logger::verylow ) << edgideq << position->edg_id() << endl;

      if( this->cm_aborted->insert(condorid) ) {
	clog << logger::setlevel( logger::null ) << "Failed insertion of aborting job in aborted list." << endl;

	throw CannotExecute( "Failed insertion of aborting job." );
      }

      this->cm_logger->reset_user_proxy( position->proxy_file() ).reset_context( position->edg_id(), position->sequence_code() );
      this->cm_logger->globus_resource_down_event();

      clog << logger::setlevel( logger::verylow ) << "Forwarding remove request to JC." << endl;

      controller.cancel( position->edg_id(), condorid.c_str(), this->cm_filename.c_str() );
      this->updateSequenceCode( position, event->eventNumber );
    }

    break;
  }
*/
  case ULOG_JOB_HELD: {
    JobHeldEvent               *jobheld = dynamic_cast<JobHeldEvent *>( event );
    string                      reason( jobheld->getReason() );
    controller::JobController   controller( *this->cm_logger );

    clog << logger::setlevel( logger::verylow ) << "Got a job held event." << endl
	 << "For cluster: " << condorid << endl;

    position = container.position_by_condor_id( condorid );

    if( position == container->end() )
      clog << logger::setlevel( logger::null ) << notsub << endl;
    else {
      clog << logger::setlevel( logger::verylow ) 
	   << "Reason = \"" << reason << "\"." << endl
#if (CONDORG_VERSION >= 653)
	   << "Code = " << jobheld->getReasonCode() << ", SubCode = " << jobheld->getReasonSubCode() << endl
#endif
	   << edgideq << position->edg_id() << endl;

      if( this->cm_aborted->search(condorid) ) // Job got a previous error, ignoring event...
	clog << "The job got a previous error, ignoring the held event..." << endl;
      else {
	if( this->cm_aborted->insert(condorid) ) {
	  clog << logger::setlevel( logger::null ) << "Failed insertion of aborting job in aborted list." << endl;

	  throw CannotExecute( "Failed insertion of aborting job." );
	}

	this->cm_logger->reset_user_proxy( position->proxy_file() ).reset_context( position->edg_id(), position->sequence_code() );
	this->cm_logger->job_held_event( reason );

	clog << logger::setlevel( logger::verylow ) << "Forwarding remove request to JC." << endl;

	controller.cancel( position->edg_id(), condorid.c_str(), this->cm_filename.c_str() );
#if (CONDORG_VERSION >= 653)
	this->updateSequenceCode( position, event->eventNumber, jobheld->getReasonSubCode() );
#else
	this->updateSequenceCode( position, event->eventNumber );
#endif
      }
    }

    break;
  }
  case ULOG_GENERIC: {
    GenericEvent   *generic = dynamic_cast<GenericEvent *>( event );
    string          info( generic->info ), code, message;

    boost::match_results<string::const_iterator>    pieces;

    clog << logger::setlevel( logger::verylow ) << "Got a generic event." << endl
	 << "For cluster " << condorid << "." << endl
	 << "Maybe our brother want say something to us ??" << endl;

    if( boost::regex_match(info, pieces, genexpr) ) {
      code.assign( pieces[1].first, pieces[1].second );
      message.assign( pieces[2].first, pieces[2].second );

      clog << logger::setlevel( logger::info ) << "It's really a message from my beloved JobController!" << endl
	   << "Message says: \"" << message << "\"." << endl;

      this->processGenericEvent( condorid, boost::lexical_cast<int>(code), message );
    }
    else if( boost::regex_match(info, pieces, lmexpr) ) {
      code.assign( pieces[1].first, pieces[1].second );
      message.assign( pieces[2].first, pieces[2].second );

      clog << logger::setlevel( logger::info ) << "No !!! It's something I was saying yo myself !!!!" << endl
	   << "Message says: \"" << message << "\"." << endl;

      this->processGenericEvent( condorid, boost::lexical_cast<int>(code), message );
    }

    break;
  }
  }

  return;
}
catch( jccommon::LoggerException &err ) {
  throw CannotExecute( err.what() );
}

ULogEvent *CondorMonitor::instantiateCondorEvent( int evn, int condorid, time_t epoch )
{
  ULogEvent                   *event = instantiateEvent( static_cast<ULogEventNumber>(evn) );

  event->cluster = condorid;
  event->proc = event->subproc = 0;

  localtime_r( &epoch, &event->eventTime );

  return event;
}

ULogEvent *CondorMonitor::instantiateGenericEvent( jccommon::generic_event_t evn, int condorid, time_t epoch )
{
  ULogEvent    *event;
  GenericEvent *gev;
  string        msg;

  event = this->instantiateCondorEvent( static_cast<int>(ULOG_GENERIC), condorid, epoch );

  switch( evn ) {
  case jccommon::force_remove:
    msg.assign( "LM: " );
    msg.append( boost::lexical_cast<string>(evn) );
    msg.append( " - Must force job removal" );

    break;
  default:
    msg.assign( "LM: 0 - Null event." );

    break;
  }

  gev = dynamic_cast<GenericEvent *>( event );
  strncpy( gev->info, msg.c_str(), 128 );

  return event;
}

void CondorMonitor::processGenericEvent( const string &condorid, int cn, const string &message )
{
  const configuration::LMConfiguration   *conf = configuration::Configuration::instance()->lm();

  jccommon::IdContainer           &container = *this->cm_container;
  jccommon::generic_event_t        code = (jccommon::generic_event_t) cn;
  ULogEvent                       *event;
  time_t                           epoch;
  char                             wbuf[30];
  string                           when;
  jccommon::IdContainer::iterator  position( container.position_by_condor_id(condorid) );
  logger::StatePusher              pusher( clog, "CondorMonitor::processGenericEvent(...)" );

  clog << logger::setlevel( logger::info );

  if( position != container->end() ) {
    switch( code ) {
    case jccommon::null_event:
    default:
      clog << "This event may be discarded." << endl;

      break;
    case jccommon::cancelled_event:
      clog << logger::setlevel( logger::info ) 
	   << "Attaching force remove timeout to cluster " << condorid << endl;

      epoch = time( NULL ) + conf->aborted_jobs_timeout();
      event = this->instantiateGenericEvent( jccommon::force_remove, boost::lexical_cast<int>(condorid), epoch );

      asctime_r( &event->eventTime, wbuf );
      when.assign( wbuf, 24 );

      clog << logger::setlevel( logger::info )
	   << "Timeout force removal will happen in " << conf->aborted_jobs_timeout() << " seconds." << endl
	   << logger::setlevel( logger::debug ) << "At: " << when << endl;

      this->cm_timer->start_timer( epoch, event );

      break;
    case jccommon::force_remove: {
      controller::JobController        controller( *this->cm_logger );

      clog << "Sending force removal request to JobController for cluster " << condorid << endl
	   << logger::setlevel( logger::debug ) << "This is try n. " << position->retry_count() << endl;

      this->cm_container->increment_pointer_retry_count( position );
      controller.cancel( boost::lexical_cast<int>(condorid) );

      break;
    }
    case jccommon::cannot_cancel_event: {
      int          retry = position->retry_count(), maxretries = conf->force_cancellation_retries();
      string       edgid( position->edg_id() );

      if( retry > maxretries ) {
	clog << logger::setlevel( logger::severe ) 
	     << "Forced cancellation retries exceeded maximum (" << retry << "/" << maxretries << ")" << endl
	     << "Job will be removed from the queue and aborted." << endl;

	this->cm_logger->reset_user_proxy( position->proxy_file() ).reset_context( position->edg_id(), position->sequence_code() );
	this->cm_logger->abort_on_error_event( string("Removal retries exceeded.") );

	jccommon::ProxyUnregistrar( edgid ).unregister();
	jccommon::JobFilePurger( edgid ).do_purge();

	this->cm_container->remove( position );
	this->cm_aborted->remove( condorid );
	this->cm_timer->remove_all_timeouts( boost::lexical_cast<int>(condorid) );
      }
      else {
	clog << "Attaching force remove timeout to cluster " << condorid << endl
	     << logger::setlevel( logger::debug ) << "This is try n. " << (retry + 1) << endl;

	epoch = time( NULL ) + conf->aborted_jobs_timeout();

	event = this->instantiateGenericEvent( jccommon::force_remove, boost::lexical_cast<int>(condorid), epoch );

	asctime_r( &event->eventTime, wbuf );
	when.assign( wbuf, 24 );

	clog << "Timeout force removal will happen in " << conf->aborted_jobs_timeout() << " seconds." << endl
	     << logger::setlevel( logger::debug ) << "At: " << when << endl;

	container.increment_pointer_retry_count( position );
	this->cm_timer->start_timer( epoch, event );
      }

      break;
    }
    }
  }
  else
    clog << logger::setlevel( logger::warning )
	 << "Job has just been removed from the queue." << endl
	 << "Action not performed..." << endl;

  return;
}

string CondorMonitor::timerFileName( const string &logpathn )
{
  const configuration::LMConfiguration *conf = configuration::Configuration::instance()->lm();

  boost::filesystem::path     logpath( logpathn, boost::filesystem::system_specific );
  boost::filesystem::path     newpath( conf->monitor_internal_dir(), boost::filesystem::system_specific );
  string                      logname( logpath.leaf() );

  logname.append( ".timer" );
  newpath <<= logname;

  return newpath.file_path();
}

CondorMonitor::CondorMonitor( const char *filename, MonitorData &data ) : 
  cm_logger( data.md_logger ), cm_container( data.md_container ), cm_aborted( data.md_aborted ),
  cm_filename( filename ), cm_logfile( new ReadUserLog() ), cm_timer( new Timer(CondorMonitor::timerFileName(filename)) )
{ this->openFile(); }

CondorMonitor::CondorMonitor( const std::string &filename, MonitorData &data ) : 
  cm_logger( data.md_logger ), cm_container( data.md_container ), cm_aborted( data.md_aborted ),
  cm_filename( filename ), cm_logfile( new ReadUserLog() ), cm_timer( new Timer(CondorMonitor::timerFileName(filename)) )
{ this->openFile(); }

CondorMonitor::~CondorMonitor( void )
{
  logger::StatePusher     pusher( clog, "CondorMonitor::~CondorMonitor()" );

  if( this->cm_sizefile.unique() && this->cm_sizefile->completed() ) {
    try {
      boost::filesystem::path   timerfile( this->cm_timer->filename(), boost::filesystem::system_specific );
      this->cm_timer.reset(); // Remove the timer and the filelist it was containing before deleting the file...

      clog << logger::setlevel( logger::medium )
	   << "Removing timer file " << timerfile.file_path() << " from staging directory." << endl;
      boost::filesystem::remove( timerfile );
      clog << logger::setlevel( logger::high ) << "Successfully removed." << endl;
    }
    catch( boost::filesystem::filesystem_error &err ) {
      clog << logger::setlevel( logger::null )
	   << "Removing of the timer file failed." << endl
	   << "Filesystem error catched: \"" << err.what() << "\"." << endl;
    }

    try { this->doRecycle(); }
    catch( boost::filesystem::filesystem_error &err ) {
      clog << logger::setlevel( logger::null )
	   << "Condor log recycling operation failed." << endl
	   << "Filesystem error catched: \"" << err.what() << "\"." << endl;
    }
    catch( ... ) {
      clog << logger::setlevel( logger::null )
	   << "Condor log recycling operation failed." << endl
	   << "Unknown error happened." << endl;
    }
  }
}

CondorMonitor::status_t CondorMonitor::process_next_event( void )
{
  status_t                              stat = event_read;
  streampos                             size;
  ULogEventOutcome                      outcome;
  ULogEvent                            *event = NULL;
  boost::scoped_ptr<ULogEvent>		event_scoped;
  Timer::EventIterator                  timerIt;
  string                                err;
  list<Timer::EventIterator>            removedTimers;
  list<Timer::EventIterator>::iterator  remIt;
  boost::filesystem::path               file( this->cm_filename, boost::filesystem::system_specific );
  logger::StatePusher                   pusher( clog, "CondorMonitor::process_next_event()" );

  try {
    size = boost::filesystem::file_size( file );
  }
  catch( boost::filesystem::filesystem_error &err ) {
    throw FileSystemError( err.what() );
  }

  if( size > this->cm_sizefile->size_field().position() ) {
    outcome = this->cm_logfile->readEvent( event );
    event_scoped.reset(event);

    switch( outcome ) {
    case ULOG_NO_EVENT:
      clog << logger::setlevel( logger::verylow ) << "Reached the end." << endl;

      this->writeCurrentPosition();

      stat = no_events;
      break;
    case ULOG_RD_ERROR:
    case ULOG_UNK_ERROR:
      clog << logger::setlevel( logger::null ) << "Error while reading log file \""
	   << this->cm_filename << "\"." << endl;

      stat = event_error;
      break;
    case ULOG_OK:
      this->processEvent( event );
      this->writeCurrentPosition();

      stat = event_read;
      break;
    default:
      clog << logger::setlevel( logger::null ) << "Reached unreachable code." << endl;

      stat = event_error;
      break;
    }
  }
  else stat = no_events;

  if( stat == no_events ) {
    for( timerIt = this->cm_timer->begin(); (timerIt != this->cm_timer->end()) && timerIt->second->expired(); ++timerIt ) {
      clog << logger::setlevel( logger::verylow ) << "Expired timeout for cluster: " 
	   << timerIt->second->to_event()->cluster << endl;

      this->processEvent( timerIt->second->to_event(), false );

      removedTimers.push_back( timerIt );
    }

    if( !removedTimers.empty() ) {
      for( remIt = removedTimers.begin(); remIt != removedTimers.end(); ++remIt )
	this->cm_timer->remove_timeout( *remIt );

      removedTimers.clear();
    }
  }

  return stat;
}

}; // Namespace logmonitor

} JOBCONTROL_NAMESPACE_END;
