#include <cstring>
#include <cerrno>

#include <fstream>

#include "edg/workload/common/configuration/Configuration.h"
#include "edg/workload/common/configuration/LMConfiguration.h"
#include "edg/workload/common/jobid/JobId.h"
#include "edg/workload/common/jobid/manipulation.h"
#include "edg/workload/common/logger/manipulators.h"
#include "edg/workload/jobcontrol/jobcontrol_namespace.h"
#include "edg/workload/jobcontrol/common/files.h"

#include "JobWrapperOutputParser.h"

using namespace std;
USING_COMMON_NAMESPACE;

JOBCONTROL_NAMESPACE_BEGIN {

namespace logmonitor {

namespace {

typedef  JobWrapperOutputParser   JWOP;

struct JWErrors {
  const char         *jwe_error;
  JWOP::status_type   jwe_status;
};

}; // Anonymous namespace

JobWrapperOutputParser::JobWrapperOutputParser( const string &edgid ) : jwop_edgid( edgid ) {}

JobWrapperOutputParser::~JobWrapperOutputParser( void ) {}

bool JobWrapperOutputParser::parseStream( istream &is, string &errors, int &retcode, status_type &stat )
{
  bool                found = false;
  char                buffer[BUFSIZ];
  struct JWErrors     jwErrors[] = { 
                                     { "There were some warnings",       good     },
                                     { "Working directory not writable", resubmit },
				     { "Requested middleware version not found",      resubmit },
				     { "GLOBUS_LOCATION undefined",      resubmit },
				     { "/etc/globus-user-env.sh",        resubmit },
				     { "not found or unreadable",        abort },
				     { "Cannot download",                resubmit },
				     { "Cannot upload",                  resubmit },
				     { "Cannot confirm that the logging",resubmit },
				     { NULL, unknown },
  };
  struct JWErrors    *errIt;
  string good_message;

  if( is.good() ) {
    do {
      is.getline( buffer, BUFSIZ );

      if( !is.eof() ) {
	if( is.good() ) {
	  for( errIt = jwErrors; errIt->jwe_error != NULL; ++errIt )
	    if( strstr(buffer, errIt->jwe_error) != NULL ) {
              if (errIt->jwe_status == good) {
                good_message.assign( buffer );
                if (found && stat == good) {
                  errors = good_message;
                }
              } else {
	        errors.assign( buffer );
	        stat = errIt->jwe_status;
	        found = true;
              }
	      break;
	    }

	  if( strstr(buffer, "job exit status = ") == buffer ) {
	    if( sscanf(buffer, "job exit status = %d", &retcode) == 1 ) {
	      if (!good_message.empty()) {
                errors = good_message;
              } else {
                //errors.assign( buffer );
                errors.clear();
              }
	      found = true;
	      stat = good;
	    }
	    else retcode = -1;
	  }
	}
	else {
	  errors.assign( "IO error while reading file: " );
	  errors.append( strerror(errno) );
	  retcode = -1;
	  stat = resubmit;

	  break;
	}
      }
    } while( !is.eof() );
  }
  else {
    errors.assign( "File not available." );
    retcode = -1;
    stat = resubmit;
  }

  return found;
}

JWOP::status_type JobWrapperOutputParser::parse_file( int &retcode, string &errors )
{
  const configuration::LMConfiguration    *lmconfig = configuration::Configuration::instance()->lm();

  bool                    found = false;
  status_type             stat = good;
  jobid::JobId            id( this->jwop_edgid );
  jccommon::Files         files( id );
  ifstream                ifs;
  logger::StatePusher     pusher( clog, "JobWrapperOutputParser::parse_file(...)" );

  errors.erase(); retcode = -1;

  clog << logger::setlevel( logger::high ) << "Going to parse standard output file." << endl;

  ifs.open( files.standard_output().file_path().c_str() );
  found = this->parseStream( ifs, errors, retcode, stat );
  ifs.close();
  if( !found ) {
    errors.assign( "Standard output does not contain useful data." );
    clog << logger::setlevel( logger::null ) << errors << endl;

    if( lmconfig->use_maradona_file() ) {
      clog << logger::setlevel( logger::high )
	   << "Standard output was not useful, passing ball to Maradona..." << endl;

      ifs.clear();
      ifs.open( files.maradona_file().file_path().c_str() );
      found = this->parseStream( ifs, errors, retcode, stat );
      ifs.close();
      if( found )
	clog << logger::setlevel( logger::null )
	     << "Got info from Maradona..." << endl
	     << logger::setlevel( logger::ugly )
	     << "Maradona makes another goal !!!" << endl
	     << "The legend goes on..." << endl
	     << logger::setlevel( logger::veryugly )
	     << "Stuttgard - Naples: 3 - 3" << endl
	     << "Naples win the UEFA cup !!!" << endl;
      else {
	errors.assign( "Cannot read JobWrapper output, both from Condor and from Maradona." );

	clog << logger::setlevel( logger::null ) << errors
	     << logger::setlevel( logger::ugly )
	     << "Maradona fails the shot !!!" << endl
	     << logger::setlevel( logger::veryugly )
	     << "100000 fans in the stadium boo him !!!" << endl;

	retcode = -1;
	stat = resubmit;
      }
    }
    else {
      clog << logger::setlevel( logger::null )
	   << "Maradona disabled, cannot check for alternate output." << endl;

      retcode = -1;
      stat = resubmit;
    }
  }

  return stat;
}

}; // Namespace logmonitor

} JOBCONTROL_NAMESPACE_END;

