#include "edg/workload/logging/client/JobStatus.h"
#include "edg/workload/logging/client/LoggingExceptions.h"
#include "edg/workload/common/jobid/JobId.h"

#include "edg/workload/logging/client/consumer.h"

#include <utility>
#include <vector>
#include <string>
#include <errno.h>

#include <stdlib.h>


EWL_BEGIN_NAMESPACE;

#define CLASS_PREFIX "edg::workload::logging::JobStatus::"


JobStatus::JobStatus(void)
{
  status = UNDEF;
  flesh = 0;
}

JobStatus::JobStatus(const JobStatus & in)
{
  status = in.status;
  flesh = in.flesh;
  if (flesh) flesh->use();
}

JobStatus & 
JobStatus::operator=(const JobStatus & in)
{
  if (flesh) flesh->release();
  status = in.status;
  flesh = in.flesh;
  if (flesh) flesh->use();
  return *this;
}

JobStatus::JobStatus(const edg_wll_JobStat & in)
{
  status = (Code)in.state;
  flesh = new CountRef<JobStatus>((void*)&in);
}
 
JobStatus & 
JobStatus::operator=(const edg_wll_JobStat & in)
{
  if(flesh)
    flesh->release();
  status = (Code)in.state;
  flesh = new CountRef<JobStatus>((void*)&in);
  return(*this);
}

JobStatus::~JobStatus()
{
  if (flesh) flesh->release();
}

edg_wll_JobStat *
JobStatus::c_ptr(void)
{
  edg_wll_JobStat *s;

  if(flesh)
    return((edg_wll_JobStat*)flesh->ptr);

  s = new edg_wll_JobStat;
  // XXX - is it neccessary? new should throw exception itself... 
  if(!s) throw(Exception(EXCEPTION_MANDATORY,
			 ENOMEM,
			 "out of memory allocating c-struct for JobStatus"));

  edg_wll_InitStatus(s);
  flesh = new CountRef<JobStatus>((void*)s);
  return(s);
}


static std::string const names[JobStatus::CODE_MAX] = {
	"undefined",
	"Submitted",
	"Waiting",
	"Ready",
	"Scheduled",
	"Running",
	"Done",
	"Cleared",
	"Aborted",
	"Cancelled",
	"Unknown",
	"Purged",
};

const std::string& JobStatus::name(void) const
{
  if (status<0 || status>=CODE_MAX) {
    STACK_ADD;
    throw(Exception(EXCEPTION_MANDATORY, 
		    EINVAL,
		    "status code invalid"));
  }
  return names[status];
}

int JobStatus::getValInt(JobStatus::Attr attr) const
{
  edg_wll_JobStat const *cstat = (edg_wll_JobStat *) flesh->ptr;
	switch (attr) {
		case JobStatus::DONE_CODE: return cstat->done_code;
		case JobStatus::CHILDREN_NUM: return cstat->children_num;
		case JobStatus::JOBTYPE: return cstat->jobtype;
		case JobStatus::EXIT_CODE: return cstat->exit_code;
		case JobStatus::CPU_TIME: return cstat->cpuTime;
		default: break;
	}

 STACK_ADD; 
 throw(Exception(EXCEPTION_MANDATORY,
		 ENOENT,
		 "no such attribute"));
 return -1;	/* make gcc shut up -- never returns */
}

bool JobStatus::getValBool(JobStatus::Attr attr) const
{
  edg_wll_JobStat const *cstat = (edg_wll_JobStat *) flesh->ptr;
	switch (attr) {
		case JobStatus::RESUBMITTED: return cstat->resubmitted;
		case JobStatus::CANCELLING: return cstat->cancelling;
		case JobStatus::EXPECT_UPDATE: return cstat->expectUpdate;
		case JobStatus::SUBJOB_FAILED: return cstat->subjob_failed;
		default: break;
	}

 STACK_ADD; 
 throw(Exception(EXCEPTION_MANDATORY,
		 ENOENT,
		 "no such attribute"));
 return -1;	/* make gcc shut up -- never returns */
}

static const char *
get_string_val(const edg_wll_JobStat *cstat,JobStatus::Attr attr)
{
	switch (attr) {
		case JobStatus::CANCEL_REASON: return cstat->cancelReason;
		case JobStatus::DESTINATION: return cstat->destination;
		case JobStatus::MATCHED_JDL: return cstat->matched_jdl;
		case JobStatus::CONDOR_ID: return cstat->condorId;
		case JobStatus::EXPECT_FROM: return cstat->expectFrom;
		case JobStatus::LOCATION: return cstat->location;
		case JobStatus::ACL: return cstat->acl;
		case JobStatus::NETWORK_SERVER: return cstat->network_server;
		case JobStatus::GLOBUS_ID: return cstat->globusId;
		case JobStatus::REASON: return cstat->reason;
		case JobStatus::SEED: return cstat->seed;
		case JobStatus::CE_NODE: return cstat->ce_node;
		case JobStatus::OWNER: return cstat->owner;
		case JobStatus::RSL: return cstat->rsl;
		case JobStatus::JDL: return cstat->jdl;
		case JobStatus::CONDOR_JDL: return cstat->condor_jdl;
		case JobStatus::LOCAL_ID: return cstat->localId;
		default: break;
	}
 STACK_ADD;
 throw(Exception(EXCEPTION_MANDATORY,
		 ENOENT,
		 "no such attribute"));
 return 0;	/* make gcc shut up -- never returns */
}

std::string 
JobStatus::getValString(JobStatus::Attr attr) const
{
  edg_wll_JobStat const *cstat = (edg_wll_JobStat *) flesh->ptr;
  std::string	ret;
  
  try{
    const char *s = get_string_val(cstat,attr);
    if (s) ret.assign(s);
    return ret;
  } catch (Exception &e) {
    STACK_ADD;
    throw;
  }
}


struct timeval 
JobStatus::getValTime(Attr attr) const
{
  edg_wll_JobStat const *cstat = (edg_wll_JobStat *) flesh->ptr;

	switch (attr) {
		case JobStatus::LAST_UPDATE_TIME: return cstat->lastUpdateTime;
		case JobStatus::STATE_ENTER_TIME: return cstat->stateEnterTime;
		default: break;
	}
 STACK_ADD;
 throw(Exception(EXCEPTION_MANDATORY,
		 ENOENT,
		 "no such attribute"));
}


const
edg_wlc_JobId 
get_val_jobid(edg_wll_JobStat const *cstat, JobStatus::Attr attr)
{
	switch (attr) {
		case JobStatus::PARENT_JOB: return cstat->parent_job;
		case JobStatus::JOB_ID: return cstat->jobId;
		default: break;
	}
 STACK_ADD;
 throw(Exception(EXCEPTION_MANDATORY,
		 ENOENT,
		 "no such attribute"));
 return 0;	/* make gcc shut up -- never returns */
}


const 
edg::workload::common::jobid::JobId  
JobStatus::getValJobId(Attr attr) const
{
  edg_wll_JobStat const *cstat = (edg_wll_JobStat *) flesh->ptr;

  try {
	  edg_wlc_JobId job_id = get_val_jobid(cstat, attr);
	  return(edg::workload::common::jobid::JobId(job_id));
  }
  catch(Exception &e) {
	  STACK_ADD;
	  throw;
  }
}

const
int *
get_val_intlist(edg_wll_JobStat const *cstat, JobStatus::Attr attr)
{
	switch (attr) {
		case JobStatus::CHILDREN_HIST: return cstat->children_hist;
		case JobStatus::STATE_ENTER_TIMES: return cstat->stateEnterTimes;
		default: break;
	}
  STACK_ADD;
  throw(Exception(EXCEPTION_MANDATORY,
		  ENOENT,
		  "no such attribute"));
  return 0;	/* make gcc shut up -- never returns */
}


const 
std::vector<int>
JobStatus::getValIntList(Attr attr) const
{
	edg_wll_JobStat const *cstat = (edg_wll_JobStat *) flesh->ptr;

	std::vector<int> result;
	const int *r;
	
	try {
		r = get_val_intlist(cstat, attr);
		if(r)
			for(int i = 1; i <= r[0]; i++)
				result.push_back(r[i]);
	} catch(Exception &e) {
		STACK_ADD;
		throw;
	}
	return result;
}


char ** const
get_val_stringlist(edg_wll_JobStat const *cstat, JobStatus::Attr attr)
{
	switch (attr) {
		case JobStatus::CHILDREN: return cstat->children;
		default: break;
	}
  STACK_ADD;
  throw(Exception(EXCEPTION_MANDATORY,
		  ENOENT,
		  "no such attribute"));
  return 0;	/* make gcc shut up -- never returns */
}


const
std::vector<std::string>
JobStatus::getValStringList(Attr attr) const
{
	edg_wll_JobStat const *cstat = (edg_wll_JobStat *) flesh->ptr;

	std::vector<std::string> result;
	char **r , **p;

	try {
		r = (char**)get_val_stringlist(cstat, attr);
		if(r)
			for(p = r; *p; p++)
				result.push_back(std::string(*p));
	} catch(Exception &e) {
		STACK_ADD;
		throw;
	}
	return result;
}


edg_wll_TagValue * const
get_val_taglist(edg_wll_JobStat const *cstat, JobStatus::Attr attr)
{
	switch (attr) {
		case JobStatus::USER_TAGS: return cstat->user_tags;
		default: break;
	}
  STACK_ADD;
  throw(Exception(EXCEPTION_MANDATORY,
		  ENOENT,
		  "no such attribute"));
  return 0;	/* make gcc shut up -- never returns */
}


const
std::vector<std::pair<std::string,std::string> >
JobStatus::getValTagList(Attr attr) const
{
	edg_wll_JobStat const *cstat = (edg_wll_JobStat *) flesh->ptr;

	std::vector<std::pair<std::string,std::string> > result;
	edg_wll_TagValue *r , *p;

	try {
		r = get_val_taglist(cstat, attr);
		if(r)
			for(p = r; p->tag ; p++)
				result.push_back(std::pair<std::string,std::string>
							(std::string(p->tag),std::string(p->value)));
	} catch(Exception &e) {
		STACK_ADD;
		throw;
	}
	return result;
}


const
edg_wll_JobStat *
get_val_stslist(edg_wll_JobStat const *cstat, JobStatus::Attr attr)
{
	switch (attr) {
		case JobStatus::CHILDREN_STATES: return cstat->children_states;
		default: break;
	}
  STACK_ADD;
  throw(Exception(EXCEPTION_MANDATORY,
		  ENOENT,
		  "no such attribute"));
  return 0;	/* make gcc shut up -- never returns */
}


const
std::vector<JobStatus>
JobStatus::getValJobStatusList(Attr attr) const
{
	edg_wll_JobStat const *cstat = (edg_wll_JobStat *) flesh->ptr;

	std::vector<JobStatus> result;
	const edg_wll_JobStat *r, *p;

	try {
		r = get_val_stslist(cstat, attr);
		if(r)
			for(p=r; p->state != EDG_WLL_JOB_UNDEF; p++) {
				edg_wll_JobStat *jsep = new edg_wll_JobStat;
				memcpy(jsep,p, sizeof(*jsep));
				result.push_back(JobStatus(*jsep));
			}
	} catch(Exception &e) {
		STACK_ADD;
		throw;
	}
	return(result);
}


static std::string const attr_names[JobStatus::ATTR_MAX] = {
	"acl",
	"cancelReason",
	"cancelling",
	"ce_node",
	"children",
	"children_hist",
	"children_num",
	"children_states",
	"condorId",
	"condor_jdl",
	"cpuTime",
	"destination",
	"done_code",
	"exit_code",
	"expectFrom",
	"expectUpdate",
	"globusId",
	"jdl",
	"jobId",
	"jobtype",
	"lastUpdateTime",
	"localId",
	"location",
	"matched_jdl",
	"network_server",
	"owner",
	"parent_job",
	"reason",
	"resubmitted",
	"rsl",
	"seed",
	"stateEnterTime",
	"stateEnterTimes",
	"subjob_failed",
	"user_tags",
};

const std::string & 
JobStatus::getAttrName(JobStatus::Attr attr) const
{
  if (attr<0 || attr>=ATTR_MAX) {
    STACK_ADD;
    throw(Exception(EXCEPTION_MANDATORY,
		    ENOENT,
		    "no such attribute"));
  }

  return attr_names[attr];
}


typedef std::pair<JobStatus::Attr,JobStatus::AttrType> tpair;
static std::vector<tpair> attrs;

static bool attrs_inited = false;

static void init_attrs(void)
{
/* XXX: currently only common attributes in JobStatus */
	attrs.push_back(tpair(JobStatus::CANCEL_REASON,JobStatus::STRING_T));
	attrs.push_back(tpair(JobStatus::PARENT_JOB,JobStatus::JOBID_T));
	attrs.push_back(tpair(JobStatus::DESTINATION,JobStatus::STRING_T));
	attrs.push_back(tpair(JobStatus::LAST_UPDATE_TIME,JobStatus::TIMEVAL_T));
	attrs.push_back(tpair(JobStatus::MATCHED_JDL,JobStatus::STRING_T));
	attrs.push_back(tpair(JobStatus::CHILDREN,JobStatus::STRLIST_T));
	attrs.push_back(tpair(JobStatus::CONDOR_ID,JobStatus::STRING_T));
	attrs.push_back(tpair(JobStatus::DONE_CODE,JobStatus::INT_T));
	attrs.push_back(tpair(JobStatus::EXPECT_FROM,JobStatus::STRING_T));
	attrs.push_back(tpair(JobStatus::LOCATION,JobStatus::STRING_T));
	attrs.push_back(tpair(JobStatus::ACL,JobStatus::STRING_T));
	attrs.push_back(tpair(JobStatus::RESUBMITTED,JobStatus::BOOL_T));
	attrs.push_back(tpair(JobStatus::NETWORK_SERVER,JobStatus::STRING_T));
	attrs.push_back(tpair(JobStatus::GLOBUS_ID,JobStatus::STRING_T));
	attrs.push_back(tpair(JobStatus::REASON,JobStatus::STRING_T));
	attrs.push_back(tpair(JobStatus::CHILDREN_NUM,JobStatus::INT_T));
	attrs.push_back(tpair(JobStatus::CHILDREN_HIST,JobStatus::INTLIST_T));
	attrs.push_back(tpair(JobStatus::CHILDREN_STATES,JobStatus::STSLIST_T));
	attrs.push_back(tpair(JobStatus::JOBTYPE,JobStatus::INT_T));
	attrs.push_back(tpair(JobStatus::CANCELLING,JobStatus::BOOL_T));
	attrs.push_back(tpair(JobStatus::STATE_ENTER_TIME,JobStatus::TIMEVAL_T));
	attrs.push_back(tpair(JobStatus::SEED,JobStatus::STRING_T));
	attrs.push_back(tpair(JobStatus::EXPECT_UPDATE,JobStatus::BOOL_T));
	attrs.push_back(tpair(JobStatus::CE_NODE,JobStatus::STRING_T));
	attrs.push_back(tpair(JobStatus::EXIT_CODE,JobStatus::INT_T));
	attrs.push_back(tpair(JobStatus::STATE_ENTER_TIMES,JobStatus::INTLIST_T));
	attrs.push_back(tpair(JobStatus::OWNER,JobStatus::STRING_T));
	attrs.push_back(tpair(JobStatus::CPU_TIME,JobStatus::INT_T));
	attrs.push_back(tpair(JobStatus::USER_TAGS,JobStatus::TAGLIST_T));
	attrs.push_back(tpair(JobStatus::RSL,JobStatus::STRING_T));
	attrs.push_back(tpair(JobStatus::JDL,JobStatus::STRING_T));
	attrs.push_back(tpair(JobStatus::JOB_ID,JobStatus::JOBID_T));
	attrs.push_back(tpair(JobStatus::CONDOR_JDL,JobStatus::STRING_T));
	attrs.push_back(tpair(JobStatus::LOCAL_ID,JobStatus::STRING_T));
	attrs.push_back(tpair(JobStatus::SUBJOB_FAILED,JobStatus::BOOL_T));
}


const std::vector<tpair>& 
JobStatus::getAttrs(void) const
{
  if (status<0 || status>=CODE_MAX) {
    STACK_ADD;
    throw(Exception(EXCEPTION_MANDATORY,
		    EINVAL,
		    "status code invalid"));
  }

/* FIXME: thread safety */
  if (!attrs_inited) {
    init_attrs();
    attrs_inited = true;
  }
  return attrs;
}

void 
JobStatus::destroyFlesh(void *p)
{
  edg_wll_JobStat *stat = (edg_wll_JobStat *) p;
  if (stat) {
    edg_wll_FreeStatus(stat);
    free(stat);
  }
}

EWL_END_NAMESPACE;
