/**
 * \file producer.c
 * \author Jan Pospisil
 */

#ident "$Header: /local/reps/lcgware/workload/logging/client/producer.c,v 1.73 2004/01/14 15:47:54 ljocha Exp $"

#include "edg/workload/common/jobid/strmd5.h"
#include "edg/workload/logging/client/consumer.h"
#include "edg/workload/logging/client/producer.h"
#include "edg/workload/logging/client/prod_proto.h"
#include "edg/workload/logging/common/context.h"
#include "edg/workload/logging/common/ulm_parse.h"

#include "edg/workload/thirdparty/trio/trio.h"

#include <globus_gss_assist.h>
#include <globus_config.h>
#include <sslutils.h>

#include <syslog.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>

/**
 *----------------------------------------------------------------------
 * Connects to local-logger and sends already formatted ULM string
 * \brief helper logging function
 * \param context	INOUT context to work with,
 * \param priority	IN priority flag (0 for async, 1 for sync)
 * \param logline	IN formated ULM string
 *----------------------------------------------------------------------
 */
static int do_log_event(
	edg_wll_Context context,
/*	int priority,  */
	edg_wll_LogLine logline)
{
	int	ret = 0, answer = 0;
	void	*cred_handle = NULL;
	SSL	*ssl = NULL;
#ifdef EDG_WLL_LOG_STUB
	char	*my_subject_name = NULL;
#endif
	edg_wll_ResetError(context);

   /* open a SSL connection to the local-logger: */

#ifdef EDG_WLL_LOG_STUB
	fprintf(stderr,"Logging to host %s, port %d\n",
			context->p_destination, context->p_dest_port);
#endif
	
	if ((cred_handle=edg_wll_ssl_init(SSL_VERIFY_PEER, 0,
			context->p_proxy_filename ? context->p_proxy_filename : context->p_cert_filename,
			context->p_proxy_filename ? context->p_proxy_filename : context->p_key_filename,
			0, 0)) == NULL) {
		edg_wll_SetError(context,ret = ECONNREFUSED,"cred_handle"); 
		goto do_log_event_end; 
	}
#ifdef EDG_WLL_LOG_STUB
	edg_wll_ssl_get_my_subject(cred_handle,&my_subject_name);
	if (my_subject_name != NULL) {
		fprintf(stderr,"Using certificate: %s\n",my_subject_name);
		free(my_subject_name);
	}
#endif
	if ((answer = edg_wll_ssl_connect(cred_handle,
			context->p_destination, context->p_dest_port, 
			&context->p_tmp_timeout, &ssl)) < 0) {
		switch(answer) {
                case EDG_WLL_SSL_ERROR_EOF:
                case EDG_WLL_SSL_ERROR_TIMEOUT:
		  edg_wll_SetError(context,ETIMEDOUT,"edg_wll_ssl_connect()");
		  break;
                case EDG_WLL_SSL_ERROR_ERRNO:
		  edg_wll_SetError(context,errno,"edg_wll_ssl_connect()");
			break;
                case EDG_WLL_SSL_ERROR_SSL:
                        edg_wll_SetError(context,EDG_WLL_ERROR_SSL,ERR_reason_error_string(ERR_get_error()));
			break;
		case EDG_WLL_SSL_ERROR_HERRNO:
			{ const char *msg1;
		          char *msg2;
			  msg1 = hstrerror(errno);
			  asprintf(&msg2, "edg_wll_ssl_connect(): %s", msg1);
                          edg_wll_SetError(context,ENOTCONN, msg2);
			  free(msg2);
			}
                        break;
		default:
			edg_wll_SetError(context,ret = ECONNREFUSED,
					"edg_wll_ssl_connect(): unknown");
			break;
		}
		goto do_log_event_end;
	}

	if ((context->p_tmp_timeout.tv_sec < 0) || (context->p_tmp_timeout.tv_usec < 0)) {
		edg_wll_SetError(context,ret = EAGAIN,"edg_wll_ssl_connect(): timeout expired");
		context->p_tmp_timeout.tv_sec = 0;
		context->p_tmp_timeout.tv_usec = 0;
		goto do_log_event_end; 
	}
	
   /* and send the message to the local-logger: */

	edg_wll_ResetError(context);
	answer = edg_wll_log_proto_client(context,ssl,logline/*,priority*/);

	switch(answer) {
		case 0:
		case EINVAL:
		case ENOSPC:
		case ENOMEM:
			break;
		default:
			/* edg_wll_SetError() has been called in log_proto_client(),
			   so fetch the text and description and place it into 
			   the generic EAGAIN error as a description */
		{
			char *err_text, *err_desc, *err;
			
			edg_wll_Error(context, &err_text, &err_desc);
			asprintf(&err, "%s - %s", err_text, err_desc);
			edg_wll_SetError(context, EAGAIN, err);
			if(err) free(err);
			if(err_text) free(err_text);
			if(err_desc) free(err_desc);
		}
		break;
	}

do_log_event_end:
	if (ssl) edg_wll_ssl_close_timeout(ssl,&context->p_tmp_timeout);
	if (cred_handle) edg_wll_ssl_free(cred_handle);
	return edg_wll_Error(context, NULL, NULL);
}


/**
 *----------------------------------------------------------------------
 * Formats a logging message and sends it to local-logger
 * \brief master logging event function
 * \param context	INOUT context to work with,
 * \param priority	IN priority flag (0 for async, 1 for sync)
 * \param event		IN type of the event,
 * \param fmt		IN printf()-like format string,
 * \param ...		IN event specific values/data according to fmt.
 *----------------------------------------------------------------------
 */
static int edg_wll_LogEventMaster(
	edg_wll_Context context,
	int priority,
	edg_wll_EventCode event,
	char *fmt, ...)
{
	va_list	fmt_args;
	int	ret,answer;
	char	*fix,*var;
	char	*source,*eventName,*lvl, *fullid,*seq;
        struct timeval start_time;
	char	date[ULM_DATE_STRING_LENGTH+1];
	edg_wll_LogLine out;
	size_t  size;
	int	i;


	i = errno  =  size = 0;
	seq = fix = var = out = source = eventName = lvl = fullid = NULL;

	edg_wll_ResetError(context);

   /* default return value is "Try Again" */
	answer = ret = EAGAIN; 

   /* format the message: */
	va_start(fmt_args,fmt);

	gettimeofday(&start_time,0);
	if (edg_wll_ULMTimevalToDate(start_time.tv_sec,start_time.tv_usec,date) != 0) {
		edg_wll_SetError(context,ret = EINVAL,"edg_wll_ULMTimevalToDate()"); 
		goto edg_wll_logeventmaster_end; 
	}
 	source = edg_wll_SourceToString(context->p_source);
// XXX: ParseEvent did not understand this:
	lvl = edg_wll_LevelToString(context->p_level);
// should work now, however if not, use this:
//	trio_asprintf(&lvl,"%d",context->p_level);
	eventName = edg_wll_EventToString(event);
	if (!eventName) { 
		ret = EINVAL; 
		goto edg_wll_logeventmaster_end; 
	}
	if (!(fullid = edg_wlc_JobIdUnparse(context->p_jobid))) { 
		ret = EINVAL;
		goto edg_wll_logeventmaster_end;
	}

	seq = edg_wll_GetSequenceCode(context);

	if (edg_wll_IncSequenceCode(context)) {
		ret = EINVAL;
		goto edg_wll_logeventmaster_end;
	}

	if (trio_asprintf(&fix,EDG_WLL_FORMAT_COMMON,
			date,context->p_host,lvl,priority,
			source,context->p_instance ? context->p_instance : "",
			eventName,fullid,seq) == -1) {
		edg_wll_SetError(context,ret = ENOMEM,"trio_asprintf"); 
		goto edg_wll_logeventmaster_end; 
	}
	if (trio_vasprintf(&var,fmt,fmt_args) == -1) { 
		edg_wll_SetError(context,ret = ENOMEM,"trio_vasprintf"); 
		goto edg_wll_logeventmaster_end; 
	}
	if (asprintf(&out,"%s%s\n",fix,var) == -1) { 
		edg_wll_SetError(context,ret = ENOMEM,"asprintf"); 
		goto edg_wll_logeventmaster_end; 
	}
	size = strlen(out);

	if (priority && (size > EDG_WLL_LOG_SYNC_MAXMSGSIZE)) {
		fprintf(stderr,"edg_wll_LogEventMaster: Message size too large for synchronous transfer");
		ret = ENOSPC;
		goto edg_wll_logeventmaster_end;
	}

#ifdef EDG_WLL_LOG_STUB
//	fprintf(stderr,"edg_wll_LogEvent (%d chars): %s",size,out);
#endif
	
	context->p_tmp_timeout.tv_sec = 0;
	context->p_tmp_timeout.tv_usec = 0;
	if (priority) {
		context->p_tmp_timeout = context->p_sync_timeout;
	}
	else {
		context->p_tmp_timeout = context->p_log_timeout;
	}

	do_log_event(context, /* priority,*/ out);
	ret = 0;

edg_wll_logeventmaster_end:
	if (seq) free(seq); 
	if (fix) free(fix); 
	if (var) free(var); 
	if (out) free(out);
	if (source) free(source);
	if (lvl) free(lvl);
	if (eventName) free(eventName);
	if (fullid) free(fullid);
	va_end(fmt_args);
	if (ret) edg_wll_SetError(context,ret,"logging library");
	return edg_wll_Error(context,NULL,NULL);
}

/**
 *----------------------------------------------------------------------
 * Formats a logging message and sends it asynchronously to local-logger
 * \brief generic asynchronous logging function
 *----------------------------------------------------------------------
 */
int edg_wll_LogEvent(
        edg_wll_Context context,
        edg_wll_EventCode event,
        char *fmt, ...)
{
	int	ret=0;
	char	*list=NULL;
	va_list	fmt_args;

	edg_wll_ResetError(context);

	va_start(fmt_args,fmt);
	if (trio_vasprintf(&list,fmt,fmt_args) == -1) { 
		edg_wll_SetError(context,ret = ENOMEM,"trio_vasprintf"); 
		goto edg_wll_logevent_end; 
	}

	ret=edg_wll_LogEventMaster(context,0,event,"%s",list);

edg_wll_logevent_end:
	va_end(fmt_args);
        if (list!=NULL) free(list);
	return ret;
}

/**
 *----------------------------------------------------------------------
 * Formats a logging message and sends it synchronously to local-logger
 * \brief generic synchronous logging function
 * \note simple wrapper around edg_wll_LogEventMaster()
 *----------------------------------------------------------------------
 */
int edg_wll_LogEventSync(
        edg_wll_Context context,
        edg_wll_EventCode event,
        char *fmt, ...)
{
	int	ret=0;
	char	*list=NULL;
	va_list	fmt_args;

	edg_wll_ResetError(context);

	va_start(fmt_args,fmt);
	if (trio_vasprintf(&list,fmt,fmt_args) == -1) { 
		edg_wll_SetError(context,ret = ENOMEM,"trio_vasprintf"); 
		goto edg_wll_logeventsync_end; 
	}

	ret=edg_wll_LogEventMaster(context,1,event,"%s",list);

edg_wll_logeventsync_end:
	va_end(fmt_args);
        if (list!=NULL) free(list);
	return ret;
}


/**
 *-----------------------------------------------------------------------
 * Instructs interlogger to to deliver all pending events related to current job
 * \brief flush events from interlogger
 * \note simple wrapper around edg_wll_LogEventMaster()
 *-----------------------------------------------------------------------
 */
int edg_wll_LogFlush(
        edg_wll_Context context,
        struct timeval *timeout)
{
	int 	ret = 0;
	edg_wll_LogLine out = NULL;
	char 	*fullid;
	char	date[ULM_DATE_STRING_LENGTH+1];
        struct timeval start_time;

	fullid = NULL;

	edg_wll_ResetError(context);

	gettimeofday(&start_time, 0);
	if (edg_wll_ULMTimevalToDate(start_time.tv_sec, start_time.tv_usec, date) != 0) {
		edg_wll_SetError(context,ret = EINVAL,"edg_wll_ULMTimevalToDate()"); 
		goto edg_wll_logflush_end; 
	}
	if (!(fullid = edg_wlc_JobIdUnparse(context->p_jobid))) { 
		ret = edg_wll_SetError(context,EINVAL,"edg_wlc_JobIdUnparse()");
		goto edg_wll_logflush_end;
	}

	if (trio_asprintf(&out, "DATE=%s HOST=\"%|Us\" PROG=internal LVL=system DG.PRIORITY=1 DG.TYPE=\"command\" DG.COMMAND=\"flush\" DG.TIMEOUT=\"%d\" DG.JOBID=\"%s\"\n", 
		    date, context->p_host, (timeout ? timeout->tv_sec : context->p_sync_timeout.tv_sec), fullid) == -1) {
		edg_wll_SetError(context,ret = EINVAL,"trio_asprintf");
		goto edg_wll_logflush_end;
	}

	if (timeout)
		context->p_tmp_timeout = *timeout;
	else
		context->p_tmp_timeout = context->p_sync_timeout;

	ret = do_log_event(context, /* 1,*/ out);

/* FIXME: handle return value: 0 - events waiting, 1 - events sent, < 0 interlogger error */

edg_wll_logflush_end:
	  if(out) free(out);
	  if(fullid) free(fullid);
	  return(ret);
}

/**
 *-----------------------------------------------------------------------
 * Instructs interlogger to to deliver all pending events
 * \brief flush all events from interlogger
 *-----------------------------------------------------------------------
 */
int edg_wll_LogFlushAll(
        edg_wll_Context context,
        struct timeval *timeout)
{
	int 	ret = 0;
	edg_wll_LogLine out = NULL;
	char	date[ULM_DATE_STRING_LENGTH+1];
        struct timeval start_time;

	edg_wll_ResetError(context);

	gettimeofday(&start_time, 0);
	if (edg_wll_ULMTimevalToDate(start_time.tv_sec, start_time.tv_usec, date) != 0) {
		edg_wll_SetError(context,ret = EINVAL,"edg_wll_ULMTimevalToDate()"); 
		goto edg_wll_logflushall_end; 
	}

	if (trio_asprintf(&out, "DATE=%s HOST=\"%|Us\" PROG=internal LVL=system DG.PRIORITY=1 DG.TYPE=\"command\" DG.COMMAND=\"flush\" DG.TIMEOUT=\"%d\"\n", 
		    date, context->p_host, (timeout ? timeout->tv_sec : context->p_sync_timeout.tv_sec)) == -1) {
		edg_wll_SetError(context,ret = ENOMEM,"trio_asprintf");
		goto edg_wll_logflushall_end;
	}

	if (timeout)
		context->p_tmp_timeout = *timeout;
	else
		context->p_tmp_timeout = context->p_sync_timeout;

	ret = do_log_event(context, /* 1,*/ out);

#ifdef EDG_WLL_LOG_STUB
	fprintf(stderr, "Return from do_log_event(...): %d\n", ret);
#endif

/* FIXME: handle return value: 0 - events waiting, 1 - events sent, < 0 interlogger error */

edg_wll_logflushall_end:
	  if(out) free(out);
	  return(ret);
}

/**
 *-----------------------------------------------------------------------
 * Set a current job for given context.
 * \note Should be called before any logging call.
 *-----------------------------------------------------------------------
 */
int edg_wll_SetLoggingJob(
	edg_wll_Context context,
	const edg_wlc_JobId job,
	const char *code,
	int flags)
{
	int	err;

	edg_wll_ResetError(context);

	if (!job) return edg_wll_SetError(context,EINVAL,"jobid is null");

	edg_wlc_JobIdFree(context->p_jobid);
	if ((err = edg_wlc_JobIdDup(job,&context->p_jobid)))
		edg_wll_SetError(context,err,"edg_wlc_JobIdDup()");

	else {
		if (!edg_wll_SetSequenceCode(context,code,flags))
			edg_wll_IncSequenceCode(context);
	}

	return edg_wll_Error(context,NULL,NULL);
}

/**
 *-----------------------------------------------------------------------
 * Register job with L&B service.
 *-----------------------------------------------------------------------
 */
static int edg_wll_RegisterJobMaster(
        edg_wll_Context         context,
	int			pri,
        const edg_wlc_JobId     job,
        enum edg_wll_RegJobJobtype	type,
        const char *            jdl,
        const char *            ns,
        int                     num_subjobs,
        const char *            seed,
        edg_wlc_JobId **        subjobs)
{
	char	*type_s = NULL,*intseed = NULL, *seq = NULL;
	int	err = 0;

	edg_wll_ResetError(context);

	intseed = seed ? strdup(seed) : 
		str2md5base64(seq = edg_wll_GetSequenceCode(context));

	free(seq);

	type_s = edg_wll_RegJobJobtypeToString(type);
	if (!type_s) return edg_wll_SetError(context,EINVAL,"jobtype");

	if ((type == EDG_WLL_REGJOB_DAG || type == EDG_WLL_REGJOB_PARTITIONED)
		&& num_subjobs > 0) 
			err = edg_wll_GenerateSubjobIds(context,job,
					num_subjobs,intseed,subjobs);

	if (err == 0 &&
		edg_wll_SetLoggingJob(context,job,NULL,EDG_WLL_SEQ_NORMAL) == 0)
			edg_wll_LogEventMaster(context,pri,
				EDG_WLL_EVENT_REGJOB,EDG_WLL_FORMAT_REGJOB,
				(char *)jdl,ns,"",type_s,num_subjobs,intseed);

	free(type_s); free(intseed);
	return edg_wll_Error(context,NULL,NULL);
}

int edg_wll_RegisterJobSync(
        edg_wll_Context         context,
        const edg_wlc_JobId     job,
        enum edg_wll_RegJobJobtype	type,
        const char *            jdl,
        const char *            ns,
        int                     num_subjobs,
        const char *            seed,
        edg_wlc_JobId **        subjobs)
{
	return edg_wll_RegisterJobMaster(context,1,job,type,jdl,ns,num_subjobs,seed,subjobs);
}

int edg_wll_RegisterJob(
        edg_wll_Context         context,
        const edg_wlc_JobId     job,
        enum edg_wll_RegJobJobtype	type,
        const char *            jdl,
        const char *            ns,
        int                     num_subjobs,
        const char *            seed,
        edg_wlc_JobId **        subjobs)
{
	return edg_wll_RegisterJobMaster(context,0,job,type,jdl,ns,num_subjobs,seed,subjobs);
}

int edg_wll_RegisterSubjobs(edg_wll_Context ctx,const edg_wlc_JobId parent,
		char const * const * jdls, const char * ns, edg_wlc_JobId const * subjobs)
{
	return edg_wll_SetError(ctx,ENOSYS,"unimpelemeted");
}


int edg_wll_ChangeACL(
		edg_wll_Context				ctx,
		const edg_wlc_JobId			jobid,
		const char				   *user_id,
		enum edg_wll_UserIdType		user_id_type,
		enum edg_wll_Permission		permission,
		enum edg_wll_PermissionType	permission_type,
		enum edg_wll_ACLOperation	operation)
{
	if ( edg_wll_SetLoggingJob(ctx, jobid, NULL, EDG_WLL_SEQ_NORMAL) == 0 )
		edg_wll_LogEventMaster(ctx, 1, EDG_WLL_EVENT_CHANGEACL, EDG_WLL_FORMAT_CHANGEACL,
				user_id, user_id_type, permission, permission_type, operation);


	return edg_wll_Error(ctx,NULL,NULL);
}

