#ident "$Header: /local/reps/lcgware/workload/logging/common/context.c,v 1.42 2004/06/14 16:09:40 dhsmith Exp $"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#include "globus_config.h"
#include "sslutils.h"

#include "edg/workload/common/jobid/strmd5.h"
#include "edg/workload/common/jobid/jobid.h"
#include "edg/workload/logging/common/context.h"
#include "edg/workload/logging/client/producer.h"

static void free_voms_groups(edg_wll_VomsGroups *);

int edg_wll_InitContext(edg_wll_Context *ctx)
{
	int i;
	edg_wll_Context out	= (edg_wll_Context) malloc(sizeof(*out));

	if (!out) return ENOMEM;
	memset(out,0,sizeof(*out));
	assert(out->errDesc == NULL);

	out->allowAnonymous = 1;

	/* XXX */
	for (i=0; i<EDG_WLL_PARAM__LAST; i++) edg_wll_SetParam(out,i,NULL);

	*ctx = out;
	return 0;
}

void edg_wll_FreeContext(edg_wll_Context ctx)
{
	struct timeval close_timeout = {0, 50000};

	if (!ctx) return;

	if (ctx->errDesc) free(ctx->errDesc);
	if (ctx->srvName) free(ctx->srvName);
	if (ctx->peerName) free(ctx->peerName);
	if (ctx->vomsGroups.len) free_voms_groups(&ctx->vomsGroups);
	if (ctx->buf) free(ctx->buf);
	if (ctx->dumpStorage) free(ctx->dumpStorage);
	if (ctx->purgeStorage) free(ctx->purgeStorage);
	if (ctx->ssl) edg_wll_ssl_close_timeout(ctx->ssl,&close_timeout);
	if (ctx->gsiCred) edg_wll_ssl_free(ctx->gsiCred);

        if (ctx->p_jobid) edg_wlc_JobIdFree(ctx->p_jobid);
        if (ctx->p_host) free(ctx->p_host);
        if (ctx->p_instance) free(ctx->p_instance);
        if (ctx->p_destination) free(ctx->p_destination);
        if (ctx->p_query_server) free(ctx->p_query_server);
        if (ctx->p_proxy_filename) free(ctx->p_proxy_filename);
        if (ctx->p_cert_filename) free(ctx->p_cert_filename);
        if (ctx->p_key_filename) free(ctx->p_key_filename);

	/* do not free (references only)
	 * ctx->job_index
	 * ctx->job_index_cols
	 * ctx->mysql */

	free(ctx);
}

static const char* const errTexts[] = {
	/* standard error strings should appear here */
	"Broken ULM",
	"Undefined event",
	"Message incomplete",
	"Duplicate ULM key",
	"Misuse of ULM key",
	"Warning: extra ULM fields",
	"XML Parse error",
	"Server response error",
	"Bad JobId format",
	"Database call failed",
	"Bad URL format",
	"MD5 key clash",
	"SSL Error",
	"No JobId specified in context",
	"No indexed condition in query",
	"Interlogger protocol error",
	"Interlogger internal error",
	"Interlogger has events pending"
};

const char *edg_wll_GetErrorText(int code) {
	return code ? 
		(code <= EDG_WLL_ERROR_BASE ?
			strerror(code) :
			errTexts[code - EDG_WLL_ERROR_BASE - 1]
		) :
		NULL;
}

int edg_wll_Error(edg_wll_Context ctx, char **errText, char **errDesc)
{
	char	*text = NULL,*desc = NULL;
	const char* et = edg_wll_GetErrorText(ctx->errCode);

	if (et) {
		text = strdup(et);
		if (ctx->errDesc) desc = (char *) strdup(ctx->errDesc);
	}

	if (errText) *errText = text; else free(text);
	if (errDesc) *errDesc = desc; else free(desc);
	return ctx->errCode;
}

int edg_wll_ResetError(edg_wll_Context ctx)
{
	if (ctx->errDesc) free(ctx->errDesc);
	ctx->errDesc = NULL;
	ctx->errCode =  0;

	return ctx->errCode;
}

int edg_wll_SetError(edg_wll_Context ctx,int code,const char *desc)
{
	edg_wll_ResetError(ctx);
	if (code) {
		ctx->errCode = code;
		if (desc) ctx->errDesc = (char *) strdup(desc);
	}

	return ctx->errCode;
}

static const char* const srcNames[] = {
	"Undefined",
	"UserInterface",
	"NetworkServer",
	"WorkloadManager",
	"BigHelper",
	"JobController",
	"LogMonitor",
	"LRMS",
	"Application",
};

edg_wll_Source edg_wll_StringToSource(const char *name)
{
	int	i;
/* XXX: remove
	for (i=1; srcNames[i] && strcmp(name,srcNames[i]); i++);
	return srcNames[i] ? i : EDG_WLL_SOURCE_NONE;
*/
        for (i=1; i<sizeof(srcNames)/sizeof(srcNames[0]); i++)
                if (strcasecmp(srcNames[i],name) == 0) return (edg_wll_Source) i;
        return EDG_WLL_SOURCE_NONE;
}

char * edg_wll_SourceToString(edg_wll_Source src)
{
        if (src < EDG_WLL_SOURCE_NONE || src >= EDG_WLL_SOURCE__LAST) return NULL;
        return strdup(srcNames[src]);
}

static const char* const queryResultNames[] = {
	"Undefined",
	"None",
	"All",
	"Limited"
};

edg_wll_QueryResults edg_wll_StringToQResult(const char *name)
{
	int	i;

	for (i=1; i<sizeof(queryResultNames)/sizeof(queryResultNames[0]); i++)
		if (strcasecmp(queryResultNames[i],name) == 0)
			return (edg_wll_QueryResults) i;

	return EDG_WLL_QUERYRES_UNDEF;
}

char * edg_wll_QResultToString(edg_wll_QueryResults res)
{
	if (res < EDG_WLL_QUERYRES_NONE || res >= EDG_WLL_QUERYRES__LAST)
		return NULL;

	return strdup(queryResultNames[res]);
}


char *edg_wll_GetSequenceCode(const edg_wll_Context ctx)
{
	unsigned int *c;
	char *ret = NULL;

	c = &ctx->p_seqcode.c[0];
	asprintf(&ret, "UI=%06d:NS=%010d:WM=%06d:BH=%010d:JSS=%06d"
				":LM=%06d:LRMS=%06d:APP=%06d",
			c[EDG_WLL_SOURCE_USER_INTERFACE],
			c[EDG_WLL_SOURCE_NETWORK_SERVER],
			c[EDG_WLL_SOURCE_WORKLOAD_MANAGER],
			c[EDG_WLL_SOURCE_BIG_HELPER],
			c[EDG_WLL_SOURCE_JOB_SUBMISSION],
			c[EDG_WLL_SOURCE_LOG_MONITOR],
			c[EDG_WLL_SOURCE_LRMS],
			c[EDG_WLL_SOURCE_APPLICATION]);
	return ret;
}

int edg_wll_SetSequenceCode(edg_wll_Context ctx,
			const char * seqcode_str, int seq_type)
{
	int res;
	unsigned int *c;
	int duplicate = 0;

	edg_wll_ResetError(ctx);

	if (seq_type == EDG_WLL_SEQ_DUPLICATE) {
		duplicate = 1;
	} else if (seq_type != EDG_WLL_SEQ_NORMAL)
		return edg_wll_SetError(ctx, EINVAL,
			"unrecognized value of seq_type parameter");
	
	if (!seqcode_str) {
		memset(&ctx->p_seqcode,0,sizeof ctx->p_seqcode);
		return 0;
	}

	c = &ctx->p_seqcode.c[0];
	res =  sscanf(seqcode_str, "UI=%d:NS=%d:WM=%d:BH=%d:JSS=%d:LM=%d:LRMS=%d:APP=%d",
			&c[EDG_WLL_SOURCE_USER_INTERFACE],
			&c[EDG_WLL_SOURCE_NETWORK_SERVER],
			&c[EDG_WLL_SOURCE_WORKLOAD_MANAGER],
			&c[EDG_WLL_SOURCE_BIG_HELPER],
			&c[EDG_WLL_SOURCE_JOB_SUBMISSION],
			&c[EDG_WLL_SOURCE_LOG_MONITOR],
			&c[EDG_WLL_SOURCE_LRMS],
			&c[EDG_WLL_SOURCE_APPLICATION]);

	assert(EDG_WLL_SOURCE__LAST == 9);
	if (res != EDG_WLL_SOURCE__LAST-1)
		return edg_wll_SetError(ctx, EINVAL, "syntax error in sequence code");

	if (duplicate) {
		if (ctx->p_source <= EDG_WLL_SOURCE_NONE || ctx->p_source >= EDG_WLL_SOURCE__LAST)
			return edg_wll_SetError(ctx,EINVAL,"context param: source missing");
		c[ctx->p_source] = time(NULL);
	}

	return edg_wll_Error(ctx, NULL, NULL);
}

int edg_wll_IncSequenceCode(edg_wll_Context ctx)
{
	if (ctx->p_source <= EDG_WLL_SOURCE_NONE || ctx->p_source >= EDG_WLL_SOURCE__LAST)
		return edg_wll_SetError(ctx,EINVAL,"context param: source missing");

	ctx->p_seqcode.c[ctx->p_source]++;
	return edg_wll_ResetError(ctx);
}

int edg_wll_GetLoggingJob(const edg_wll_Context ctx,edg_wlc_JobId *out)
{
	return edg_wlc_JobIdDup(ctx->p_jobid,out);
}

int edg_wll_GenerateSubjobIds(
	edg_wll_Context	ctx,
	const edg_wlc_JobId	parent,
	int			num_subjobs,
	const char *		seed,
	edg_wlc_JobId **	subjobs)
{
	int subjob, ret;
	char *p_unique, *p_bkserver, *intseed;
	char *unhashed, *hashed;
	unsigned int p_port;
	edg_wlc_JobId *retjobs;


	if (num_subjobs < 1)
		return edg_wll_SetError(ctx, EINVAL,
				"num_subjobs < 1 in edg_wll_GenerateSubjobIds");
	if (seed == NULL)
		intseed = edg_wll_GetSequenceCode(ctx);
	else
		intseed = strdup(seed);

	p_unique = edg_wlc_JobIdGetUnique(parent);
	edg_wlc_JobIdGetServerParts(parent, &p_bkserver, &p_port);

	retjobs = calloc(num_subjobs+1, sizeof(edg_wlc_JobId));

	if (p_unique == NULL ||
		intseed == NULL ||
		p_bkserver == NULL ||
		retjobs == NULL)
		return edg_wll_SetError(ctx, ENOMEM, NULL);

	for (subjob = 0; subjob < num_subjobs; subjob++) {

		asprintf(&unhashed, "%s,%s,%d", p_unique, intseed, subjob);
		if (unhashed == NULL) {
			edg_wll_SetError(ctx, ENOMEM, NULL);
			goto handle_error;
		}

		hashed = str2md5base64(unhashed);
		free(unhashed);
		if (hashed == NULL) {
			edg_wll_SetError(ctx, ENOMEM, NULL);
			goto handle_error;
		}

		ret = edg_wlc_JobIdRecreate(p_bkserver, p_port, hashed, &retjobs[subjob]);
		free(hashed);
		if (ret != 0) {
			edg_wll_SetError(ctx, ret, "in edg_wlc_JobIdRecreate");
			goto handle_error;
		}
	}
	
	free(intseed);
	free(p_unique);
	free(p_bkserver);

	*subjobs = retjobs;
	return 0;

    handle_error:
	free(intseed);
	free(p_unique);
	free(p_bkserver);
	for ( subjob-- ;subjob >= 0; subjob--)
		edg_wlc_JobIdFree(retjobs[subjob]);
	return edg_wll_Error(ctx, NULL, NULL);
}

static void
free_voms_groups(edg_wll_VomsGroups *groups)
{
   size_t len;

   if (groups == NULL)
      return;

   for (len = 0; len < groups->len; len++) {
      if (groups->val[len].vo)
	 free(groups->val[len].vo);
      if (groups->val[len].name)
	 free(groups->val[len].name);
   }
}
