#ident "$Header: /local/reps/lcgware/workload/logging/locallogger/event_queue.c,v 1.24 2003/04/17 12:17:43 michal Exp $"

/* 
 *   - general queue handling routines (insert, get)
 */

#include <netdb.h>
#include <sys/socket.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "edg/workload/logging/locallogger/interlogd.h"
#include "edg/workload/common/jobid/jobid.h"

struct event_queue_msg {
  struct server_msg *msg;
  struct event_queue_msg *prev;
};

struct event_queue *
event_queue_create(char *server_name)
{
  struct event_queue *eq;
  char *p;

  p = strchr(server_name, ':');
  
  if(p) 
    *p++ = 0;

  if((eq = malloc(sizeof(*eq))) == NULL) {
    set_error(IL_NOMEM, ENOMEM, "event_queue_create: error allocating event queue");
    return(NULL);
  }

  memset(eq, 0, sizeof(*eq));

  eq->dest_name = strdup(server_name);

  if(p)
    *(p-1) = ':';
  
  eq->dest_port = p ? atoi(p)+1 : EDG_WLC_JOBID_DEFAULT_PORT+1;

  /* create all necessary locks */
  if(pthread_rwlock_init(&eq->update_lock, NULL)) {
    set_error(IL_SYS, errno, "event_queue_create: error creating update lock");
    free(eq);
    return(NULL);
  }
  if(pthread_mutex_init(&eq->cond_lock, NULL)) {
    set_error(IL_SYS, errno, "event_queue_create: error creating cond mutex");
    free(eq);
    return(NULL);
  }
  if(pthread_cond_init(&eq->ready_cond, NULL)) {
    set_error(IL_SYS, errno, "event_queue_create: error creating cond variable");
    free(eq);
    return(NULL);
  }
  if(pthread_cond_init(&eq->flush_cond, NULL)) {
    set_error(IL_SYS, errno, "event_queue_create: error creating cond variable");
    free(eq);
    return(NULL);
  }

  return(eq);
}


int
event_queue_free(struct event_queue *eq)
{
  assert(eq != NULL);

  if(!event_queue_empty(eq))
    return(-1);

  if(eq->thread_id)
    pthread_cancel(eq->thread_id);


  pthread_rwlock_destroy(&eq->update_lock);
  pthread_mutex_destroy(&eq->cond_lock);
  pthread_cond_destroy(&eq->ready_cond);
  pthread_cond_destroy(&eq->flush_cond);

  free(eq);

  return(0);
}


int
event_queue_empty(struct event_queue *eq)
{
  int ret;

  assert(eq != NULL);

  event_queue_lock_ro(eq);
  ret = eq->head == NULL;
  event_queue_unlock(eq);

  return(ret);
}


int
event_queue_insert(struct event_queue *eq, struct server_msg *msg)
{
  struct event_queue_msg *el, *tail;
  
  assert(eq != NULL);

  if((el = malloc(sizeof(*el))) == NULL) 
    return(set_error(IL_NOMEM, ENOMEM, "event_queue_insert: not enough room for queue element"));

  el->msg = server_msg_copy(msg);
  if(el->msg == NULL) {
    free(el);
    return(-1);
  };

  /* this is critical section */
  event_queue_lock(eq);
  if(server_msg_is_priority(msg)) {
    /* priority messages go first */
    tail = eq->tail_ems;
    if(tail) {
      el->prev = tail->prev;
      tail->prev = el;
      if (tail == eq->tail)
	eq->tail = el;
    } else {
      el->prev = eq->head;
      eq->head = el;
      if(eq->tail == NULL)
	eq->tail = el;
    }
    eq->tail_ems = el;
  } else {
    /* normal messages */
    if(eq->tail)
      eq->tail->prev = el;
    else
      eq->head = el;
    eq->tail = el;
    el->prev = NULL;
  }
  /* if we are inserting message between mark_prev and mark_this,
     we have to adjust mark_prev accordingly */
  if(eq->mark_this && (el->prev == eq->mark_this)) 
    eq->mark_prev = el;
  event_queue_unlock(eq);
  /* end of critical section */

  return(0);
}


int
event_queue_get(struct event_queue *eq, struct server_msg **msg)
{
  struct event_queue_msg *el;

  assert(eq != NULL);
  assert(msg != NULL);
  
  event_queue_lock_ro(eq);
  el = eq->head;
  /* this message is marked for removal, it is first on the queue */
  eq->mark_this = el;
  eq->mark_prev = NULL;
  event_queue_unlock(eq);

  if(el == NULL)
    return(-1);

  *msg = el->msg;

  return(0);
}


int 
event_queue_remove(struct event_queue *eq)
{
  struct event_queue_msg *el, *prev;

  assert(eq != NULL);

  /* this is critical section */
  event_queue_lock(eq);
  el = eq->mark_this;
  prev = eq->mark_prev;

  if(el == NULL) {
    event_queue_unlock(eq);
    return(-1);
  }

  if(prev == NULL) {
    /* removing from head of the queue */
    eq->head = el->prev;
  } else {
    /* removing from middle of the queue */
    prev->prev = el->prev;
  }
  if(el == eq->tail) {
    /* we are removing the last message */
    eq->tail = prev;
  }
  if(el == eq->tail_ems) {
    /* we are removing last priority message */
    eq->tail_ems = prev;
  }
  event_queue_unlock(eq);
  /* end of critical section */
    
  server_msg_free(el->msg);
  free(el);

  return(0);
}

