#ident "$Header: /local/reps/lcgware/workload/logging/locallogger/il_error.c,v 1.3 2003/07/25 07:27:14 michal Exp $"

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <assert.h>
#include <pthread.h>
#include <stdlib.h>

#include <err.h> // SSL header file

#include "edg/workload/logging/locallogger/il_error.h"
#include "edg/workload/logging/common/dgssl.h"


extern int log_level;

static pthread_key_t err_key;

static int IL_ERR_MSG_LEN = 128;

static
void
error_key_delete(void *err)
{
  free(err);
}

static 
void
error_key_create()
{
  pthread_key_create(&err_key, error_key_delete);
}

static 
struct error_inf *
error_get_err ()
{
  struct error_inf *err;

  /* get thread specific error structure */
  err = (struct error_inf *)pthread_getspecific(err_key);
  assert(err != NULL);

  return(err);
}

int
init_errors(int level)
{
  static pthread_once_t error_once = PTHREAD_ONCE_INIT;
  struct error_inf *err;

  /* create the key for thread specific error only once */
  pthread_once(&error_once, error_key_create);
  
  /* there is no thread error yet, try to create one */
  if((err = (struct error_inf *)malloc(sizeof(*err)))) {
    /* allocation successfull, make it thread specific data */
    if(pthread_setspecific(err_key, err)) {
	free(err);
	return(-1);
    }
  } else 
    return(-1);

  err->code_maj = 0;
  err->code_min = 0;
  err->msg = malloc(IL_ERR_MSG_LEN + 1);
  if(err->msg == NULL) 
	  return(-1);

  if(level)
    log_level = level;

  return(0);
}

int 
set_error(int code, long minor, char *msg)
{
  struct error_inf *err;

  err = error_get_err();

  err->code_maj = code;
  err->code_min = minor;

  switch(code) {

  case IL_SYS:
    snprintf(err->msg, IL_ERR_MSG_LEN, "%s: %s", msg, strerror(err->code_min));
    break;

  case IL_HOST:
    snprintf(err->msg, IL_ERR_MSG_LEN, "%s: %s", msg, hstrerror(err->code_min));
    break;

  case IL_AUTH:
    snprintf(err->msg, IL_ERR_MSG_LEN, "%s: %s", msg, ERR_error_string(err->code_min, NULL));
    break;

  case IL_DGSSL:
    switch(err->code_min) {

    case EDG_WLL_SSL_ERROR_SSL:
      snprintf(err->msg, IL_ERR_MSG_LEN, "%s: %s", msg, ERR_error_string(ERR_get_error(), NULL));
      break;

    case EDG_WLL_SSL_ERROR_TIMEOUT:
      snprintf(err->msg, IL_ERR_MSG_LEN, "%s: Timeout in SSL connection.", msg);
      break;

    case EDG_WLL_SSL_ERROR_EOF:
      snprintf(err->msg, IL_ERR_MSG_LEN, "%s: Connection lost.", msg);
      break;

    case EDG_WLL_SSL_ERROR_ERRNO:
      snprintf(err->msg, IL_ERR_MSG_LEN, "%s: %s", msg, strerror(errno));
      break;
      
    case EDG_WLL_SSL_ERROR_HERRNO:
      snprintf(err->msg, IL_ERR_MSG_LEN, "%s: %s", msg, hstrerror(errno));
      break;
    }

  default:
	  strncpy(err->msg, msg, IL_ERR_MSG_LEN);
  }

  return(code);
}


int 
clear_error() {
  struct error_inf *err;

  err = error_get_err();

  err->code_maj = IL_OK;
  err->code_min = 0;
  *(err->msg) = 0;

  return(0);
} 


int
error_get_maj()
{
  struct error_inf *err;

  err = error_get_err();

  return(err->code_maj);
}


long
error_get_min()
{
  struct error_inf *err;

  err = error_get_err();

  return(err->code_min);
}


char * 
error_get_msg()
{
  struct error_inf *err;

  err = error_get_err();

  return(err->msg);
}
