/*
 * File: listjobmatch.cpp
 * Author: Salvatore Monforte <Salvatore.Monforte@ct.infn.it>
 * Author: Marco Pappalardo <Marco.Pappalardo@ct.infn.it>
 * Copyright (c) 2002 EU DataGrid.
 * For license conditions see http://www.eu-datagrid.org/license.html
 */

// $Id: listjobmatch.cpp,v 1.17 2006/02/27 17:38:28 dhsmith Exp $ 

#include <classad_distribution.h>
#include <boost/scoped_ptr.hpp>
#include <boost/filesystem/exception.hpp>
#include "edg/workload/common/configuration/Configuration.h"
#include "edg/workload/common/configuration/CommonConfiguration.h"
#include "edg/workload/planning/broker/RBSimpleImpl.h"
#include "edg/workload/planning/broker/RBMaximizeFilesImpl.h"
#include "edg/workload/planning/broker/RBMinimizeAccessCostImpl.h"
#include "edg/workload/planning/broker/matchmaking/matchmaker.h"
#include "edg/workload/planning/broker/matchmaking/utility.h"
#include "edg/workload/planning/broker/matchmaking/exceptions.h"
#include "edg/workload/networkserver/commands/Command.h"
#include "edg/workload/networkserver/commands/logging.h"
#include "edg/workload/networkserver/common/const.h"
#include "edg/workload/common/utilities/edgstrstream.h"
#include "edg/workload/common/utilities/classad_utils.h"
#include "edg/workload/common/requestad/ManipulationExceptions.h"
#include "edg/workload/common/requestad/JobAdManipulation.h"
#include "edg/workload/common/requestad/JDLAttributes.h"
#include "edg/workload/common/logger/edglog.h"
#include "edg/workload/common/logger/manipulators.h"
#include "edg/workload/common/jobid/JobIdExceptions.h"

using namespace std;

namespace broker      = edg::workload::planning::broker;
namespace matchmaking = broker::matchmaking;
namespace utilities   = edg::workload::common::utilities;
namespace requestad   = edg::workload::common::requestad;
namespace logger      = edg::workload::common::logger;
namespace jobid       = edg::workload::common::jobid;
namespace configuration = edg::workload::common::configuration;

typedef matchmaking::brokerinfo::BrokerInfo<matchmaking::brokerinfo::brokerinfoGlueImpl> BrokerInfo;

namespace edg {
namespace workload {
namespace networkserver {
namespace commands {

bool listjobmatch(Command* cmd)
{
  edglog_fn(CommandFactoryServerImpl()::ListJobMatch());

  std::string request_str;
  cmd->getParam("jdl", request_str);
  boost::scoped_ptr<classad::ClassAd> ad;  

  edglog(medium) << "Received listjobmatch request." << std::endl;
  edglog(info) << request_str << std::endl;

  try 
  {
    ad.reset( utilities::parse_classad(request_str) );
  }
  catch( utilities::CannotParseClassAd& e )
  {
    edglog(medium) << "Cannot Parse classAd: " << request_str << std::endl;
    return false;	  
  }

  const configuration::CommonConfiguration *commonconf;
  commonconf = configuration::Configuration::instance()->common();
  bool use_cached_info_for_gris = commonconf->use_cache_instead_of_gris();

  broker::ResourceBroker rb( new broker::RBSimpleImpl(use_cached_info_for_gris) );
  bool input_data_exists = false;
  std::vector<std::string> input_data;
  requestad::get_input_data(*ad, input_data, input_data_exists);
  boost::scoped_ptr<BrokerInfo> BI(new BrokerInfo);
  boost::shared_ptr<BrokerInfo *> BIp(new BrokerInfo*);
  *BIp = BI.get();
  BrokerInfo::tp_BIp.reset(new boost::shared_ptr<BrokerInfo *>(BIp));

  if(input_data_exists) {

	// Here we have to check if the rank expression in the request
        // is rank = other.dataAccessCost and change the implementation
        // of the broker (RBMinimizeAccessCost)
        classad::ExprTree* rank_expr = ad->Lookup("rank");
        if( rank_expr ) {

            std::vector<std::string> rankAttributes;
            utilities::insertAttributeInVector(&rankAttributes, rank_expr, utilities::is_reference_to("other"));
            if( rankAttributes.size() == 1 &&
                rankAttributes.front()=="DataAccessCost" ) {
		// RBMinimizeAccessCostImpl doesn't rank based on the
		// Info Service. So the use_cached_info_for_gris 
		// flag isn't needed.
                rb.changeImplementation( new broker::RBMinimizeAccessCostImpl( BI.get() ) );
            }
	    else {
	 	rb.changeImplementation( new broker::RBMaximizeFilesImpl( BI.get(), use_cached_info_for_gris ) );
            }
  	}
  }

  boost::scoped_ptr<matchmaking::match_table_t> suitableCEs;
  std::string mm_error("");
  bool error = false;
  try {
    suitableCEs.reset(rb.findSuitableCEs(ad.get()));
  } catch (const std::string& e) {
    mm_error = std::string(e);
    error = true;
  } catch (matchmaking::ISNoResultError const& e) {
    mm_error = std::string( 
     "The user is not authorized on any resource currently registered in "
     + e.host());
    error = true;
  } catch (matchmaking::InformationServiceError const& e) {
    mm_error = std::string(e.what());  
    error = true;
  } catch (matchmaking::RankingError const& e) {
    mm_error = std::string(e.what());
    error = true;
  } catch (requestad::CannotGetAttribute const& e) {
    mm_error = std::string("Attribute " + e.parameter() + 
			   " does not exist or has the wrong type (expected unknown)"); 
    error = true;
  } catch (requestad::CannotSetAttribute const& e) {
    mm_error = std::string("Cannot set attribute " + e.parameter());
    error = true;
  } catch (jobid::JobIdException& jide) {
    edglog( error ) << jide.what() << std::endl;
    mm_error = std::string("Invalid value unknown for attribute " + requestad::JDL::JOBID + 
			   " (expecting valid jobid)");
    error = true;			   
  } catch( boost::filesystem::filesystem_error& fse ) {
    edglog( error ) << fse.what() << std::endl;
    mm_error = std::string(fse.what());
    error = true;
  } catch (std::exception& e) {
    //  catching  (matchmaking::MatchMakingError& e)
    //  catching  (matchmaking::ISConnectionError& e)
    //  catching  (matchmaking::ISQueryError& e)
    //  catching  (matchmaking::ISClusterQueryError& e)
    //  catching  (matchmaking::BadCEIdFormatEx& e)
    edglog( error ) << e.what() << std::endl;
    mm_error = std::string(e.what());
    error = true;
  } catch (...) {
    mm_error = std::string("Unknown exception caught.");
    error = true;
  } 

  if ( !error && suitableCEs.get()==0 ) {
    error = true;
    mm_error = std::string("Matchmaking returned null result. Expression you used may be unsupported.");
  }

  if (error) {
    cmd->setParam("MatchMakingDone", false);
    cmd->setParam("MatchMakingError", mm_error );
    std::vector<std::string> match_list;
    match_list.push_back(EDG_WL_NSMATCHMAKINGERROR);
    match_list.push_back(mm_error);
    cmd->setParam( "MatchResult", match_list );
    return true;
  }

  if (!suitableCEs->empty()) {
	  
    matchmaking::match_vector_t suitableCEs_vector(suitableCEs->begin(), suitableCEs->end());
    std::stable_sort(suitableCEs_vector.begin(), suitableCEs_vector.end(), matchmaking::rank_greater_than_comparator());
    std::vector<std::string> match_list;
    for( matchmaking::match_vector_t::const_iterator it = suitableCEs_vector.begin(); it != suitableCEs_vector.end(); it++) {
      
      utilities::oedgstrstream s;
      s << it -> first << " = " << it -> second.getRank();
      match_list.push_back( s.str() );
    }
    cmd->setParam("MatchResult", match_list);
  }  
  
  edglog(medium) << "ListJobMatch done." << std::endl;
  cmd->setParam("MatchMakingDone", true);
  return true;
}

}}}} // namespace edg::workload::networkserver::commands

