/*
 * File: NS2WMProxy.cpp
 * Author: Marco Pappalardo
 * Copyright (c) 2002 EU DataGrid.
 * For license conditions see http://www.eu-datagrid.org/license.html
 */

// $Id: NS2WMProxy.cpp,v 1.35 2004/11/25 16:08:13 dhsmith Exp $

#include "edg/workload/networkserver/daemon/NS2WMProxy.h"
#include <classad_distribution.h>
#include "edg/workload/common/jobid/JobId.h"
#include "edg/workload/networkserver/commands/Command.h"
#include "edg/workload/networkserver/commands/logging.h"
#include "edg/workload/logging/client/producer.h"
#include "edg/workload/logging/client/context.h"
#include "edg/workload/common/utilities/classad_utils.h"
#include "edg/workload/common/logger/edglog.h"
#include "edg/workload/common/logger/manipulators.h"
#include "edg/workload/common/utilities/edgstrstream.h"
#include "edg/workload/logging/client/producer.h"
#include "edg/workload/networkserver/daemon/logging_fn.h"
#include <boost/shared_ptr.hpp>

#include "edg/workload/common/requestad/JobAdManipulation.h"
#include "edg/workload/common/requestad/PrivateAdManipulation.h"
#include "edg/workload/common/requestad/ManipulationExceptions.h"
#include "edg/workload/common/configuration/Configuration.h"
#include "edg/workload/common/configuration/WMConfiguration.h"
#include <cstdlib>

#define SEQFILENAME ".edg_wll_seq"
#define QUOTAFILE ".quid"

namespace common     = edg::workload::common;
namespace jobid      = common::jobid;
namespace utilities  = common::utilities;
namespace logger     = common::logger;
namespace commands   = edg::workload::networkserver::commands;

namespace {

void f_forward(utilities::FileList<std::string>& filelist,
               utilities::FileListMutex& mutex,
               std::string const& ad)
{
  utilities::FileListLock lock(mutex);
  filelist.push_back(ad);
}

}

namespace edg {
namespace workload {
namespace networkserver {
namespace daemon {

NS2WMProxy::NS2WMProxy()
{
}

NS2WMProxy::~NS2WMProxy()
{
}

void NS2WMProxy::init(const std::string &filename)
{
  edglog_fn("NS2WM::init");
  edglog(fatal) << "Initializing NS2WM Proxy..." << std::endl;
  edglog(fatal) << "FileQueue is: " << filename << std::endl;

  m_filelist.reset(new utilities::FileList<std::string>(filename));
  m_mutex.reset(new utilities::FileListMutex(*(m_filelist.get())));

  edglog(fatal) << "NS2WMProxy Initialization Done." << std::endl;
}

void NS2WMProxy::storeQuotaId (classad::ClassAd* cmdAd)
{
  edglog_fn("NS2WM::stQuotaId");
  edglog(null) << "Storing QuotaId for Dynamic Quota Management." << std::endl;

  std::string quid;
  try {
   if (!utilities::evaluate_expression(*cmdAd, "Arguments.Uid", quid) ) {
     return;
   }
 
   std::string quid_path(utilities::evaluate_expression(*cmdAd, "Arguments.JobPath"));
   std::string quid_file(quid_path + "/" + QUOTAFILE);

   edglog(debug) << "Quota Uid: " << quid << std::endl;
   edglog(info)  << "Quota file: " << quid_file << std::endl;
      
   std::ofstream ostr( quid_file.c_str());
   if ( ostr ) {	
     ostr << quid << std::endl;
     ostr.close();
   } else {
     edglog(error) << "Error during quota file creation." << std::endl;
   } 
  }
  catch(utilities::InvalidValue& e)
  {
    edglog(fatal) << "Job Path not found." << quid << std::endl;
  }
  catch ( std::exception& e ) {
    edglog(fatal) << "Exception during quota id storaging. " << e.what() << std::endl;
  }
}


void NS2WMProxy::storeSequenceCode (classad::ClassAd* cmdAd, edg_wll_Context* log_ctx)
{
  edglog_fn("NS2WM::stSeqCode");
  edglog(fatal) << "Storing Seq_Code." << std::endl;

  std::string seq_code;
  try {
   if( log_ctx ) {
     char* seqstr = edg_wll_GetSequenceCode(*log_ctx);
     seq_code.assign( std::string(seqstr) );
     free( seqstr );
   }
   else seq_code.assign( utilities::evaluate_expression(*cmdAd, "Arguments.SeqCode") );
    
   std::string seq_path(utilities::evaluate_expression(*cmdAd, "Arguments.JobPath"));
   std::string seq_file(seq_path + "/" + SEQFILENAME);
      
   edglog(info) << "Sequence Code: " << seq_code << std::endl;
   edglog(info) << "Sequence Code file: "  << seq_file << std::endl;
      
   std::ofstream ostr( seq_file.c_str());
   if ( ostr ) {	
     ostr << seq_code << std::endl;
     ostr.close();
   } 
   else {
     edglog(info) << "Error during sequence file creation." << std::endl;
   } 
  }
  catch(utilities::InvalidValue& e)
  {
    edglog(fatal) << "Job path not found."<< seq_code << std::endl;
  }
  catch ( std::exception& e ) {
    edglog(fatal) << "Exception during sequence code storaging. " << e.what() << std::endl;
  }
}

std::string NS2WMProxy::convertProtocol(classad::ClassAd* cmdAd) 
{
  edglog_fn("NS2WM::convertProtocol");
  edglog(fatal) << "Converting to common protocol." << std::endl;

  std::string convertedString;
  std::string name( utilities::evaluate_attribute(*cmdAd, "Command") );
  if (name == "JobSubmit") {
    /* This is the output we want to forward to WM
     * for submit case.
     * [
     *   version = "1.0.0";
     *   command = "jobsubmit";
     *   arguments = [
     *     ad = [
     *        type = \"job\";
     *        executable = \"ls\";
     *        dg_jobid=\"abcakjflkjdlksjf\";
     *        lb_sequence_code=\".....\"
     *        inputsandboxpath="...";
     *        outputsandboxpath="...";
     *        x509userproxy="road to user proxy file.";
     *     ]
     *   ]
     * ]
     */
    std::string jdl      (utilities::evaluate_expression(*cmdAd, "Arguments.jdl"));
    std::string isb      (utilities::evaluate_expression(*cmdAd, "Arguments.InputSandboxPath"));
    std::string osb      (utilities::evaluate_expression(*cmdAd, "Arguments.OutputSandboxPath"));
    std::string proxy    (utilities::evaluate_expression(*cmdAd, "Arguments.X509UserProxy"));
    std::string seq_code (utilities::evaluate_expression(*cmdAd, "Arguments.SeqCode"));
    
    classad::ClassAd internal, total;
    classad::ClassAd* jdlAd(utilities::parse_classad(jdl)); 
    internal.Insert( "ad", jdlAd );
    
    jdlAd->InsertAttr("lb_sequence_code",  seq_code); 
    jdlAd->InsertAttr("InputSandboxPath",  isb );
    jdlAd->InsertAttr("OutputSandboxPath", osb);
    jdlAd->InsertAttr("X509UserProxy",     proxy);
   
    // edglog(null) << "Classad modificato: " << *ad  << std::endl;

    // normalization of Input Sandbox
    std::vector<std::string> ISB;
    std::vector<std::string> ISB_normalized;

    utilities::EvaluateAttrListOrSingle( *jdlAd, "InputSandBox", ISB );
    for( std::vector<std::string>::const_iterator it = ISB.begin(); it != ISB.end(); it++) {
	// TODO: use boost::filesystem here
	// it should be worth to do just: 
	// fs::path p(*it, fs::system_specific);
	// ISB_normalized( push_back(p.leaf()) );
	std::string full_path_2_file( *it );
	size_t slash_pos = full_path_2_file.rfind("/");
	std::string filename( full_path_2_file.substr(slash_pos + 1, full_path_2_file.length() - slash_pos -1 ) );	
	ISB_normalized.push_back( filename );	
    }
    utilities::InsertAttrList( *jdlAd, "InputSandBox", ISB_normalized);
    total.InsertAttr( "command", std::string("jobsubmit") );
    total.InsertAttr( "version", std::string("1.0.0") );
    total.Insert( "arguments", internal.Copy() );
    convertedString.assign( utilities::unparse_classad(total) );

  } else if (name == "JobCancel") {
    /* This is the output we want to forward to WM
     * for cancel case.
     * [ 
     *   version = "1.0.0";
     *   command = "jobcancel";
     *   arguments = [
     *     dg_jobid="abcakjflkjdlksjf";
     *     lb_sequence_code=\".....\"
     *   ] 
     * ] 
     */
    std::string jobid   (utilities::evaluate_expression(*cmdAd, "Arguments.jobid"));
    std::string seq_code(utilities::evaluate_expression(*cmdAd, "Arguments.SeqCode")); 
    
    utilities::edgstrstream strstr;
    strstr << "[ version = \"1.0.0\"; command = \"jobcancel\"; arguments = [ id = \"" 
	   << jobid << "\"; lb_sequence_code = \"" << seq_code << "\" ] ]";
    
    convertedString.assign( std::string(strstr.str()) );

  } else {
    edglog(fatal) << "Error: Unforwardable command! Command is: " << name << std::endl;
  }

  edglog(fatal) << "Converted String: " << convertedString << std::endl;
  return convertedString;
}

void NS2WMProxy::submit(classad::ClassAd* cmdAd)
{
  edglog_fn("NS2WM::submit");
  edglog(fatal) << "Forwarding Submit Request." << std::endl;
  
  std::string seq_code(utilities::evaluate_expression(*cmdAd, "Arguments.SeqCode"));
  std::string jdl     (utilities::evaluate_expression(*cmdAd, "Arguments.jdl"));
  boost::scoped_ptr<classad::ClassAd> jdlAd(utilities::parse_classad(jdl));
  std::string jobid   (utilities::evaluate_attribute(*jdlAd, "edg_jobid"));
  //
  // Initialize the logging context
  //
  edg_wll_Context      log_ctx;
  edg_wlc_JobId        wlc_jid; 
  if( edg_wll_InitContext( &log_ctx ) != 0 ||
      edg_wlc_JobIdParse ( jobid.c_str(), &wlc_jid ) != 0 ||  
      edg_wll_SetParam( log_ctx, EDG_WLL_PARAM_SOURCE,  EDG_WLL_SOURCE_NETWORK_SERVER  ) != 0 ||
      edg_wll_SetLoggingJob( log_ctx, wlc_jid, seq_code.c_str(), EDG_WLL_SEQ_NORMAL ) != 0 )  {
	  char      *et,*ed;
          edg_wll_Error(log_ctx,&et,&ed);
          edglog(fatal) << " unable to initialize logging context " << et << ": " << ed;
          free(et); free(ed);
  }
  char* seq_str = edg_wll_GetSequenceCode(log_ctx);
  ((classad::ClassAd*)(cmdAd->Lookup("Arguments"))) -> InsertAttr("SeqCode", std::string(seq_str));
  free(seq_str);

  edglog(debug) << "start: " << jdl << std::endl;
  std::string proxy (utilities::evaluate_expression(*cmdAd, "Arguments.X509UserProxy"));
  std::string command_str( convertProtocol(cmdAd) );
  {
      boost::scoped_ptr<classad::ClassAd> command_ad(utilities::parse_classad(command_str));
      classad::ClassAd* ptr = ((classad::ClassAd*) (((classad::ClassAd*)(command_ad->Lookup("Arguments"))) -> Lookup ("ad")));
      // add sync running event flag, to allow for shallow resubmissions
      bool sync_running_event = true;
      int max_shallow_count = get_max_shallow_count();
      int max_retry_count = get_max_retry_count();
      bool good;
      int job_retry_count = common::requestad::get_retry_count(*ptr, good);
  
      /* Use sync logging for LRMS running event only if there is a possibility of shallow resubmit */
      if (max_shallow_count == 0 || max_retry_count<0 || (good && job_retry_count<0)) {
        sync_running_event = false;
      }

      /* sync logging only supported for job type 'normal' */
      std::string jobtype(common::requestad::get_job_type(*ptr,good));
      if (good) {
        std::string ljobtype(jobtype);
        std::transform(ljobtype.begin(), ljobtype.end(), ljobtype.begin(), ::tolower);
        if (ljobtype != "normal") {
          sync_running_event = false;
        }
      } else {  
        sync_running_event = false;
      }
      common::requestad::set_edg_sync_running_event(*ptr, sync_running_event, good);

      command_str.assign(utilities::unparse_classad(*command_ad));
      jdl.assign(utilities::unparse_classad(*ptr));
  }

  edglog(debug) << "Logged jdl: " << jdl << std::endl;

  edg_wlc_JobIdFree(wlc_jid);                                                      

  try {
    f_forward(*(m_filelist.get()), *(m_mutex.get()), command_str);
    // Take the result of f_forward and do what for LogEnQueued in CommandFactoryServerImpl
    /*
    edg_wll_LogEnQueued( log_ctx,
			(*m_filelist).filename().c_str(),
			jdl.c_str(),
			"OK",
		        "");
    */
    LogEnqueuedJob( log_ctx, jdl, proxy, (*m_filelist).filename().c_str(), true, "", true, true);
    edglog(null) << "Submit EnQueued OK." << std::endl; 
  } catch (std::exception &e) {
    // LogEnQueued FAIL if exception occurs
    /*
    edg_wll_LogEnQueued( log_ctx,
			(*m_filelist).filename().c_str(),
                        jdl.c_str(),
                        "FAIL",
                        e.what());
    */
    LogEnqueuedJob( log_ctx, jdl, proxy, (*m_filelist).filename().c_str(), false, e.what(), true, true);
    edglog(null) << "Submit EnQueued FAIL." << std::endl; 
  }

  // Save the sequence code on an hidden file.
  storeSequenceCode(cmdAd, &log_ctx);

  // Free the context
  edg_wll_FreeContext(log_ctx);
  edglog(null) << "Submit Forwarded." << std::endl; 
}

void NS2WMProxy::cancel(classad::ClassAd* cmdAd)
{
  edglog_fn("NS2WM::cancel");
  edglog(fatal) << "Forwarding Cancel Request." << std::endl;

  std::string command_ad(convertProtocol(cmdAd));
  f_forward(*(m_filelist.get()), *(m_mutex.get()), command_ad);

  // Save the sequence code on an hidden file.
  storeSequenceCode(cmdAd, 0);

  edglog(null) << "Cancel Forwarded." << std::endl;
}

int NS2WMProxy::get_max_retry_count()
{   
  common::configuration::Configuration const* const config
    = common::configuration::Configuration::instance();

  if (!config) {
    edglog(fatal) << "empty or invalid configuration" << std::endl;
    std::abort();
  }
  common::configuration::WMConfiguration const* const wm_config = config->wm();
  if (!wm_config) {
    edglog(fatal) << "empty WM configuration" << std::endl;
    std::abort();
  }

  return wm_config->max_retry_count();
}

int NS2WMProxy::get_max_shallow_count()
{
  common::configuration::Configuration const* const config
    = common::configuration::Configuration::instance();

  if (!config) {
    edglog(fatal) << "empty or invalid configuration" << std::endl;
    std::abort();
  }
  common::configuration::WMConfiguration const* const wm_config = config->wm();
  if (!wm_config) {
    edglog(fatal) << "empty WM configuration" << std::endl;
    std::abort();
  }
   
  return wm_config->max_shallow_count();
}


} // namespace daemon closure
} // namespace networkserver closure
} // namespace workload closure
} // namespace edg closure
