/*
 * File: NetworkServerReal.cpp
 * Author: Monforte Salvatore <Salvatore.Monforte@ct.infn.it>
 * Author: Marco Pappalardo <marco.pappalardo@ct.infn.it>
 * Copyright (c) 2001 EU DataGrid.
 * For license conditions see http://www.eu-datagrid.org/license.html
 */

// $Id: NetworkServerReal.cpp,v 1.39 2005/10/12 12:34:37 dhsmith Exp $

#include <boost/lexical_cast.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/scoped_ptr.hpp>
#include <classad_distribution.h>

#include "edg/workload/networkserver/daemon/NetworkServerReal.h"
#include "edg/workload/networkserver/daemon/NS2WMProxy.h"
#include "edg/workload/networkserver/daemon/NetworkServer.h"
#include "edg/workload/networkserver/daemon/exceptions.h"
#include "edg/workload/networkserver/commands/Command.h"
#include "edg/workload/networkserver/commands/CommandFactory.h"
#include "edg/workload/networkserver/commands/CommandFactoryServerImpl.h"
#include "edg/workload/networkserver/commands/logging.h"
#include "edg/workload/common/task/Task.h"
#include "edg/workload/common/utilities/Exceptions.h"
#include "edg/workload/common/socket++/GSISocketAgent.h"
#include "edg/workload/common/socket++/GSISocketServer.h"
#include "edg/workload/common/socket++/exceptions.h"
#include "edg/workload/common/configuration/Configuration.h"
#include "edg/workload/common/configuration/ModuleType.h"
#include "edg/workload/common/configuration/exceptions.h"
#include "edg/workload/common/configuration/CommonConfiguration.h"
#include "edg/workload/common/configuration/NSConfiguration.h"
#include "edg/workload/common/configuration/WMConfiguration.h"
#include "edg/workload/common/process/process.h"
#include "edg/workload/common/process/user.h"
#include "edg/workload/common/utilities/classad_utils.h"
#include "edg/workload/common/logger/edglog.h" 
#include "edg/workload/common/logger/manipulators.h" 
 
namespace common        = edg::workload::common;
namespace task          = common::task;
namespace socket_pp     = common::socket_pp;
namespace process       = common::process;
namespace configuration = common::configuration;
namespace logger        = common::logger; 
namespace utilties	= common::utilities;
namespace commands      = edg::workload::networkserver::commands;

namespace edg {
namespace workload {
namespace networkserver {
namespace daemon {

class Listener: public task::PipeWriter< socket_pp::GSISocketAgent* >
{
public:
  virtual void run(void)
  {
    socket_pp::GSISocketAgent* agent = 0;
    edglog_fn("Listener::run");
    try {
      while(true) {
        try {
	   agent =  UIserver -> Listen();
           if( agent ) {
             edglog(debug) << "Handling incoming connection on socket #" << agent->SocketDescriptor() << std::endl;
             write_end().write( agent );
           }
        } 
	catch( edg::workload::common::utilities::Exception& e) {
	  edglog(fatal) << "Exception Caught: " << e.what() << std::endl;
	}
      }	
    }
    catch (task::SigPipe&) {
      if( agent ) UIserver -> KillAgent( agent );
      edglog(fatal) << "Pipe Closed." << std::endl;
    }
  }
  static GSISocketServerPtr UIserver;
};

GSISocketServerPtr Listener::UIserver;

class Manager: public task::PipeForwarder< socket_pp::GSISocketAgent*, classad::ClassAd* >
{

public:
  virtual void run(void)
  {
    commands::Command *cmd=NULL;
    edglog_fn("Manager::run");
    try {
      while (true) {
	      socket_pp::GSISocketAgent* agent = read_end().read();
              std::string from_host( agent -> HostName() );

              // Setting Timeout to address 
              // LCG Bug #2716
              agent -> SetTimeout(configuration::Configuration::instance() -> ns() -> connection_timeout());
              edglog(info) << "Connection from: " << from_host << std::endl;

              try { 
                if (!UIserver -> AuthenticateAgent(agent)) {
                  edglog(warning) << "Authentication with host " << from_host << " failed" << std::endl;
                  continue;
                } else {
                  std::string cs(agent -> CertificateSubject() );
                  std::string gn(agent -> GridmapName() );
                  edglog(info) << "Authentication with host " << from_host << " succeeded for " << cs << " mapped to " << gn << std::endl;
                }
              } catch( edg::workload::common::socket_pp::Exception& e) {
                edglog(fatal) << "Exception Caught during client authentication: " << e.what() << std::endl;
                UIserver -> KillAgent(agent);
                continue; 
              } catch (...) {
                edglog(fatal) << "Exception Caught during client authentication" << std::endl;
                UIserver -> KillAgent(agent);
                continue; 
              }

              // Resetting Timeouts
              agent -> SetTimeout(configuration::Configuration::instance() -> ns() -> command_timeout());

              std::string cmdname;
              try {
                  agent -> Receive(cmdname);
              }
	      catch( edg::workload::common::socket_pp::Exception& e) {
                  edglog(fatal) << "While trying to read command: " << e.what() << std::endl;
                  UIserver -> KillAgent(agent);
                  continue; 
              }
              catch(...) {
	          edglog(fatal) << "Caught unhandled exception while trying to read command." << std::endl;
                  UIserver -> KillAgent(agent);
                  continue; 
              }

	      commands::CommandFactory<commands::CommandFactoryServerImpl> factory;
	      bool is_forwarded = false;
	      try {
		  cmd =  factory.create(cmdname);
		  cmd -> serialize( agent );
		  assert( !cmd -> isDone() );
		  do {
			  
		    if( cmd -> state().forwardRequest() ) {
                            static boost::mutex forwardIdMtx;
                            static unsigned long int nextForwardId=1;
                            std::string thisForwardId;

			    is_forwarded = true;
		    	    char* seq_str = edg_wll_GetSequenceCode(*cmd->getLogContext());
			    std::string seq_code(seq_str);		    
			    free( seq_str );
			    cmd->setParam("SeqCode", seq_code);	    

                            {
                              boost::mutex::scoped_lock lk(forwardIdMtx);
                              thisForwardId = boost::lexical_cast<std::string>(nextForwardId++);
                            }

                            std::auto_ptr< classad::ClassAd > cmdAd( cmd->asClassAd().Copy() );

                            cmdAd->InsertAttr("ForwardId", thisForwardId);

			    write_end().write( cmdAd.get() );
                            cmdAd.release();
			    edglog(fatal) << "Command #" << thisForwardId << " Forwarded." << std::endl;
		    }
		    
		  } while( cmd -> execute() && !cmd -> isDone() );
	      } 
	      catch (commands::bad&) {
	        edglog(fatal) << "Bad Command." << std::endl;
	      }
	      catch( edg::workload::common::socket_pp::Exception& e) {
	      	edglog(fatal) << "Exception Caught during client communication via socket: " 
		       << e.what() << std::endl;
		// Log abort so that the job wont remain in waiting status
		// if and only if the job has not been forwarded yet.
	      	edglog(fatal) << "Logging Aborted" << std::endl;
		if( !is_forwarded ) {
		  int retry = 10; // this should be taken from the configuration file...
		  while( edg_wll_LogAbort(*(cmd->getLogContext()), e.what()) && retry-- );
	        }
	      }
	      catch( edg::workload::common::utilities::Exception& e) {
		edglog(fatal) << "Exception Caught: " << e.what() << std::endl;
	      }
	      UIserver -> KillAgent(agent);
	      if (cmd) {
		delete cmd;
		cmd = NULL;
              }
              edglog(info) << "Command done" << std::endl;
      }
    } 
    catch (task::Eof&) {
      edglog(fatal) << "Pipe Closed for reading." << std::endl;
    } 
    catch (task::SigPipe&) {
      edglog(fatal) << "Pipe Closed for writing." << std::endl;
    } 
    catch (commands::bad&) {
      edglog(fatal) << "Bad Command." << std::endl;
    } 
    catch( socket_pp::Exception& e) {
      edglog(fatal) << "Manager: " << e.what() << std::endl;
    } 
    catch( std::exception& e ) {
      edglog(fatal) << "Manager: " << e.what() << std::endl;
    } 
    catch (...) {
      edglog(fatal) << "Uncaught Exception: plese check." << std::endl;
    } 
  }
  static GSISocketServerPtr UIserver;
};

GSISocketServerPtr Manager::UIserver;

class Dispatcher: public task::PipeReader< classad::ClassAd* >
{
public:
  virtual void run(void)
  {
    try {
      edglog_fn("Dispatcher::run");
      while (true) {
	try {
          boost::scoped_ptr< classad::ClassAd > cmdAd( read_end().read() );
	  std::string cmdName;
	  try {
	    cmdName.assign(utilities::evaluate_attribute(*cmdAd, "Command"));
            std::string reqId;
            if (utilities::evaluate_attribute(*cmdAd, "ForwardId", reqId)) {
              cmdAd -> Delete("ForwardId");
  	      edglog(critical) << "Command #" << reqId << " to dispatch: " << cmdName << std::endl;
            } else {
  	      edglog(critical) << "Command to dispatch: " << cmdName << std::endl;
            }
	  }		
	  catch(utilities::InvalidValue &e) {
	    edglog(fatal) << "Missing Command name." << std::endl;
	  }
	  
	  if ( cmdName == "JobSubmit") { 
	    singleton_default<NS2WMProxy>::instance().submit(cmdAd.get());
	  }
	  else if (cmdName == "JobCancel") {
	    singleton_default<NS2WMProxy>::instance().cancel(cmdAd.get());
	  }
	  else {
	    edglog(fatal) << "No forwarding procedure defined for this command." << std::endl;
	  }
	} 
	catch(utilities::Exception& e) {
	  edglog(fatal) << "Exception Caught:" << e.what() << std::endl;
	} 
	catch( std::exception& ex ) {
	  edglog(fatal) << "Exception Caught:" << ex.what() << std::endl;
	}
      }
    } catch (task::Eof&) {
      edglog(fatal) << "Pipe Closed." << std::endl;
    }
  }
};

NetworkServerReal::NetworkServerReal()
{
}
 
NetworkServerReal::~NetworkServerReal()
{
}

void NetworkServerReal::init()
{ 
  edglog_fn("  NSR::init");

  try {
#ifndef DISABLE_DROP_PRIVILEGES
    drop_privileges();
#endif
#ifndef DISABLE_MAKE_DAEMON
    process::Process::self()->make_daemon();
#endif
    NSconf = configuration::Configuration::instance() -> ns();

    UIserver.reset( new socket_pp::GSISocketServer(NSconf->listening_port(), NSconf->backlog_size() ));
    UIserver -> Open();
    edglog(fatal) << "Listening on port: " << NSconf->listening_port() << std::endl;
    
    Listener::UIserver = UIserver;
    Manager::UIserver  = UIserver;

  } catch (CannotStartException e) {
    edglog(fatal) << "CannotStart Daemon: " << e.what() << std::endl;
    std::cout << "CannotStart Daemon: " << e.what() << std::endl;
    exit (-1);
  }
  catch( configuration::ModuleType &type ) {
    edglog(fatal) << "Unknown module type: " << type.get_stringtype() << std::endl;
    std::cout << "Unknown module type: " << type.get_stringtype() << std::endl;
    throw;
  }
  catch( configuration::InvalidExpression &error ) {
    std::cout << "Invalid expression: " << error << std::endl;
    edglog(fatal) << "Invalid expression: " << error << std::endl;
    throw;
  }
  catch( configuration::CannotOpenFile &file ) {
    edglog(fatal) << "Cannot open file: " << file << std::endl;
    std::cout << "Cannot open file: " << file << std::endl;
    throw;
  }
  catch( configuration::CannotConfigure &error ) {
    edglog(fatal) << "Cannot configure: " << error << std::endl;
    std::cout << "Cannot configure: " << error << std::endl;
    throw;
  }
  catch( socket_pp::IOException& e) {
    edglog(fatal) << "IOException caught: " << e.what() << std::endl;
    std::cout << "IOException caught: " << e.what() << std::endl;
    throw;
  } 	  
}

void NetworkServerReal::run()
{ 
  edglog_fn("NSR::run");
  if(UIserver.get()) {

    task::Pipe< socket_pp::GSISocketAgent*   > p1;
    task::Pipe< classad::ClassAd*            > p2;
    
    Listener   listener;
    Manager    manager;
    Dispatcher dispatcher;
    
    
    task::Task listeners   (listener,   p1,     1);
    
    try {
      task::Task managers    (manager, p1, p2, NSconf->master_threads());
      task::Task dispatchers (dispatcher, p2, NSconf->dispatcher_threads());
      boost::mutex::scoped_lock lk( mtx_shutdown );
      do_shutdown.wait(lk);
      exit(0);
    } catch ( configuration::InvalidExpression& error ) {
      edglog(fatal) << "Invalid Expression: " << error << std::endl;
    }
  }
  exit(-1);
}

void NetworkServerReal::shutdown()
{
  do_shutdown.notify_one();
  edglog_fn("NSR::shutdown");
  edglog(fatal) << "Network Server Shutdown..." << std::endl;
}

void NetworkServerReal::drop_privileges()
{
  edglog_fn("  NSR::drop");
  process::User currentUser;
  const configuration::CommonConfiguration *commonconf = configuration::Configuration::instance()->common(); 
  
  if( currentUser.uid() == 0 ) { 
    edglog(fatal) << "Running as root, trying to lower permissions..." << std::endl;
 
    if( process::Process::self()->drop_privileges(commonconf->dguser().c_str()) ) {
       edglog(fatal) << "Failed: reason \"" << strerror(errno) << "\"" << std::endl;
       throw CannotStartException( "Cannot drop privileges, avoiding running as root." );
    }
  
    edglog(fatal) << "Running as user " << commonconf->dguser() << "..." << std::endl;
  } 
  else edglog(fatal) << "Already running in an unprivileged account..." << std::endl;
}


const configuration::NSConfiguration& NetworkServerReal::configuration()
{
 return *NSconf;
}

} // namespace daemon 
} // namespace networkserver
} // namespace workload
} // namespace edg
