/*
 * File: main.cpp
 * Author: Monforte Salvatore <Salvatore.Monforte@ct.infn.it>
 * Author: Marco Pappalardo <Marco.Pappalardo@ct.infn.it>
 * Copyright (c) 2001 EU DataGrid.
 * For license conditions see http://www.eu-datagrid.org/license.html
 */
 
#include "edg/workload/networkserver/daemon/NetworkServer.h"
#include "edg/workload/networkserver/daemon/NS2WMProxy.h"
#include "edg/workload/networkserver/commands/logging.h"
#include "edg/workload/common/configuration/Configuration.h"
#include "edg/workload/common/configuration/WMConfiguration.h"
#include "edg/workload/common/configuration/ModuleType.h"
#include "edg/workload/common/logger/edglog.h"
#include "edg/workload/common/logger/manipulators.h"
#include "edg/workload/common/ssl_helpers/ssl_inits.h"
#include "edg/workload/common/ssl_helpers/ssl_pthreads.h"
#include "edg/workload/common/utilities/LineParser.h"
#include "edg/workload/common/utilities/LineParserExceptions.h"
#include "edg/workload/common/utilities/classad_utils.h"
#include "edg/workload/planning/plugin/edg-classad-plugin-loader.h"

#include <globus_common.h>

#include <boost/scoped_ptr.hpp>

using namespace edg::workload::networkserver::daemon;
using namespace edg::workload::common::utilities;
namespace configuration = edg::workload::common::configuration;
namespace logger = edg::workload::common::logger;

LineOption  options[] = {
	{ 'c', 1, 	    "conf_file", "\t use conf_file as configuration file." },
	{ 'd', no_argument, "daemonize", "\t run as daemon." },
};

int main(int argc, char* argv[])
{
  std::vector<LineOption>        optvec( options, options + sizeof(options)/sizeof(LineOption) );
  LineParser                     options( optvec, 0 );
  std::string conf_file;	    
  
  try {
  
	  options.parse( argc, argv );
      	  conf_file.assign( options.is_present('c') ? options['c'].getStringValue() : "edg_wl.conf" );
          std::auto_ptr<configuration::Configuration> conf;
          conf.reset(new configuration::Configuration( conf_file, "NetworkServer"));

	  logger::threadsafe::edglog.open( 
	        configuration::Configuration::instance()->ns()->log_file(),
     		static_cast<logger::level_t>(configuration::Configuration::instance()->ns()->log_level()) );

	  edglog_fn("   NS::main");
	  edglog(fatal) << "--------------------------------------" << std::endl;
	  edglog(fatal) << "Starting Network Server..." << std::endl;
	
	  logger::threadsafe::edglog.activate_log_rotation (
		configuration::Configuration::instance()->ns()->log_file_max_size(),
		configuration::Configuration::instance()->ns()->log_rotation_base_file(),
	        configuration::Configuration::instance()->ns()->log_rotation_max_file_number() );

	  if (globus_module_activate(GLOBUS_COMMON_MODULE) != GLOBUS_SUCCESS)   {
	    edglog_fn("NS::main");
	    edglog(fatal) << "Failed to initialize Globus common module" << std::endl;
	    exit( -1 );
	  }
       
	  if( edg_wlc_SSLInitialization() != 0) {
	    edglog(fatal) << "Failed to initialize SSL" << std::endl;
	    exit( -2 );	     	
	  }

	  if (edg_wlc_SSLLockingInit() != 0) {
	    edglog(fatal) << "Failed to initialize SSL Locking" << std::endl;
	    exit( -3 );
	  }	
	
	  singleton_default<NetworkServer>::instance().init();
	  singleton_default<NS2WMProxy>::instance().init(configuration::Configuration::instance() -> wm() -> input() );
	  edglog(fatal) << "Initialization Done. Running Server..." << std::endl;
	  edglog(fatal) << "--------------------------------------" << std::endl;
	  singleton_default<NetworkServer>::instance().run();
  } 
  catch( LineParsingError &error ) {
    std::cerr << error << std::endl;
    exit( error.return_code() );
  }
  catch(std::exception& e) {
    edglog(fatal) << "Exception caught: " << e.what() << std::endl;
  } 
  catch( ... ) {
    edglog(fatal) << "Uncaught exception...." << std::endl;	 
  }
  
}
