// File: RBSimpleImpl.cpp
// Author: Salvatore Monforte <Salvatore.Monforte@ct.infn.it>
// Copyright (c) 2002 EU DataGrid.
// For license conditions see http://www.eu-datagrid.org/license.html

// $Id: RBMaximizeFilesImpl.cpp,v 1.15 2004/11/06 13:33:34 dhsmith Exp $

#include <algorithm>
#include <functional>
#include <numeric>
#include <vector> 

#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/tuple/tuple.hpp>

#include "edg/workload/common/requestad/ManipulationExceptions.h"
#include "edg/workload/common/requestad/PrivateAdManipulation.h"
#include "edg/workload/common/requestad/JobAdManipulation.h"

#include "edg/workload/planning/broker/matchmaking/matchmakerGlueImpl.h"
#include "edg/workload/planning/broker/matchmaking/glue_attributes.h"
#include "edg/workload/planning/broker/matchmaking/jdl_attributes.h"
#include "edg/workload/planning/broker/RBMaximizeFilesImpl.h"
#include "edg/workload/planning/broker/utility.h"

#include "edg/workload/common/logger/edglog.h"
#include "edg/workload/common/logger/manipulators.h"

#define edglog(level) logger::threadsafe::edglog << logger::setlevel(logger::level)
#define edglog_fn(name) logger::StatePusher    pusher(logger::threadsafe::edglog, #name);

using namespace std;

namespace edg {
namespace workload {

namespace requestad     = common::requestad;
namespace logger        = common::logger;

namespace planning {
namespace broker {

using namespace matchmaking;
using namespace brokerinfo;
	
RBMaximizeFilesImpl::RBMaximizeFilesImpl(matchmaking::brokerinfo::BrokerInfo<matchmaking::brokerinfo::brokerinfoGlueImpl> *bi, bool do_prefetch)
{
	BI = bi;
	m_prefetch = do_prefetch;
}
 
RBMaximizeFilesImpl::~RBMaximizeFilesImpl()
{
}

match_table_t* RBMaximizeFilesImpl::findSuitableCEs(const classad::ClassAd* requestAd)
{
  edglog_fn(RBMaximizeFilesImpl::findSuitableCEs);
  bool retry_previous_destinations_disallowed = false;
  bool good;

  if (requestAd &&
      requestad::get_edg_retry_previous_destinations_disallowed(*requestAd, good) && good) {
    retry_previous_destinations_disallowed = true;
  }

  match_table_t* suitableCEs = 0;
  if( requestAd ) { 
	
      	// retrieves all the CEs which satisfy user requiremets...	
	MatchMaker<matchmakerGlueImpl> MM;
  	suitableCEs = new match_table_t;
        if (m_prefetch) {
          MM.prefetchCEInfo(requestAd, *suitableCEs);
        }

  	MM.checkRequirement(requestAd, *suitableCEs, m_prefetch);
	
	// Now we have to filter this CEs to check whether they 
	// satisfy data requirements, or not.
	
	// Instantiates a BI object and collects all the information
	// about SFNs and involved SEs.
	// NOTE: this information is not CE dependent.
	BI->retrieveSFNsInfo(*requestAd);
	BI->retrieveSEsInfo (*requestAd);
	
	vector<string> deletingCEs;
	vector<string> data_access_protocols;
	requestad::get_data_access_protocol(*requestAd, data_access_protocols);
	map<string, size_t > CEs2nFiles;
	
	for(match_const_iterator ce = suitableCEs->begin();
	    ce != suitableCEs->end(); ce++) {
		
		BI->retrieveCloseSEsInfo(ce->first);
		BrokerInfoData::SE_container_type compatibleCloseSEs( (*BI)->getCompatibleCloseSEs(data_access_protocols) );
		if(compatibleCloseSEs.empty()) {
		        
			edglog( debug ) << ce->first << " has no compatible CloseSE..." << endl;		
			deletingCEs.push_back(ce->first);	
		}
		else {
			// Now we have to count the number of files the CloseSEs supply the CE with.
			BrokerInfoData::LFN_container_type providedLFNs( (*BI)->getProvidedLFNs(compatibleCloseSEs) );
			size_t n = providedLFNs.size();
			edglog( debug ) << ce->first << " has #" << compatibleCloseSEs.size()
				<< " providing #" << n << " accessible file(s)" << endl;

                        CEs2nFiles[ce->first] = n;
		}
	}
	
	// Remove the CEs which have no CloseSEs speaking the requested protocols
	std::for_each(deletingCEs.begin(), deletingCEs.end(), removeCEFromMatchTable(suitableCEs) );
  	MM.checkRank       (requestAd, *suitableCEs, m_prefetch);
	
	// Remove CEs with undefined rank
	deletingCEs.clear();
	std::accumulate( suitableCEs -> begin(), suitableCEs -> end(), &deletingCEs, insertUnRankedCEsInVector() );
	std::for_each(deletingCEs.begin(), deletingCEs.end(), removeCEFromMatchTable(suitableCEs));

        //Check for previously matched CEs
        deletingCEs.clear();
        std::accumulate( suitableCEs -> begin(), suitableCEs -> end(), &deletingCEs, insertPrevMatchedCEsInVector() );
        if (deletingCEs.size() < suitableCEs->size()) {
          std::for_each(deletingCEs.begin(), deletingCEs.end(), removeCEFromMatchTable(suitableCEs) );
          for( std::vector<std::string>::const_iterator it = deletingCEs.begin(); it != deletingCEs.end(); it++) {
            edglog( info ) << "Will not consider " << *it << " because of previous match" << endl << flush;
          }
        } else if (deletingCEs.size() > 0) {
          if (retry_previous_destinations_disallowed) {
            edglog( info ) << "All possible destinations already tried, not retrying the same ones again" << endl << flush;
            suitableCEs->clear();
          } else {
            edglog( info ) << "Not removing previously matched destinations" << endl << flush;
          }
        }

	map<size_t, set<string> > nFiles2CEs;
	size_t max_files = 0;
	for(match_const_iterator ce = suitableCEs->begin();
	    ce != suitableCEs->end(); ce++) {

		size_t n = CEs2nFiles[ce->first];
		if( n > max_files ) max_files=n;
                nFiles2CEs[n].insert(ce->first);
        }
 		
	const set<string>& CEs_class( nFiles2CEs[max_files] );
	
	deletingCEs.clear();
	std::accumulate( suitableCEs -> begin(), suitableCEs -> end(), &deletingCEs, insertNotInClassCEsInVector(CEs_class) );
  	std::for_each(deletingCEs.begin(), deletingCEs.end(), removeCEFromMatchTable(suitableCEs));
  }
  return suitableCEs;
}

} // namespace broker
} // namespace planning
} // namespace workload
} // namespace edg
