// File: RBSimpleImpl.cpp
// Author: Salvatore Monforte <Salvatore.Monforte@ct.infn.it>
// Copyright (c) 2002 EU DataGrid.
// For license conditions see http://www.eu-datagrid.org/license.html

// $Id: 

#include <algorithm>
#include <functional>
#include <numeric>
#include <vector> 

#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/tuple/tuple.hpp>

#include "edg/workload/common/requestad/ManipulationExceptions.h"
#include "edg/workload/common/requestad/PrivateAdManipulation.h"

#include "edg/workload/planning/broker/matchmaking/matchmakerGlueImpl.h"
#include "edg/workload/planning/broker/matchmaking/glue_attributes.h"
#include "edg/workload/planning/broker/matchmaking/jdl_attributes.h"
#include "edg/workload/planning/broker/RBSimpleImpl.h"
#include "edg/workload/planning/broker/utility.h"

#include "edg/workload/common/logger/edglog.h"
#include "edg/workload/common/logger/manipulators.h"

#define edglog(level) logger::threadsafe::edglog << logger::setlevel(logger::level)
#define edglog_fn(name) logger::StatePusher    pusher(logger::threadsafe::edglog, #name);

using namespace std;

namespace edg {
namespace workload {

namespace requestad     = common::requestad;
namespace logger        = common::logger;

namespace planning {
namespace broker {

RBSimpleImpl::RBSimpleImpl(bool do_prefetch)
{
  m_prefetch = do_prefetch;
}
 
RBSimpleImpl::~RBSimpleImpl()
{
}

matchmaking::match_table_t* RBSimpleImpl::findSuitableCEs(const classad::ClassAd* requestAd)
{
  edglog_fn(RBSimpleImpl::findSuitableCEs);
  bool retry_previous_destinations_disallowed = false;
  bool good;

  if (requestAd &&
      requestad::get_edg_retry_previous_destinations_disallowed(*requestAd, good) && good) {
    retry_previous_destinations_disallowed = true;
  } 

  matchmaking::match_table_t* suitableCEs = 0;
  if( requestAd ) { 
	matchmaking::MatchMaker<matchmaking::matchmakerGlueImpl> MM;
  	suitableCEs = new matchmaking::match_table_t;
	if (m_prefetch) {
  	  MM.prefetchCEInfo(requestAd, *suitableCEs);
	}
  	MM.checkRequirement(requestAd, *suitableCEs, m_prefetch);
  	MM.checkRank       (requestAd, *suitableCEs, m_prefetch);
  	//Remove CEs with undefined rank 
  	std::vector<std::string> deletingCEs;
  	std::accumulate( suitableCEs -> begin(), suitableCEs -> end(), &deletingCEs, insertUnRankedCEsInVector() );
  	std::for_each(deletingCEs.begin(), deletingCEs.end(), removeCEFromMatchTable(suitableCEs) ); 

        //Check for previously matched CEs
        deletingCEs.clear();
        std::accumulate( suitableCEs -> begin(), suitableCEs -> end(), &deletingCEs, insertPrevMatchedCEsInVector() );
        if (deletingCEs.size() < suitableCEs->size()) {
          std::for_each(deletingCEs.begin(), deletingCEs.end(), removeCEFromMatchTable(suitableCEs) );
          for( std::vector<std::string>::const_iterator it = deletingCEs.begin(); it != deletingCEs.end(); it++) {
            edglog( info ) << "Will not consider " << *it << " because of previous match" << endl << flush;
          }
        } else if (deletingCEs.size() > 0) {
          if (retry_previous_destinations_disallowed) {
            edglog( info ) << "All possible destinations already tried, not retrying the same ones again" << endl << flush;
            suitableCEs->clear();
          } else {
            edglog( info ) << "Not removing previously matched destinations" << endl << flush;
          }
        }
  }
  return suitableCEs;
}

} // namespace broker
} // namespace planning
} // namespace workload
} // namespace edg

