// File: brokerinfoGlueImpl.cpp
// Author: Salvatore Monforte <Salvatore.Monforte@ct.infn.it>
// Copyright (c) 2002 EU DataGrid.
// For license conditions see http://www.eu-datagrid.org/license.html

// $Id: brokerinfoGlueImpl.cpp,v 1.35 2005/11/11 16:03:01 dhsmith Exp $

#include <dlfcn.h>                 // for dlopen(), dlclose(), etc.

#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/thread/tss.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/regex.hpp>

#include <classad_distribution.h>

#include "edg/workload/common/logger/edglog.h"
#include "edg/workload/common/logger/manipulators.h"

#include "edg/workload/common/requestad/ManipulationExceptions.h"
#include "edg/workload/common/requestad/JobAdManipulation.h"

#include "edg/workload/common/utilities/Exceptions.h"
#include "edg/workload/planning/broker/matchmaking/brokerinfo.h"
#include "edg/workload/planning/broker/matchmaking/brokerinfoGlueImpl.h"

#include "edg/workload/planning/broker/matchmaking/glue_attributes.h"
#include "edg/workload/planning/broker/matchmaking/jdl_attributes.h"

#define edglog(level) logger::threadsafe::edglog << logger::setlevel(logger::level)
#define edglog_fn(name) logger::StatePusher    pusher(logger::threadsafe::edglog, #name);

using namespace std;

namespace edg {
namespace workload {

namespace configuration = common::configuration;
namespace ldif2classad  = common::ldif2classad;
namespace utilities     = common::utilities;
namespace logger        = common::logger;
namespace requestad     = common::requestad;

namespace planning {

namespace rls         = broker::rls;
namespace dli	      = broker::dli;

namespace broker {
namespace matchmaking { 
namespace brokerinfo {

// Initialize static data members
boost::mutex	brokerinfoGlueImpl::mtx;
dli::create_t*	brokerinfoGlueImpl::createDli=NULL;
dli::destroy_t*	brokerinfoGlueImpl::destroyDli=NULL;
void *		brokerinfoGlueImpl::dliLibHandle=NULL;
rls::create_t*	brokerinfoGlueImpl::createRls=NULL;
rls::destroy_t*	brokerinfoGlueImpl::destroyRls=NULL;
void *		brokerinfoGlueImpl::rlsLibHandle=NULL;

boost::thread_specific_ptr<boost::shared_ptr<BrokerInfo<brokerinfoGlueImpl> *> > BrokerInfo<brokerinfoGlueImpl>::tp_BIp;

extern "C" {
  boost::thread_specific_ptr<boost::shared_ptr<BrokerInfo<brokerinfoGlueImpl> *> > *get_BIp()
  {
    return &BrokerInfo<brokerinfoGlueImpl>::tp_BIp;
  }
}  

brokerinfoGlueImpl::brokerinfoGlueImpl()
{
  m_close_se_info_avail = false;
  m_sa_info_avail = false;

  const common::configuration::NSConfiguration* NSconf = common::configuration::Configuration::instance() -> ns();
  m_IIconnection.reset( new common::ldif2classad::LDAPSynchConnection( NSconf -> ii_dn(),
                         NSconf -> ii_contact(),
                         NSconf -> ii_port(),
                         NSconf -> ii_timeout()) );
}

brokerinfoGlueImpl::~brokerinfoGlueImpl()
{
}

void brokerinfoGlueImpl::retrieveCloseSAsInfo(const BrokerInfoData::VO_name_type& VO, 
		BrokerInfoData& bid,
		std::vector<std::string>* additional_attrs)
{
 edglog_fn(retrieveCloseSAsInfo); 	

 // see if we the common broker info object in the static per thread variable
 boost::shared_ptr<BrokerInfo<brokerinfoGlueImpl> *> *BI_h = BrokerInfo<brokerinfoGlueImpl>::tp_BIp.get();
 if (BI_h != 0 && BI_h->use_count() == 2 && (*BI_h->get())->is_impl(this)) {
   // retrieve SA information for all SEs and cache
   bool reload = !m_sa_info_avail;

   if (m_sa_info_avail) {
     bool ok = true;

     if (!additional_attrs && !m_sa_extra_attrs.empty()) ok=false;
     else if (additional_attrs && additional_attrs->size() != m_sa_extra_attrs.size()) ok = false;

     if (additional_attrs && ok) {
       if (m_sa_vo != VO) ok = false;
     }

     if (additional_attrs && ok) {
       for(std::vector<std::string>::const_iterator s_itr = additional_attrs->begin(); ok && s_itr != additional_attrs->end(); ++s_itr) {
         if (std::find(m_sa_extra_attrs.begin(), m_sa_extra_attrs.end(), *s_itr) == m_sa_extra_attrs.end()) ok = false;
       }
     }

     if (!ok) reload = true;
   }
  
   if (reload) {
     m_sa_info_avail = false;
     m_se_to_sa_info.clear();
     m_sa_vo = VO;
     retrieveCloseSAsInfo(VO, additional_attrs);
   }

   bid.m_referredVO.assign( VO );

   if (!m_sa_info_avail) {
     edglog(warning) << "Failed to get any CloseSA information" << endl;
     return;
   }

   for(BrokerInfoData::CloseSEInfo_iterator it = bid.m_CloseSEInfo_map.begin();it != bid.m_CloseSEInfo_map.end(); it++) {
     const string CloseSE(it->first);
     if (m_se_to_sa_info.find(CloseSE) != m_se_to_sa_info.end()) {
       bid.m_CloseSEInfo_map[CloseSE]->Update(*m_se_to_sa_info[CloseSE]);
     } else {
       edglog(warning) << "No (cached) GlueSA information found for SE " << CloseSE << endl;
     }
   }
   return;
 }

 // query only for SA information for current SE
 bid.m_referredVO.assign( VO );

 try {
        if (! m_IIconnection -> is_established())
          m_IIconnection -> open();

        std::vector<std::string> CloseSAattr;
        CloseSAattr.push_back( GS_GLUESA_STATE_AVAILABLESPACE );
                
        if( additional_attrs && !additional_attrs->empty() ) {
 
                CloseSAattr.insert( CloseSAattr.begin(),
                                additional_attrs->begin(),
                                additional_attrs->end() );
        }                       
        for(BrokerInfoData::CloseSEInfo_iterator it = bid.m_CloseSEInfo_map.begin();
                       it != bid.m_CloseSEInfo_map.end(); it++) {
                const string CloseSE(it->first);
                std::string CloseSAInfo_filter;
                CloseSAInfo_filter.assign("(&(&("+GS_OC+"="+GS_OC_GLUESA+")("+GS_GLUECHUNK_KEY+"="+GS_GLUESE_UNIQUEID+"=" + CloseSE +
                        ")("+GS_GLUESA_ACCESSCONTROL_BASERULE+"=" + VO + ")))");
                
                ldif2classad::LDAPQuery CloseSAInfo_query(m_IIconnection.get(), CloseSAInfo_filter, CloseSAattr);
                try {
                                
                        CloseSAInfo_query.execute();
                        if( !CloseSAInfo_query.tuples() -> empty() ) {
                
                                ldif2classad::LDAPForwardIterator CloseSAInfo_it( CloseSAInfo_query.tuples() );
                                CloseSAInfo_it.first();
                                boost::scoped_ptr<classad::ClassAd> ad((*CloseSAInfo_it).asClassAd());
                                bid.m_CloseSEInfo_map[CloseSE]->Update(*ad);
                        }
                        else {
                                edglog(warning) << "InformationIndex search (no tuples): "
                                        <<  CloseSAInfo_query.what() << endl;
                        }
                }
                catch( ldif2classad::QueryException& e) {
                                
                        edglog( warning ) << e.what() << endl;
                }
                catch( ldif2classad::LDAPNoEntryEx&) {
                        
                        edglog(warning) << "InformationIndex search (no entry): " << CloseSAInfo_query.what() << endl;
                }
        }
 }
 catch( ldif2classad::ConnectionException& e) {
                                
      edglog(warning) << e.what() << endl;
 }
}

void brokerinfoGlueImpl::retrieveCloseSAsInfo(const BrokerInfoData::VO_name_type& VO, 
		std::vector<std::string>* additional_attrs)
{
 edglog_fn(retrieveCloseSAsInfo_allse); 	
 m_sa_info_avail = false;

 try {
   if (! m_IIconnection -> is_established())
     m_IIconnection -> open();

   std::vector<std::string> CloseSAattr;
   CloseSAattr.push_back( GS_GLUESA_STATE_AVAILABLESPACE );
   CloseSAattr.push_back( GS_GLUECHUNK_KEY );
	
   if( additional_attrs && !additional_attrs->empty() ) {

     CloseSAattr.insert( CloseSAattr.begin(), 
		         additional_attrs->begin(), 
		         additional_attrs->end() );
     m_sa_extra_attrs = *additional_attrs;
   } else {
     m_sa_extra_attrs.clear();
   }

   // from here on any failure will not cause invalidation of this cache refresh
   m_sa_info_avail = true;

   std::string CloseSAInfo_filter;
   CloseSAInfo_filter.assign("(&("+GS_OC+"="+GS_OC_GLUESA+")("+GS_GLUESA_ACCESSCONTROL_BASERULE+"=" + VO + "))");
   ldif2classad::LDAPQuery CloseSAInfo_query(m_IIconnection.get(), CloseSAInfo_filter, CloseSAattr);
   try {
     CloseSAInfo_query.execute();
     if( !CloseSAInfo_query.tuples() -> empty() ) {
       ldif2classad::LDAPForwardIterator CloseSAInfo_it( CloseSAInfo_query.tuples() );
       CloseSAInfo_it.first();
       boost::smatch pieces;
       while( CloseSAInfo_it.current() ) {
         const boost::regex  expression( "^\\s*([^\\s=]+)=([^\\s]+).*" );
         std::string chunk_key;

         (*CloseSAInfo_it).EvaluateAttribute(GS_GLUECHUNK_KEY, chunk_key);

         if( boost::regex_match(chunk_key, pieces, expression) ) {
           std::string p[2];
           p[0].assign(pieces[1].first,pieces[1].second);
           p[1].assign(pieces[2].first,pieces[2].second);
           if (p[0] == GS_GLUESE_UNIQUEID) {
             const std::string CloseSE = p[1];
             BrokerInfoData::CloseSEInfo_type CloseSAInfo((*CloseSAInfo_it).asClassAd());
             CloseSAInfo->Delete(GS_GLUECHUNK_KEY);
             m_se_to_sa_info[CloseSE] = CloseSAInfo;
           }
         }
         CloseSAInfo_it.next();
       }
     } else {
       edglog(warning) << "InformationIndex search (no tuples): " 
                       <<  CloseSAInfo_query.what() << endl;
     }
   } catch( ldif2classad::QueryException& e) {
     edglog( warning ) << e.what() << endl;
   } catch( ldif2classad::LDAPNoEntryEx&) { 
     edglog(warning) << "InformationIndex search (no entry): " << CloseSAInfo_query.what() << endl;    
   }    
 } catch( ldif2classad::ConnectionException& e) {
   edglog(warning) << e.what() << endl;
 }
}

void brokerinfoGlueImpl::retrieveCloseSEsInfo(const BrokerInfoData::CEid_type& CEid,
                BrokerInfoData& bid, std::vector<std::string>* additional_attrs)
{
  edglog_fn(retrieveCloseSEsInfo);

  // see if we the common broker info object in the static per thread variable
  boost::shared_ptr<BrokerInfo<brokerinfoGlueImpl> *> *BI_h = BrokerInfo<brokerinfoGlueImpl>::tp_BIp.get();
  if (BI_h != 0 && BI_h->use_count() == 2 && (*BI_h->get())->is_impl(this)) {
    // query for all SEs and cache the results
    bool reload = !m_close_se_info_avail;

    if (m_close_se_info_avail) {
      bool ok = true;

      if (!additional_attrs && !m_close_se_extra_attrs.empty()) ok=false;
      else if (additional_attrs && additional_attrs->size() != m_close_se_extra_attrs.size()) ok = false;

      if (additional_attrs && ok) {
        for(std::vector<std::string>::const_iterator s_itr = additional_attrs->begin(); ok && s_itr != additional_attrs->end(); ++s_itr) {
          if (std::find(m_close_se_extra_attrs.begin(), m_close_se_extra_attrs.end(), *s_itr) == m_close_se_extra_attrs.end()) ok = false;
        }
      }

      if (!ok) reload = true;
    }

    if (reload) {
      m_close_se_info_avail = false;
      if (additional_attrs && !additional_attrs->empty()) {
        m_close_se_extra_attrs = *additional_attrs;
      } else {
        m_close_se_extra_attrs.clear();
      }
      retrieveCloseSEsInfo(m_ce_to_close_se_info, bid, &m_close_se_extra_attrs);
      m_close_se_info_avail = true;
    }

    bid.m_CloseSEInfo_map.clear();
    bid.m_referredCEid.assign( CEid );

    if (m_ce_to_close_se_info.find(CEid) != m_ce_to_close_se_info.end()) {
      BrokerInfoData::CloseSEInfo_map_type &map = m_ce_to_close_se_info[CEid];
      bid.m_CloseSEInfo_map = map;
    } else {
      // no entry for CE
      edglog(warning) << "Could not get CloseSEsInfo for CE: " << CEid << endl;
    }
    return;
  }

  // we are not common broker info object: query only per host and don't cache the result
  bid.m_CloseSEInfo_map.clear();
  bid.m_referredCEid.assign( CEid );
  std::vector<std::string> CloseSEattr;
  CloseSEattr.push_back(GS_GLUECE_SEBIND_GROUP_SEUNIQUEID);
  std::string CloseSE_filter;
  CloseSE_filter = "(&("+GS_OC+"="+GS_OC_GLUECE_SEBIND_GROUP+")("+GS_GLUECE_SEBIND_GROUP_CEUNIQUEID+"=" + CEid + "))";
 
  ldif2classad::LDAPQuery CloseSE_query(m_IIconnection.get(), CloseSE_filter, CloseSEattr);

  try {
      if ( ! m_IIconnection -> is_established())
        m_IIconnection -> open();

      CloseSE_query.execute();
      if( !CloseSE_query.tuples() -> empty() ) {
          try {
                                                             
              ldif2classad::LDAPForwardIterator CloseSE_it( CloseSE_query.tuples() );
              CloseSE_it.first();
              while( CloseSE_it.current() ) {
 
                  vector<string> CloseSEs;
                  (*CloseSE_it).EvaluateAttribute(GS_GLUECE_SEBIND_GROUP_SEUNIQUEID,CloseSEs);
      
                  vector<string> CloseSEInfo_attr;
                  CloseSEInfo_attr.push_back(GS_GLUECE_SEBIND_ACCESSPOINT);
                                                             
                  for(vector<string>::const_iterator it=CloseSEs.begin();
                                  it!=CloseSEs.end(); it++) {

                        string CloseSEInfo_filter("(&("+GS_OC+"="+GS_OC_GLUECE_SEBIND+")(" +
                                 GS_GLUECE_SEBIND_SEUNIQUEID + "=" + *it + "))");
                  
                        ldif2classad::LDAPQuery CloseSEInfo_query(m_IIconnection.get(), CloseSEInfo_filter, CloseSEInfo_attr);
                        try {
                                CloseSEInfo_query.execute();
                                if( !CloseSEInfo_query.tuples() -> empty() ) {
                  
                                        ldif2classad::LDAPForwardIterator CloseSEInfo_it( CloseSEInfo_query.tuples() );
                                        CloseSEInfo_it.first();
                                        ldif2classad::LDIFObject ldif_CloseSE = (*CloseSEInfo_it);
                                        if( additional_attrs && !additional_attrs->empty() ) {

                                                string SE_info_filter("(&("+GS_OC+"="+GS_OC_GLUESE+")("+GS_GLUESE_UNIQUEID+"=" + *it + "))");
                                                edglog( debug ) << "Filtering Information Index: " << SE_info_filter << endl;
                                                ldif2classad::LDAPQuery SE_info_query(m_IIconnection.get(), SE_info_filter, *additional_attrs);
                                                try {
                                                        SE_info_query.execute();
                                                        if(!SE_info_query.tuples()->empty()) {

                                                                ldif2classad::LDAPForwardIterator SE_info_it(SE_info_query.tuples());
                                                                SE_info_it.first();
                                                                if(SE_info_it.current()) {

                                                                        ldif_CloseSE.merge(*SE_info_it);
                                                                }
                                                        }
                                                }
                                                catch( ldif2classad::QueryException& e) {

                                                        edglog( warning ) << e.what() << endl;
                                                }
                                        }
                                        utilities::ii_attributes::const_iterator attrs_begin, attrs_end;
                                        boost::tie(attrs_begin,attrs_end) = utilities::ii_attributes::multiValued();
                                        BrokerInfoData::CloseSEInfo_type CloseSEInfo( ldif_CloseSE.asClassAd(attrs_begin, attrs_end) );
                                        bid.m_CloseSEInfo_map[*it] = CloseSEInfo;
                                }
                                else {

                                        edglog(warning) << "InformationIndex search (no tuples): " <<  CloseSEInfo_query.what() << endl;
                                }
                        }
                        catch( ldif2classad::QueryException& e) {

                                edglog( warning ) << e.what() << endl;
                        }
                  }
                  CloseSE_it.next();  
              }
          }
          catch( ldif2classad::LDAPNoEntryEx& ) {
                         
              edglog(warning) << "InformationIndex search (no entry): " << CloseSE_query.what() << endl;
          }
      }
      else {
                   
              edglog(warning) << "InformationIndex search (no tuples): " << CloseSE_query.what() << endl;
      }
  }
  catch( ldif2classad::QueryException& e) {
                         
      edglog(warning) << e.what() << endl;
  }
  catch( ldif2classad::ConnectionException& e) {

      edglog(warning) << e.what() << endl;
  } 
}

void brokerinfoGlueImpl::retrieveCloseSEsInfo(std::map<BrokerInfoData::CEid_type, BrokerInfoData::CloseSEInfo_map_type> &ce_to_ses,
                                              BrokerInfoData& bid, std::vector<std::string>* additional_attrs)
{ 
  edglog_fn(retrieveCloseSEsInfo_allce);

  ce_to_ses.clear();
  
  std::vector<std::string> QueryAttr;
  QueryAttr.push_back(GS_OC);
  QueryAttr.push_back(GS_GLUECE_SEBIND_GROUP_CEUNIQUEID);
  QueryAttr.push_back(GS_GLUECE_SEBIND_GROUP_SEUNIQUEID);
  QueryAttr.push_back(GS_GLUECE_SEBIND_SEUNIQUEID);
  QueryAttr.push_back(GS_GLUECE_SEBIND_ACCESSPOINT);
  
  std::string query_filter;
  query_filter = "(|("+GS_OC+"="+GS_OC_GLUECE_SEBIND_GROUP+")("+GS_OC+"="+GS_OC_GLUECE_SEBIND+"))";

  if (additional_attrs && !additional_attrs->empty()) {
    query_filter = "(|" + query_filter + "("+GS_OC+"="+GS_OC_GLUESE+"))";
    std::vector<std::string>::const_iterator itr;
    for(itr = additional_attrs->begin(); itr != additional_attrs->end(); ++itr) {
      QueryAttr.push_back(*itr);
    }
  }
  
  ldif2classad::LDAPQuery query(m_IIconnection.get(), query_filter, QueryAttr);
  
  try {
    if ( ! m_IIconnection -> is_established())
      m_IIconnection -> open();

    query.execute();    
    if( !query.tuples() -> empty() ) {

      ldif2classad::LDAPForwardIterator query_it( query.tuples() );
      query_it.first();

      typedef std::map<std::string, ldif2classad::LDIFObject> obj_map_t;

      obj_map_t CE_to_CESEBindGroup;
      obj_map_t SE_to_CESEBind;
      obj_map_t SE_to_GlueSE;

      bool adv = false;
      while( (adv && query_it.next()) || (!adv && query_it.current() && (adv=true)) ) {
        vector<string> objectClass;
        query_it->EvaluateAttribute(GS_OC,objectClass);

        if (std::find(objectClass.begin(), objectClass.end(), GS_OC_GLUECE_SEBIND_GROUP) != objectClass.end()) {
          string CEid;
          if (query_it->EvaluateAttribute(GS_GLUECE_SEBIND_GROUP_CEUNIQUEID,CEid)) {
            CE_to_CESEBindGroup[CEid] = (*query_it);
          }
        } else if (std::find(objectClass.begin(), objectClass.end(), GS_OC_GLUECE_SEBIND) != objectClass.end()) {
          string SEid;
          if (query_it->EvaluateAttribute(GS_GLUECE_SEBIND_SEUNIQUEID,SEid)) {
            SE_to_CESEBind[SEid] = (*query_it);
          }
        } else if (additional_attrs && !additional_attrs->empty() &&
                   std::find(objectClass.begin(), objectClass.end(), GS_OC_GLUESE) != objectClass.end()) {
          string SEid;
          if (query_it->EvaluateAttribute(GS_GLUESE_UNIQUEID,SEid)) {
            SE_to_GlueSE[SEid] = (*query_it);
          }
        }
      }

      for(obj_map_t::const_iterator ce_it=CE_to_CESEBindGroup.begin(); ce_it != CE_to_CESEBindGroup.end(); ++ce_it) {

        string CEid = ce_it->first;

        const ldif2classad::LDIFObject &CESEBindGroup = ce_it->second;
                
        vector<string> CloseSEs;
        CESEBindGroup.EvaluateAttribute(GS_GLUECE_SEBIND_GROUP_SEUNIQUEID,CloseSEs);
			      
	for(vector<string>::const_iterator it=CloseSEs.begin(); it!=CloseSEs.end(); it++) {
       
          obj_map_t::const_iterator se_it;

          if ( (se_it = SE_to_CESEBind.find(*it)) != SE_to_CESEBind.end() ) {

            const ldif2classad::LDIFObject &CESEBind = se_it->second;

            ldif2classad::LDIFObject ldif_CloseSE;
            vector<string> s;
            CESEBind.EvaluateAttribute(GS_GLUECE_SEBIND_ACCESSPOINT, s);
            for(vector<string>::const_iterator s_itr = s.begin(); s_itr != s.end(); ++s_itr) {
              ldif_CloseSE.add(GS_GLUECE_SEBIND_ACCESSPOINT, *s_itr);
            }

	    if( additional_attrs && !additional_attrs->empty() ) {

              obj_map_t::const_iterator se_extra_it;

              if ( (se_extra_it = SE_to_GlueSE.find(*it)) != SE_to_GlueSE.end() ) {

                const ldif2classad::LDIFObject &GlueSE = se_extra_it->second;

                ldif2classad::LDIFObject ldif_se_info;
                for(vector<string>::const_iterator s_itr = additional_attrs->begin(); s_itr != additional_attrs->end(); ++s_itr) {
                  vector<string> attrs;
                  GlueSE.EvaluateAttribute(*s_itr, attrs);
                  for(vector<string>::const_iterator a_itr = attrs.begin(); a_itr != attrs.end(); ++a_itr) {
                    ldif_se_info.add(*s_itr, *a_itr);
                  }
                }
                ldif_CloseSE.merge(ldif_se_info);
              }
            }
 
            utilities::ii_attributes::const_iterator attrs_begin, attrs_end;
            boost::tie(attrs_begin,attrs_end) = utilities::ii_attributes::multiValued();
            BrokerInfoData::CloseSEInfo_type CloseSEInfo( ldif_CloseSE.asClassAd(attrs_begin, attrs_end) );

            if (ce_to_ses.find(CEid) == ce_to_ses.end()) {
              BrokerInfoData::CloseSEInfo_map_type map;
              ce_to_ses[CEid] = map;
            }

            (ce_to_ses[CEid])[*it] = CloseSEInfo;

          } else {
            edglog(warning) << "Did not find object of class " << GS_OC_GLUECE_SEBIND << " for SE: " << *it << endl;
          }
        } // for loop 'it' over CloseSEs
      } // for loop 'ce_it' over CE_to_CESEBindGroup

    } else {
      // no results on query result
      edglog(warning) << "InformationIndex search (no tuples): " << query.what() << endl;
    }
  }
  catch( ldif2classad::QueryException& e) {
	  
      edglog(warning) << e.what() << endl;
  }
  catch( ldif2classad::ConnectionException& e) {
	  
      edglog(warning) << e.what() << endl;
  }
}

void brokerinfoGlueImpl::retrieveSEsInfo(const classad::ClassAd& requestAd, BrokerInfoData& bid)
{
  edglog_fn(retrieveSEsInfo);	
  
  vector<string> attributes;
  
  attributes.push_back(GS_GLUESE_ACCESSPROTOCOL_TYPE);
  attributes.push_back(GS_GLUESE_ACCESSPROTOCOL_PORT);
  BrokerInfoData::SE_const_iterator se_begin, se_end;
  boost::tie(se_begin,se_end) = bid.involvedSEs();
  for (BrokerInfoData::SE_const_iterator it = se_begin; it != se_end; it++) {    
    string filter;	
    filter = "(&("+GS_OC+"="+GS_GLUESE_ACCESSPROTOCOL+")("+GS_GLUECHUNK_KEY+"="+GS_GLUESE_UNIQUEID+"=" + (*it) + "))";
    
    ldif2classad::LDAPQuery query(m_IIconnection.get(),filter,attributes);
    
    try {
      
      if ( ! m_IIconnection -> is_established())
        m_IIconnection -> open();

      query.execute();
      
      if( !query.tuples() -> empty() ) {
	
	try {
	  
	  ldif2classad::LDAPForwardIterator ldap_it( query.tuples() );   
	  ldap_it.first();
	  while( ldap_it.current() ) {
	    
	    BrokerInfoData::protocol_name protocol;
	    std::string str_port;
	    
	    (*ldap_it).EvaluateAttribute(GS_GLUESE_ACCESSPROTOCOL_TYPE, protocol);
	    (*ldap_it).EvaluateAttribute(GS_GLUESE_ACCESSPROTOCOL_PORT, str_port);
	    
	    bid.m_SE2Protocol_map[(*it)].push_back( std::make_pair(protocol,std::atoi(str_port.c_str())) );
	    ldap_it.next();
	  }
	} 
	catch( ldif2classad::LDAPNoEntryEx&) {
	  edglog(warning) << "InformationIndex search (no entry): " <<  query.what() << endl;
	}
      } 
      else {
	edglog(warning) << "InformationIndex search (no tuples): " << query.what() << endl;  
      }
    } 
    catch( ldif2classad::QueryException& e) {
      edglog(warning) << e.what() << endl;
    }
    catch ( ldif2classad::ConnectionException& e) {
      edglog(warning) << e.what() << endl;
    }
  }
}

void brokerinfoGlueImpl::retrieveSFNsInfo(const classad::ClassAd& requestAd, BrokerInfoData& bid)
{ 
  // DEFAULT currently to try RLS //
  bool     rlsConfig = true;  // RLS is to be checked for lfn or guid
  bool     dliConfig = false; // DLI to be used for all input data types
  bool     dliUsage;
  string   vo;
  string   endpoint;

  edglog_fn(retrieveSFNsInfo);	
  try {    
    
    // The InputData field is used to specify a list of LFN, GUID, LDS or Query
    //  
    BrokerInfoData::LFN_container_type input_data;
    requestad::get_input_data(requestAd, input_data);
    vo = requestad::get_virtual_organisation(requestAd);

    // Check if we have the DLI/RLS configured for the VO.
    //
    if (checkDliUsage(vo) == 0) {
      rlsConfig = false;
      dliConfig = true;
    } else if (checkRlsUsage(vo) == 0) {
      rlsConfig = true;
      dliConfig = false;
    }
    
    // Check which InputDataType we got from the JDL. We have the following
    // possibilities:
    //    - If RLS is not to be tried: use DLI for all InputDataTypes  such
    //                                 as lfn, guid, lds and query
    //    - If RLS is to be tried: We have these options: 
    //        * lfn, guid   -> use RLS 
    //        * lds, query  -> use DLI  
    // Currently, we can either use the RLS or the DLI interface but
    // we don't consult both for one job.
    //
    dliUsage = false;
    if (rlsConfig == true ) {
      // If RLS is configured, we might still want to use the DLI in case
      // lds or query is used in the JDL.
      //
      for (unsigned int i = 0; i < input_data.size(); i++) {
	if ((input_data[i].find("lds:") == 0) ||
	    (input_data[i].find("query:") == 0)) {
	  dliUsage = true;
	}
      }
    } else {
      dliUsage = true;
    }

    // Load the client dynamically as a plug-in, but don't make
    // external symbols visible outside the plugin, in order to
    // avoid SOAP version clashes between DLI and RLS
    //
    if (dliUsage) {
      // Dynamically load the DLI plug-in, if not already loaded
      //
      boost::mutex::scoped_lock lk(mtx);
      if (dliLibHandle == NULL) {
        createDli = NULL;
        destroyDli = NULL;
        string dliLib = "libedg-wl-dli.so";
        dliLibHandle = dlopen (dliLib.c_str(), RTLD_NOW);
        if (dliLibHandle == NULL) {
  	  edglog(warning) << "cannot load DLI helper lib " << dliLib << endl;
	  edglog(warning) << "dlerror returns: " << dlerror() << endl;
	  return; 
        }
      }

      if (createDli == NULL || destroyDli == NULL) {
        createDli = (dli::create_t*)dlsym(dliLibHandle,"create");
        destroyDli = (dli::destroy_t*)dlsym(dliLibHandle,"destroy");
        if (createDli == NULL || destroyDli == NULL) {
	  edglog(warning) << "cannot load DLI helper symbols" << endl;
	  edglog(warning) << "dlerror returns: " << dlerror() << endl;
          createDli = NULL;
          destroyDli = NULL;
          int dlstat = dlclose(dliLibHandle);
          if (dlstat == 0) dliLibHandle=NULL;
	  return;  
        } 
      }

      // Determine the endpoint for the catalogue that implements the DLI
      // We have two options: - specify ReplicaCatalog in JDL
      //                      - or get it from the Information Serice (IS)
      //                      
      try {
	endpoint = requestad::get_replica_catalog(requestAd);
      } catch (...) {
	// Try to get URL from Information system
	//
	endpoint = getDLIurl(vo,bid);
	if (endpoint == "") {
	  edglog(warning) << "No DataLocationInterface found in IS." 
			  << " Cannot retrieve replica location. " << endl;
	  return;
	}
      }
    } else {
      // Dynamically load the RLS plug-in, if not already loaded
      //
      boost::mutex::scoped_lock lk(mtx);
      if (rlsLibHandle == NULL) {
        createRls = NULL;
        destroyRls = NULL;
        string rlsLib = "libedg-wl-replicaservice.so"; 
        rlsLibHandle = dlopen (rlsLib.c_str(), RTLD_NOW);
        if (rlsLibHandle == NULL) {
 	  edglog(warning) << "cannot load RLS helper lib " << rlsLib << endl;
	  edglog(warning) << "dlerror returns: " << dlerror() << endl;
	  return; 
        }
      }

      if (createRls == NULL || destroyRls == NULL) {
        createRls = (rls::create_t*)dlsym(rlsLibHandle,"create");
        destroyRls = (rls::destroy_t*)dlsym(rlsLibHandle,"destroy");
        if (createRls == NULL || destroyRls == NULL) {
	  edglog(warning) << "cannot load RLS helper symbols" << endl;
	  edglog(warning) << "dlerror returns: " << dlerror() << endl;
          createRls = NULL;
          destroyRls = NULL;
          int dlstat = dlclose(rlsLibHandle);
          if (dlstat == 0) rlsLibHandle=NULL;
	  return;  
        }
      }
    } // dliUsage
      
    for(BrokerInfoData::LFN_container_type::const_iterator lfn = input_data.begin();
	lfn != input_data.end(); lfn++) {
      try {
	BrokerInfoData::SFN_container_type resolved_sfn;
	edglog(debug) << "Listing replica(s) for " << *lfn; 
	if (dliUsage) { edglog(debug) << " using DLI " << endl; }
	else          { edglog(debug) << " using RLS " << endl; }
	
	// If DLI (DataLocationInterface) is used, call its listReplicas 
	// method, otherwise call the equivalent RLS method
	//
	if (dliUsage) {
	  dli::DataLocationInterfaceSOAP* dli(createDli(vo, endpoint));

	  // Extract the data type from the 'lfn' string
	  //
          size_t delin = lfn->find(":");
          std::string inputDataType = "none";
          if (delin != lfn->npos) {
            inputDataType = lfn->substr(0,delin);
          }
          resolved_sfn = dli->listReplicas(inputDataType, *lfn);
	  destroyDli(dli);
	}
	else { 
	  rls::ReplicaServiceReal* replica(createRls(vo));
	  resolved_sfn = replica->listReplica(*lfn);
	  destroyRls(replica);
	}

	if( !resolved_sfn.empty() ) {
	  
	  bid.m_LFN2SFN_map[*lfn] = resolved_sfn;

	  // static boost::regex  expression( "(.*):[\\s/]*([^\\s/]+)/.*" );
	  static boost::regex  expression( "^\\s*([^:]*):[\\s/]*([^\\s:/]+)(:[0-9]+)?/.*" );

	  // Here we check that each returned SFN corresponds to a URL
	  // of the general form:  protocol://SE-hostname/filepath
	  // Only if this is the case, the an SE is included into the bid.
	  // If the return SFN is only a hostname, we check if the SEid
	  // is registered in the Information System: if yes, fine too.
	  // 
	  for(BrokerInfoData::SFN_container_type::const_iterator sfn = resolved_sfn.begin();
	      sfn != resolved_sfn.end(); sfn++) {
	    edglog(debug) << *sfn << endl;

	    try {
	      
	      boost::smatch pieces;
	      std::string   SE_name;

	      if( boost::regex_match(*sfn, pieces, expression) ) {
		SE_name.assign(pieces[2].first, pieces[2].second);
		bid.m_involvedSEs.insert(SE_name);
	      } else {
		// If the SFN doesn't match the regular expression, we assume
		// that only the SE name (SEid) has been returned. We check
		// if the string really corresponds to a valid SE in the IS.
		//
		if (validSE(*sfn,bid) == 0) {
		  bid.m_involvedSEs.insert(*sfn);
		} 
	      }
	    }
	    catch( std::exception& ex ) {
	         edglog(warning) << ex.what() << endl;
	    }
	  }
	}
	else {
	  edglog(debug) << "No replica(s) found!" << endl;
	}
      } 
      catch( std::exception& ex ) {
	   edglog(warning) << ex.what() << endl;
      }
      catch(char *faultstring) {
	// Catch exceptions from the DataLocationInterface
	//
	edglog(warning) << "Exception from DLI: " << faultstring << endl;
      }
    } // for

    // Unload the plug-in(s) ?
    //

    // Commented out for now - don't unload the pluggins //
  /*
    {
      boost::mutex::scoped_lock lk(mtx);

      if (dliLibHandle != NULL) {
        int dlstat = dlclose(dliLibHandle);
        if (dlstat == 0) {
          dliLibHandle = NULL;
          createDli = NULL;
          destroyDli = NULL;
        } else {
	  edglog(warning) << "DLI plug-in not closed correctly. \n";
	  edglog(warning) << "dlerror returns: " << dlerror() << endl;
        }
      }
      if (rlsLibHandle != NULL) {
        int dlstat = dlclose(rlsLibHandle);
        if (dlstat == 0) {
          rlsLibHandle = NULL;
          createRls = NULL;
          destroyRls = NULL;
        } else {
	  edglog(warning) << "RLS plug-in not closed correctly. \n";
	  edglog(warning) << "dlerror returns: " << dlerror() << endl;
        } 
      }
    }
  */

  } 
  catch( std::exception& ex ) {
	   edglog(warning) << ex.what() << endl;
  }
}

/**
 * Contact the Information Service (IS) and check if the given SE (SEid)
 * is registered in the IS. If yes, the SE is valid and 0 is returned, 
 * otherwise -1. 
 */
int brokerinfoGlueImpl::validSE(std::string SEid,BrokerInfoData& bid)
{
  int valid = -1; // default value: SE is not valid and not in IS

  edglog_fn(validSE);	
  
  vector<string> attributes;
  
  attributes.push_back(GS_GLUESE_UNIQUEID);
  
  string filter;	
  filter = "(&("+GS_OC+"="+GS_OC_GLUESE+"))";

  ldif2classad::LDAPQuery query(m_IIconnection.get(),filter,attributes);
    
  try {
    if ( ! m_IIconnection -> is_established())
      m_IIconnection -> open();

    query.execute();
      
    if( !query.tuples() -> empty() ) {
      try {
	ldif2classad::LDAPForwardIterator ldap_it( query.tuples() );   
	ldap_it.first();
	
	while( ldap_it.current() ) {
	  std::string currentSE;
	  (*ldap_it).EvaluateAttribute(GS_GLUESE_UNIQUEID, currentSE); 
	  if (currentSE == SEid) { valid = 0; }
	  ldap_it.next();
	}
      } 
      catch( ldif2classad::LDAPNoEntryEx&) {
	edglog(warning) << "InformationIndex search (no entry): " <<  query.what() << endl;
      }
    }
    else {
      edglog(warning) << "InformationIndex search (no tuples): " << query.what() << endl;  
    }
  } // try connection 
  catch( ldif2classad::QueryException& e) {
    edglog(warning) << e.what() << endl;
  }
  catch ( ldif2classad::ConnectionException& e) {
    edglog(warning) << e.what() << endl;
  }
  
  return valid;
} // validSE


/**
 * Contact the Information Service (IS) and return the URL (endpoint) of the
 * server the provides the DataLocationInterface(DLI).
 * If no service is found, "" is returned.
 */
std::string brokerinfoGlueImpl::getDLIurl(std::string vo, BrokerInfoData& bid)
{

  string url = "";

  edglog(debug) << "Contact IS for DataLocationInterface. " << endl;
  
  vector<string> attributes;
  attributes.push_back(GS_GLUESERVICE_ACCESSPOINTURL);
  attributes.push_back(GS_GLUESERVICE_TYPE);
  attributes.push_back(GS_GLUESERVICE_ACCESSCONTROLRULE);

  string filter;	
  filter = "(&("+GS_OC+"="+GS_OC_GLUESERVICE+")" +
           string("(" + GS_GLUESERVICE_TYPE + "=data-location-interface)") +
           string("(" + GS_GLUESERVICE_ACCESSCONTROLRULE + "=") + vo + string("))");

  ldif2classad::LDAPQuery query(m_IIconnection.get(),filter,attributes);
    
  try {
    if ( ! m_IIconnection -> is_established())
      m_IIconnection -> open();

    query.execute();
      
    if( !query.tuples() -> empty() ) {
      try {
	ldif2classad::LDAPForwardIterator ldap_it( query.tuples() );   
	ldap_it.first();
	
	while( ldap_it.current() ) {
	  std::string currentURL, currentType, currentVO;
	  (*ldap_it).EvaluateAttribute(GS_GLUESERVICE_ACCESSPOINTURL, currentURL); 
	  url=currentURL;
	  ldap_it.next();
	}
      } 
      catch( ldif2classad::LDAPNoEntryEx&) {
	edglog(warning) << "InformationIndex search (no entry): " <<  query.what() << endl;
      }
    }
    else {
      edglog(warning) << "InformationIndex search (no tuples): " << query.what() << endl;  
    }
  } // try connection 
  catch( ldif2classad::QueryException& e) {
    edglog(warning) << e.what() << endl;
  }
  catch ( ldif2classad::ConnectionException& e) {
    edglog(warning) << e.what() << endl;
  }
  return url;
} // getDLIurl


/*
 * Check the configuration file of the Networkserver if DLI is used for
 * a certain VO. In case the DLI is used, 0 is returned, Otherwise -1.
 */
int brokerinfoGlueImpl::checkDliUsage(std::string vo)
{
  const configuration::NSConfiguration* NSconf = configuration::Configuration::instance() -> ns();

  std::vector<std::string> dli = NSconf->dlicatalog();

  for (unsigned int i=0; i < dli.size(); i++) {
    if (dli[i] == vo) { return 0;}
  }

  return -1;
} // checkDliUsage

/*
 * Check the configuration file of the Networkserver if RLS is used for
 * a certain VO. In case the RLS is used, 0 is returned, Otherwise -1.
 */
int brokerinfoGlueImpl::checkRlsUsage(std::string vo)
{
  const configuration::NSConfiguration* NSconf = configuration::Configuration::instance() -> ns();

  std::vector<std::string> rls = NSconf->rlscatalog();

  for (unsigned int i=0; i < rls.size(); i++) {
    if (rls[i] == vo) { return 0;}
  }

  return -1;
} // checkRlsUsage

std::string brokerinfoGlueImpl::findLCGApplicationDir(const std::string &ceid, BrokerInfoData& bid)
{
  edglog_fn(findLCGApplicationDir);	
  
  string filter, application_dir;
  const string vo = "dteam";
  vector<string> attributes;

  attributes.push_back(GS_GLUECE_INFO_APPLICATIONDIR);
  attributes.push_back(GS_GLUEVOVIEW_LOCALID);

  filter.assign("(&(&("+GS_OC+"="+GS_OC_GLUEVOVIEW+")("+GS_GLUECHUNK_KEY+"="+GS_GLUECE_UNIQUEID+"=" +
                        ceid +")("+GS_GLUECE_ACCESSCONTROL_BASERULE+"=VO:"+vo+")))");

  ldif2classad::LDAPQuery query(m_IIconnection.get(),filter,attributes);
    
  try {
      
    if ( ! m_IIconnection -> is_established())
      m_IIconnection -> open();

    query.execute();
      
    if( !query.tuples() -> empty() ) {
  
      try {
	  
        ldif2classad::LDAPForwardIterator ldap_it( query.tuples() );   
	ldap_it.first();
	while( application_dir.empty() && ldap_it.current() ) {
	    
	  std::string dir,localid;

	  (*ldap_it).EvaluateAttribute(GS_GLUECE_INFO_APPLICATIONDIR, dir);
	  (*ldap_it).EvaluateAttribute(GS_GLUEVOVIEW_LOCALID, localid);

          if (localid == vo)
            application_dir = dir;
            
	  ldap_it.next();
	}
      } 
      catch( ldif2classad::LDAPNoEntryEx&) {
        edglog(warning) << "InformationIndex search (no entry): " <<  query.what() << endl;
      }
    } 
    else {
      edglog(warning) << "InformationIndex search (no tuples): " << query.what() << endl;  
    }
  } 
  catch( ldif2classad::QueryException& e) {
    edglog(warning) << e.what() << endl;
  }
  catch ( ldif2classad::ConnectionException& e) {
    edglog(warning) << e.what() << endl;
  }

  return application_dir;
}

} // namespace brokerinfo
} // namespace matchmaking
} // namespace broker
} // namespace planning
} // namespace workload
} // namespace edg
