// File: RequestStateMachine.cpp
// Author: Francesco Giacomini <Francesco.Giacomini@cnaf.infn.it>
// Copyright (c) 2002 EU DataGrid.
// For license conditions see http://www.eu-datagrid.org/license.html

// $Id: RequestStateMachine.cpp,v 1.7 2003/11/11 14:18:57 giaco Exp $

#include <boost/scoped_ptr.hpp>
#include <classad_distribution.h>
#include "edg/workload/planning/helper/RequestStateMachine.h"
#include "edg/workload/planning/helper/Helper.h"
#include "edg/workload/common/utilities/classad_utils.h"

namespace utilities = edg::workload::common::utilities;

namespace edg {
namespace workload {
namespace planning {
namespace helper {

typedef RequestStateMachine::state_type state_type;
typedef RequestStateMachine::expression_type expression_type;
typedef RequestStateMachine::initial_states_type initial_states_type;
typedef RequestStateMachine::transition_table_type transition_table_type;
typedef RequestStateMachine::final_states_type final_states_type;

namespace {

const std::string state_machine_reqs("[requirements="
                                     "islist(other.initial_states)"
                                     "&& islist(other.transition_table)"
                                     "&& islist(other.final_states)]");

void init_initial_states(initial_states_type& states,
                         classad::ExprList const* l)
{
  states.insert(std::make_pair(state_type("SimpleLogicalJob"),
                               expression_type("other.type==\"job\"")));
}

void init_transition_table(transition_table_type& table,
                           classad::ExprList const* l)
{
  table.insert(
    std::make_pair(
      state_type("SimpleLogicalJob"),
      boost::make_tuple(
        std::string("BrokerHelper"),
        state_type("SimplePhysicalJob")
      )
    )
  );
  table.insert(
    std::make_pair(
      state_type("SimplePhysicalJob"),
      boost::make_tuple(
        std::string("JobAdapterHelper"),
        state_type("AdaptedJob")
      )
    )
  );
}

void init_final_states(final_states_type& states,
                       classad::ExprList const* l)
{
  states.insert(state_type("AdaptedJob"));
}

bool state_is_final(state_type const& state,
                    final_states_type const& final_states)
{
  return final_states.find(state) != final_states.end();
}

}

RequestStateMachine::RequestStateMachine(classad::ClassAd const* config)
{
  init_initial_states(m_initial_states, 0);
  init_transition_table(m_transition_table, 0);
  init_final_states(m_final_states, 0);
}

RequestStateMachine::~RequestStateMachine()
{
}

class MatchAd: public std::unary_function<initial_states_type::value_type, bool>
{
  classad::ClassAd const* m_ad;
public:
  MatchAd(classad::ClassAd const* ad): m_ad(ad) {}
  result_type operator()(argument_type const& v) const
  {
    expression_type expr = "[requirements=" + v.second + "]";
    boost::scoped_ptr<classad::ClassAd> rhs(utilities::parse_classad(expr));
    
    return utilities::left_matches_right(*m_ad, *rhs);
  }
};

void
RequestStateMachine::start(classad::ClassAd const* ad)
{
  initial_states_type::const_iterator it = find_if(m_initial_states.begin(),
                                                   m_initial_states.end(),
                                                   MatchAd(ad));
  if (it == m_initial_states.end()) {
    throw NoValidState();
  }
  m_current_state = it->first;
}

classad::ClassAd*
RequestStateMachine::next_step(classad::ClassAd const* ad)
{
  assert(ad != 0);

  classad::ClassAd* result = 0;

  transition_table_type::const_iterator it = m_transition_table.find(m_current_state);
  if (it == m_transition_table.end()) {
    throw NoValidState();
  }

  std::string helper;
  state_type next_state;
  boost::tie(helper, next_state) = it->second;

  result = Helper(helper).resolve(ad);
  m_current_state = next_state;

  assert(result != 0);

  return result;
}

RequestStateMachine::operator void*() const
{
  return state_is_final(m_current_state, m_final_states) ? (void*)0 : (void*)(-1);
}

int
RequestStateMachine::operator!() const
{
  return state_is_final(m_current_state, m_final_states);
}

}}}} // edg::workload::planning::helper
