/***************************************************************************
 *  filename  : JobAdapter.cpp
 *  authors   : Elisabetta Ronchieri <elisbetta.ronchieri@cnaf.infn.it>
 *  Copyright (c) 2002 CERN and INFN on behalf of the EU DataGrid.
 *  For license conditions see LICENSE file or
 *  http://www.edg.org/license.html
 ***************************************************************************/

#include <string>
#include <iostream>
#include <algorithm>
#include <cctype>
#include <cassert>
#include <fstream>
#include <vector>

#include <unistd.h>
#include <errno.h>
#include <netdb.h>

#include <memory>

#include <classad_distribution.h>

#include <boost/lexical_cast.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/exception.hpp>

#include "edg/workload/planning/common/URL.h"
#include "edg/workload/planning/jobadapter/JobAdapter.h"
#include "edg/workload/planning/jobadapter/JobWrapper.h"
#include "edg/workload/planning/jobadapter/MpiLsfJobWrapper.h"
#include "edg/workload/planning/jobadapter/MpiPbsJobWrapper.h"
#include "edg/workload/planning/jobadapter/InteractiveJobWrapper.h"
#include "edg/workload/common/jobid/JobId.h"
#include "edg/workload/common/jobid/manipulation.h"
#include "edg/workload/common/configuration/Configuration.h"
#include "edg/workload/common/configuration/JCConfiguration.h"
#include "edg/workload/common/configuration/LMConfiguration.h"
#include "edg/workload/common/configuration/NSConfiguration.h"
#include "edg/workload/common/configuration/WMConfiguration.h"
#include "edg/workload/common/utilities/classad_utils.h"
#include "edg/workload/common/utilities/utils.h"
#include "edg/workload/common/utilities/boost_fs_add.h"
#include "edg/workload/common/requestad/JobAdManipulation.h"
#include "edg/workload/common/requestad/PrivateAdManipulation.h"
#include "edg/workload/common/requestad/ManipulationExceptions.h"
#include "edg/workload/common/requestad/JDLAttributes.h"
#include "edg/workload/common/requestad/PrivateAttributes.h"
#include "edg/workload/planning/helper/exceptions.h"
#include "edg/workload/planning/jobadapter/exceptions.h"

using namespace std;
using namespace classad;

namespace config = edg::workload::common::configuration;
namespace jobid = edg::workload::common::jobid;
namespace jobwrapper = edg::workload::planning::jobadapter;
namespace url = edg::workload::planning::common;
namespace utilities = edg::workload::common::utilities;
namespace requestad = edg::workload::common::requestad;

namespace {
std::string const helper_id("JobAdapterHelper");
}

namespace edg {
namespace workload {
namespace planning {
namespace jobadapter {

namespace {
class Beginning {
public:
  Beginning(const string& str) 
    : m_what(str)
    {
    }

  bool operator()(const string& str) const
  {
    return str.find(m_what) != string::npos;
  }
private:
  string m_what;
};
}

JobAdapter::JobAdapter(ClassAd const* ad)
 : m_ad(ad)
{
}

JobAdapter::~JobAdapter(void)
{
}

ClassAd*
JobAdapter::resolve(void)  
try {
  auto_ptr<ClassAd> result(new ClassAd);
  
  /* Mandatory */
  string executable(requestad::get_executable(*m_ad));
  if (executable.empty()) {
    throw helper::InvalidAttributeValue(requestad::JDL::EXECUTABLE,
                                        executable,
                                        "not empty",
                                        helper_id);
  }
  
  /* Mandatory */
  /* It is renamed as usersubjectname and reinserted with */
  string certificatesubject(requestad::get_certificate_subject(*m_ad));
  if (certificatesubject.empty()) {
    throw helper::InvalidAttributeValue(requestad::JDL::CERT_SUBJ,
                                        certificatesubject,
                                        "not empty",
                                        helper_id);
  }  

  requestad::set_user_subject_name(*result, certificatesubject);
  
  /* Not Mandatory */
  vector<string>  outputsandbox;
  utilities::EvaluateAttrListOrSingle(*m_ad, "outputsandbox", outputsandbox); 
  
  /* Not Mandatory */
  vector<string>  inputsandbox;
  utilities::EvaluateAttrListOrSingle(*m_ad, "inputsandbox", inputsandbox);

  /* Mandatory */
  string inputsandboxpath(requestad::get_input_sandbox_path(*m_ad));
  if (inputsandboxpath.empty()) { 
    throw helper::InvalidAttributeValue(requestad::JDLPrivate::INPUT_SANDBOX_PATH,
                                        inputsandboxpath,
                                        "not empty",
                                        helper_id);
  }
	  
  /* Mandatory */
  string outputsandboxpath(requestad::get_output_sandbox_path(*m_ad));
  if (outputsandboxpath.empty()) {
    throw helper::InvalidAttributeValue(requestad::JDLPrivate::OUTPUT_SANDBOX_PATH,
                                        outputsandboxpath,
                                        "not empty",
                                        helper_id);
  }
  
  /* Not Mandatory */
  bool   b_arg;
  string arguments(requestad::get_arguments(*m_ad, b_arg));
 
  /* Not Mandatory */
  vector<string> env;
  utilities::EvaluateAttrListOrSingle(*m_ad, requestad::JDL::ENVIRONMENT, env);

  /* Mandatory */
  requestad::set_copy_to_spool(*result, false);
  requestad::set_transfer_executable(*result, true);
  requestad::set_universe(*result, "Globus");
 
  /* Mandatory */
  requestad::set_notification(*result, "never");
  
  /* Not Mandatory */
  bool b_std;
  string stdinput(requestad::get_std_input(*m_ad, b_std));
  string stderror(requestad::get_std_error(*m_ad, b_std));
  string stdoutput(requestad::get_std_output(*m_ad, b_std));
  
  // TEMP patch: forward ce_id
  bool   b_ce_id;
  string ce_id(requestad::get_ce_id(*m_ad, b_ce_id));
  requestad::set_ce_id(*result, ce_id, b_ce_id);
  
  /* Mandatory */
  /* It is renamed in globusscheduler. */
  string globusresourcecontactstring(requestad::get_globus_resource_contact_string(*m_ad));
  if (globusresourcecontactstring.empty()) {
    throw helper::InvalidAttributeValue(requestad::JDL::GLOBUSRESOURCE,
                                        globusresourcecontactstring,
                                        "not empty",
                                        helper_id);
  }
  requestad::set_globus_scheduler(*result, globusresourcecontactstring);
  
  /* Mandatory */
  /* job id is mandatory to build the globusrsl string and to create the */
  /* job wrapper                                                         */
  string job_id(requestad::get_edg_jobid(*m_ad));
  if (job_id.empty()) {
    throw helper::InvalidAttributeValue(requestad::JDL::JOBID,
                                        job_id,
                                        "not empty",
                                        helper_id);
  }
  requestad::set_edg_jobid(*result, job_id);

  /* Mandatory */
  string requirements("EDG_WL_JDL_REQ");
  requirements.append(" '");
  string reqvalue(requestad::unparse_requirements(*m_ad));
  if (reqvalue.empty()) {
    throw helper::InvalidAttributeValue(requestad::JDL::REQUIREMENTS,
                                        reqvalue,
                                        "not empty",
                                        helper_id);
  }
  utilities::replace(reqvalue, "\"", "\\\"");
  requirements.append(reqvalue);
  requirements.append("'");
  
  /* Mandatory */
  /* queuname is mandatory to build the globusrsl string. */ 
  string queuename(requestad::get_queue_name(*m_ad));
  if (queuename.empty()) {
    throw helper::InvalidAttributeValue(requestad::JDL::QUEUENAME,
                                        queuename,
                                        "not empty",
                                        helper_id);
  }

  string globusrsl("(queue=");
  globusrsl.append(queuename);
  globusrsl.append(")(jobtype=single)");

  //globusrsl.append("(environment=(");
  //globusrsl.append(requirements);
  //globusrsl.append("))");

  /* Not Mandatory */
  {
    globusrsl.append("(environment=(");
    string jid("EDG_WL_JOBID");
    jid.append(" '");
    string jidvalue(job_id);
    utilities::replace(jidvalue, "\"", "\\\"");
    jid.append(jidvalue);
    jid.append("'");
    globusrsl.append(jid);
    globusrsl.append(")");

    bool good_version;
    string mwversion(requestad::get_mwversion(*m_ad, good_version));
    if (good_version && mwversion != "") {
      globusrsl.append("(EDG_MW_VERSION '");
      globusrsl.append(mwversion);
      globusrsl.append("')");
    }

    globusrsl.append(")");   
  }

  /* Mandatory */
  /* x509 user proxy is mandatory for the condor submit file. */
  string userproxy(requestad::get_x509_user_proxy(*m_ad));
  if (userproxy.empty()) {
    throw helper::InvalidAttributeValue(requestad::JDLPrivate::USERPROXY,
                                        userproxy,
                                        "not empty",
                                        helper_id);
  }
  requestad::set_x509_user_proxy(*result, userproxy);
  
  /* Mandatory */
  string type(requestad::get_type(*m_ad));
  if (type.empty()) {
    throw helper::InvalidAttributeValue(requestad::JDL::TYPE,
                                        type,
                                        "not empty",
                                        helper_id);
  }
  requestad::set_type(*result, type);
  
  /* Mandatory */
  /* job type is mandatory to build the correct job wrapper file */
  string jobtype(requestad::get_job_type(*m_ad));
  if (jobtype.empty()) {
    throw helper::InvalidAttributeValue(requestad::JDL::JOBTYPE,
                                        jobtype,
                                        "not empty",
                                        helper_id);
  }
  
  /* start preparing JobWrapper file */
  JobWrapper* jw = 0;
   
  /* convert the jobid into filename */
  string jobid_to_file(jobid::to_filename(job_id));

  /* check if there is '/' in the executable */
  if (executable[0] != '/') {
    executable.insert(0, "./", 2);
  }    
	
  // Figure out the local host name (this should really come from
  // some common entity)
  char   hostname[1024];
  string local_host_name("");
  
  if (gethostname(hostname, sizeof(hostname)) >= 0) {
    hostent resolved_host;
    hostent *resolver_result=NULL;
    int resolver_work_buffer_size = 2048;
    char* resolver_work_buffer = new char[resolver_work_buffer_size];
    int resolve_errno = ERANGE;
    while ((gethostbyname_r(hostname, &resolved_host,
                            resolver_work_buffer,
                            resolver_work_buffer_size,
                            &resolver_result,
                            &resolve_errno) ) == ERANGE ) {
      delete[] resolver_work_buffer;
      resolver_work_buffer_size *= 2;
      resolver_work_buffer = new char[resolver_work_buffer_size];
    }
    if (resolver_result != NULL) {
      local_host_name.assign(resolved_host.h_name);
    }
    delete[] resolver_work_buffer;
  }
  
  /* lowercase all jobtype characters */
  string ljobtype(jobtype);
  transform(ljobtype.begin(), ljobtype.end(), ljobtype.begin(), ::tolower); 
  
  if (ljobtype == "mpich") {
    /* Mandatory */	  
    /* lrms type is mandatory for the mpich job */  
    string lrmstype(requestad::get_lrms_type(*m_ad));
    /* lowercase all lrmstype characters */
    string llrmstype(lrmstype);
    std::transform(llrmstype.begin(), llrmstype.end(), llrmstype.begin(), ::tolower);    
    if (llrmstype != "lsf" && llrmstype != "pbs") {
      throw helper::InvalidAttributeValue(requestad::JDL::LRMS_TYPE,
                                          lrmstype,
                                          "lsf or pbs",
                                          helper_id);
    }
    
    /* Mandatory */
    /* node number is mandatory for the mpich job */
    int    nodenumber = requestad::get_node_number(*m_ad);
    string nn(boost::lexical_cast<string>(nodenumber));
    
    globusrsl.append("(count=");
    globusrsl.append(nn);
    globusrsl.append(")(hostCount=");
    globusrsl.append(nn);
    globusrsl.append(")");
    
    if (llrmstype == "lsf") {
      jw = new jobwrapper::MpiLsfJobWrapper(executable);
    }
    else if (llrmstype == "pbs") {
      jw = new jobwrapper::MpiPbsJobWrapper(executable);
    } else {
      // not possible;
    }
    jw->nodes(nodenumber);

  } else if (ljobtype == "interactive") {
    if (find_if(env.begin(), env.end(), 
	Beginning("BYPASS_SHADOW_PORT=")) == env.end())
    {
      std::ostringstream env_str;
      std::copy(env.begin(), env.end(),
                std::ostream_iterator<std::string>(env_str, ":"));
      throw helper::InvalidAttributeValue(requestad::JDL::ENVIRONMENT,
                                          env_str.str(),
                                          "contains BYPASS_SHADOW_PORT",
                                          helper_id);
    }
    if (find_if(env.begin(), env.end(),
	Beginning(string("BYPASS_SHADOW_HOST="))) == env.end())
    {
      std::ostringstream env_str;
      std::copy(env.begin(), env.end(),
                std::ostream_iterator<std::string>(env_str, ":"));
      throw helper::InvalidAttributeValue(requestad::JDL::ENVIRONMENT,
                                          env_str.str(),
                                          "contains BYPASS_SHADOW_HOST",
                                          helper_id);
    }

    string::size_type pos = executable.find("./");
    if (pos == string::npos) {
      jw = new jobwrapper::InteractiveJobWrapper(executable);
    } else {
      jw = new jobwrapper::InteractiveJobWrapper(executable.substr(pos+2));
    }
  } else {
    jw = new jobwrapper::JobWrapper(executable);
  }

  /* Define the VO for the wrapper script, if available */
  string vo(requestad::get_virtual_organisation(*m_ad));
  if (!vo.empty()) {
    jw->vo(vo);
  }

  // Define maximum output sandbox size
  config::WMConfiguration const* wmconfig = config::Configuration::instance()->wm();
  double osbl = wmconfig->max_output_sandbox_size();
  jw->set_max_output_size((int64_t)osbl);

  // Define the LCG application directory, not mandatory
  bool b_lad;
  string app_dir(requestad::get_lcg_application_dir(*m_ad, b_lad));
  if (b_lad && !app_dir.empty()) {
    jw->app_dir(app_dir);
  }
  
  // OutputData is not mandatory
  bool b_od;
  ExprTree* odtree = requestad::get_output_data(*m_ad, b_od);
  string    otptdata("Warning");
  if (b_od) {
    ExprList* odtree_list = dynamic_cast<ExprList*>(odtree);

    if (odtree_list) {
      jw->outputdata(odtree_list);
    } else {
      ClassAd *odtree_classad = dynamic_cast<ClassAd*>(odtree);
      if (odtree_classad) {
        auto_ptr<ExprList> list(new ExprList);
	list->Append(odtree->Copy());
	jw->outputdata(list.get());
      } else {
       throw helper::InvalidAttributeValue(requestad::JDL::OUTPUTDATA,
		                           otptdata,
					   "either list of classad or a classad",
					   helper_id);
      }
    }  
    /* Virtual Organization is mandatory if OutputData is in the JDL */
    if (vo.empty()) {
      throw helper::InvalidAttributeValue(requestad::JDL::VIRTUAL_ORGANISATION,
                                          vo,
                                          "not empty",
                                          helper_id);
    }
    jw->dsupload(url::URL(job_id));
  }

  requestad::set_globus_rsl(*result, globusrsl);
  
  jw->standard_input(stdinput);
  jw->standard_output(stdoutput);
  jw->standard_error(stderror);
  jw->brokerinfo();
  jw->create_subdir();
  jw->arguments(arguments);
  jw->job_Id(job_id);
  jw->job_id_to_filename(jobid_to_file); 
  jw->environment(env);
  jw->brokername(local_host_name);

  // check if we need to use sync running event logging from LRMS
  bool good;
  if (m_ad && requestad::get_edg_sync_running_event(*m_ad, good) && good) {
    jw->set_sync_running_event();
  }
  
  // Get the gatekeeper hostname
  string::size_type pos = globusresourcecontactstring.find(":");
  if (pos == 0 || pos == string::npos) {
    throw helper::InvalidAttributeValue(requestad::JDL::GLOBUSRESOURCE,
                                        globusresourcecontactstring,
                                        "contains a hostname before a :",
                                        helper_id);
  }
  jw->gatekeeper_hostname(globusresourcecontactstring.substr(0, pos));
  
  //check if there is the protocol in the inputsandbox path. 
  //if no the protocol gsiftp:// is added to the inputsandboxpath.
  if (inputsandboxpath.find("://") == string::npos) {
    string new_inputsandboxpath("gsiftp://");
    new_inputsandboxpath.append(local_host_name);
    new_inputsandboxpath.append(inputsandboxpath);
    jw->input_sandbox(url::URL(new_inputsandboxpath), inputsandbox);    
  } else { 
    jw->input_sandbox(url::URL(inputsandboxpath), inputsandbox);
  }

  //check if there is the protocol in the outputsandbox path. 
  //if no the protocol gsiftp:// is added to the outputsandboxpath.
  if (outputsandboxpath.find("://") == string::npos) {
    string new_outputsandboxpath("gsiftp://");     
    new_outputsandboxpath.append(local_host_name);
    new_outputsandboxpath.append(outputsandboxpath);
    jw->output_sandbox(url::URL(new_outputsandboxpath), outputsandbox);
  } else {  
    jw->output_sandbox(url::URL(outputsandboxpath), outputsandbox);
  }
  
  // end preparation JobWrapper file

  config::LMConfiguration const* logconfig = config::Configuration::instance()->lm();
  
  // Mandatory
  // Maradone file path
  string maradonapr(logconfig->maradona_transport_protocol());

  config::NSConfiguration const* nsconfig = config::Configuration::instance()->ns();
  
  string maradona_name(jobid_to_file);
  maradona_name.append("/");
  maradona_name.append("Maradona.output");
  boost::filesystem::path maradona_path(
    boost::filesystem::normalize_path(nsconfig->sandbox_staging_path()),
    boost::filesystem::system_specific);
  maradona_path <<= jobid::get_reduced_part(job_id);
  maradona_path <<= maradona_name;
  jw->maradonaprotocol(maradonapr, maradona_path.file_path());

  config::JCConfiguration const* jwconfig = config::Configuration::instance()->jc();

  // read the submit file path
  string jw_name("JobWrapper.");
  jw_name.append(jobid_to_file);
  jw_name.append(".sh");
 
  boost::filesystem::path jw_path(
    boost::filesystem::normalize_path(jwconfig->submit_file_dir()),
    boost::filesystem::system_specific);
  
  jw_path <<= jobid::get_reduced_part(job_id);
  jw_path <<= jw_name;

  try { 
    boost::filesystem::create_parents(jw_path.branch());
  } catch(boost::filesystem::filesystem_error &err) {
    throw CannotCreateJobWrapper(jw_path.file_path());
  }

  // write the JobWrapper in the submit file path
  ofstream ofJW(jw_path.file_path().c_str());
  if (!ofJW) {
    throw CannotCreateJobWrapper(jw_path.file_path());
  }	  
  ofJW << *jw;

  delete jw;
	
  // The requestad::get_output_data function returns a _copy_ of the object.
  delete odtree;
  
  // Mandatory
  // the new executable value is the jobwrapper file path
  requestad::set_executable(*result, jw_path.file_path());

  // Prepare the output file path
  string output_file_path(jwconfig->output_file_dir());
  output_file_path.append("/");
  output_file_path.append(jobid::get_reduced_part(job_id));
  output_file_path.append("/");
  output_file_path.append(jobid_to_file);
  output_file_path.append("/");

  // New parameter Mandatory
  // Output file path is mandatory
  string output(output_file_path);
  output.append("StandardOutput");
  requestad::set_output(*result, output);

  // New parameter Mandatory
  // Error file path is mandatory
  string error(output_file_path);
  error.append("StandardError");
  requestad::set_error_(*result, error);

  /* Mandatory */
  requestad::set_stream_output(*result, false);
  requestad::set_stream_error(*result, false);

  // Mandatory
  // Condor Log file path
  string condor_file_path(logconfig->condor_log_dir());
  string log(condor_file_path);
  log.append("/");
  log.append("CondorG.log");
  requestad::set_log(*result, log);
  
  return result.release();

} catch (requestad::CannotGetAttribute const& e) {
  throw helper::CannotGetAttribute(e, helper_id);
} catch (requestad::CannotSetAttribute const& e) {
  throw helper::CannotSetAttribute(e, helper_id);
} 

} // namespace jobadapter
} // namespace planning
} // namespace workload
} // namespace edg

