/***************************************************************************
 *  filename  : JobWrapper.cpp
 *  authors   : Alessio Gianelle <alessio.gianelle@pd.infn.it>
 *              Francesco Giacomini <francesco.giacomini@cnaf.infn.it>
 *              Rosario Peluso <rosario.peluso@pd.infn.it>
 *              Elisabetta Ronchieri <elisabetta.ronchieri@cnaf.infn.it>
 *  Copyright (c) 2002 CERN and INFN on behalf of the EU DataGrid.
 *  For license conditions see LICENSE file or
 *  http://www.edg.org/license.html
 ***************************************************************************/

#include <algorithm>
#include <cassert>
#include <string>

#include <classad_distribution.h>
#include "edg/workload/planning/jobadapter/JobWrapper.h"
#include "edg/workload/common/requestad/JobAdManipulation.h"

using namespace std;
using namespace classad;

namespace url = edg::workload::planning::common;
namespace requestad = edg::workload::common::requestad;

namespace edg {
namespace workload {
namespace planning {
namespace jobadapter {

const string JobWrapper::s_brokerinfo_default = ".BrokerInfo";

JobWrapper::JobWrapper(const string& job)
  : m_job(job), m_create_subdir(false), m_sync_running_event(false), m_nodes(0)
{  
}

JobWrapper::~JobWrapper(void)
{
}

void
JobWrapper::standard_input(const string& file)
{
  m_standard_input = file;
}

void
JobWrapper::standard_output(const string& file)
{
  m_standard_output = file;
}

void
JobWrapper::standard_error(const string& file)
{
  m_standard_error = file;
}

void
JobWrapper::input_sandbox(const url::URL& base_url,
                          const vector<string>& files)
{
  m_input_base_url = base_url;
  copy(files.begin(), files.end(), back_inserter(m_input_files));
}

void
JobWrapper::output_sandbox(const url::URL& base_url,
                           const vector<string>& files)
{
  m_output_base_url = base_url;
  copy(files.begin(), files.end(), back_inserter(m_output_files));
}
 
void
JobWrapper::create_subdir(void)
{
  m_create_subdir = true;
}

void
JobWrapper::set_sync_running_event(void)
{
  m_sync_running_event = true;
}

void
JobWrapper::set_max_output_size(const int64_t &max_size)
{
  m_max_size = max_size;
}

void
JobWrapper::brokerinfo(const string& file)
{
  assert(file.find_first_of("/") == string::npos);
  assert(file != "");

  m_brokerinfo = file;
}

void
JobWrapper::job_Id(const string& jobid)
{
  m_jobid = jobid;
}

void
JobWrapper::job_id_to_filename(const string& jobid_to_filename)
{
  m_jobid_to_filename = jobid_to_filename;
}

void
JobWrapper::arguments(const string& args)
{
  m_arguments = args;
}

void
JobWrapper::maradonaprotocol(const string& protocol, const string& filename)
{
  m_maradonaprotocol.assign(protocol);
  m_maradonaprotocol.append(filename); 
}

void
JobWrapper::gatekeeper_hostname(const string& gatekeeper)
{
  m_gatekeeper_hostname = gatekeeper;
}

void 
JobWrapper::environment(const vector<string>& env)
{
  m_environment = env;
}

void
JobWrapper::nodes(int node)
{
  m_nodes = node;
}

void
JobWrapper::outputdata(const ExprList* output_data)
{
  m_outputdata.reset(output_data->Copy());
}

void
JobWrapper::vo(const string& vo)
{
  m_vo = vo;
}

void
JobWrapper::app_dir(const string& app_dir)
{
  m_app_dir = app_dir;
}

void
JobWrapper::brokername(const string& rbname)
{
  m_rbname = rbname;
}

void
JobWrapper::dsupload(const common::URL& id)
{
  m_dsupload.assign("DSUpload_");
  m_dsupload.append(id.path().substr(1));
  m_dsupload.append(".out");
}

ostream&
JobWrapper::set_shell(ostream& os) const
{
  return os << "#!/usr/bin/env bash" << endl 
	    << endl;
}

ostream&
JobWrapper::transfer_output_sb(ostream& os, const int64_t &max_size, const vector<string> &file, const common::URL &returl) const
{
  os        << "sboutcmd='" << endl
            << "use strict;" << endl
            << "use Fcntl \":mode\";" << endl
            << endl
            << "my $max_size = " << max_size << ";" << endl
            << "my @in_files = ( ";

  for (vector<string>::const_iterator it = file.begin();
       it != file.end(); ) {
    std::string s;
    for(size_t i=0;i<it->length();i++) {
      const char c = (*it)[i];
      if (c == '\'') {
        s += "\\'\"'\"'";
      } else if (c == '\\') {
        s += "\\\\";
      } else {
        s += c;
      }
    }
    os << "'\"'\"'" << s << "'\"'\"'";
    ++it;
    if (it != file.end()) {
      os << ", ";
    }
  }

  return os << " );" << endl
            << "my $maradona = \"'$maradona'\";" << endl
            << "my $returl = \"" << returl.as_string() << "\";" << endl
            << "my $workdir = \"'$workdir'\";" << endl
            << "my $max_retries = 6;" << endl
            << "my $user_status = \"'$status'\";" << endl
            << "my @warning_list;" << endl
            << endl
            << "sub doPerlExit {" << endl
            << "  my ($code) = @_;" << endl
            << "  print $ENV{EDG_WL_SEQUENCE_CODE}.\"\\n\";" << endl
            << "  exit($code);" << endl
            << "}" << endl
            << endl
            << "sub log_something {" << endl
            << "  my ($info_only, $text, $status_code, $ret) = @_;" << endl
            << "  if (!$info_only) {" << endl
            << "    print STDERR $text.\"\\n\";" << endl
            << "    if (open(DAT,\">> \".$maradona)) {" << endl
            << "      print DAT $text.\"\\n\";" << endl
            << "      close(DAT);" << endl
            << "    }" << endl
            << "  }" << endl
            << "  $text =~ s/[\\$\\`\\\\\"]/\\\\$&/g;" << endl
            << "  if (defined $status_code && defined $ret && $status_code ne \"\") {" << endl
            << "    if (open(CMD, $ENV{EDG_WL_LOCATION}.\"/bin/edg-wl-logev\"." << endl
            << "                                        \" --jobid=\\\"\".$ENV{EDG_WL_JOBID}.\"\\\"\"." << endl
            << "                                        \" --source=LRMS\"." << endl
            << "                                        \" --sequence=\".$ENV{EDG_WL_SEQUENCE_CODE}." << endl
            << "                                        \" --event=\\\"Done\\\"\"." << endl
            << "                                        \" --reason=\\\"\".$text.\"\\\"\"." << endl
            << "                                        \" --status_code=\".$status_code." << endl
            << "                                        \" --exit_code=\".$user_status.\" 2>/dev/null |\")) {" << endl
            << "      chomp(my $value = <CMD>);" << endl
            << "      close(CMD);" << endl
            << endl
            << "      my $result = $?;" << endl
            << "      my $exit_value = $result >> 8;" << endl
            << endl
            << "      $ENV{EDG_WL_SEQUENCE_CODE} = $value if ($exit_value == 0);" << endl
            << endl
            << "      doPerlExit($ret);" << endl
            << "    } else {" << endl
            << "      doPerlExit($ret);" << endl
            << "    }" << endl
            << "  } else {" << endl
            << "    if (open(CMD, $ENV{EDG_WL_LOCATION}.\"/bin/edg-wl-logev\"." << endl
            << "                                        \" --jobid=\\\"\".$ENV{EDG_WL_JOBID}.\"\\\"\"." << endl
            << "                                        \" --source=LRMS\"." << endl
            << "                                        \" --sequence=\".$ENV{EDG_WL_SEQUENCE_CODE}." << endl
            << "                                        \" --event=\\\"CurDescr\\\"\"." << endl
            << "                                        \" --descr=\\\"\".$text.\"\\\" 2>/dev/null |\")) {" << endl
            << "      chomp(my $value = <CMD>);" << endl
            << "      close(CMD);" << endl
            << endl
            << "      my $result = $?;" << endl
            << "      my $exit_value = $result >> 8;" << endl
            << endl
            << "      $ENV{EDG_WL_SEQUENCE_CODE} = $value if ($exit_value == 0);" << endl
            << "    }" << endl
            << "  }" << endl
            << "}" << endl
            << endl
            << "my @out_files;" << endl
            << "my %sizes;" << endl
            << "my $file;" << endl
            << "my $missing_files=0;" << endl
            << endl
            << "foreach $file (@in_files) {" << endl
            << "  my ($mode, $size) = (stat($file))[2,7];" << endl
            << "  if (!defined $mode) {" << endl
            << "    my $error = \"Did not find output sandbox file $file (ignoring)\";" << endl
            << "    log_something(1,$error);" << endl
            << "    $missing_files++;" << endl
            << "    next;" << endl
            << "  }" << endl
            << "  if (S_ISREG($mode) && -r \"$file\") {" << endl
            << "    $sizes{$file} = $size;" << endl
            << "    push(@out_files,$file);" << endl
            << "  } else {" << endl
            << "    my $error = \"Output sandbox file $file not readable or not a regular file (ignoring)\";" << endl
            << "    log_something(1,$error);" << endl
            << "    $missing_files++;" << endl
            << "  }" << endl
            << "}" << endl
            << endl
            << "my @sorted_files = sort { $sizes{$a} <=> $sizes{$b} } @out_files;" << endl
            << endl
            << "my ($i,$total);" << endl
            << endl
            << "@out_files=();" << endl
            << endl
            << "for($i=0, $total=0;$i<scalar(@sorted_files);$i++) {" << endl
            << "  $file = $sorted_files[$i];" << endl
            << "  last if ($max_size>=0 && $total + $sizes{$file} > $max_size);" << endl
            << "  $total += $sizes{$file};" << endl
            << "  push(@out_files,$file);" << endl
            << "}" << endl
            << endl
            << "my $remaining = scalar(@sorted_files) - $i;" << endl
            << "if ($remaining>0) {" << endl
            << endl
            << "  push(@warning_list,\"some output sandbox file(s) needed truncating\");" << endl
            << "  my $free = $max_size - $total;" << endl
            << "  my $trunc_len = int($free / $remaining);" << endl
            << endl
            << "  for(my $j=$i;$j<scalar(@sorted_files);$j++) {" << endl
            << "    $file = $sorted_files[$j];" << endl
            << "    my $trunc_ok=0;" << endl
            << "    if (open(F,\">> \".$file)) {" << endl
            << "      $trunc_ok=1 if (truncate(F, $trunc_len));" << endl
            << "      close(F);" << endl
            << "    }" << endl
            << "    if ($trunc_ok) {" << endl
            << "      log_something(1,\"Truncated output sandbox file $file from \".$sizes{$file}.\" to $trunc_len bytes\");" << endl
            << "      push(@out_files,$file);" << endl
            << "      $free -= $trunc_len;" << endl
            << "    } else {" << endl
            << "      log_something(1,\"Could not truncate output sandbox file $file, will ignore it\");" << endl
            << "      $missing_files++;" << endl
            << "    }" << endl
            << "    $remaining--;" << endl
            << "    $trunc_len = int($free / $remaining) if $remaining>0;" << endl
            << "  }" << endl
            << "}" << endl
            << endl
            << "if ($missing_files) {" << endl
            << "  push(@warning_list,\"some file(s) listed in the output sandbox were not available and were ignored\");" << endl
            << "}" << endl
            << endl
            << "my $num_tries=0;" << endl
            << "my $wait_time=300;" << endl
            << "my $ret;" << endl
            << endl
            << "for($i=0;$i<scalar(@out_files);$i++) {" << endl
            << "  $file = $out_files[$i];" << endl
            << "  while(1) {" << endl
            << "    my @copy_lines;" << endl
            << "    my @pelem = split(\"/\",$file);" << endl
            << "    my $fflast = $pelem[$#pelem];" << endl
            << "    my $ff = $file;" << endl
            << "    $ff =~ s/[\\$\\`\\\\\"]/\\\\$&/g;" << endl
            << "    $fflast =~ s/[\\$\\`\\\\\"]/\\\\$&/g;" << endl
            << "    if (open(CMD, \"globus-url-copy \\\"file://\".$workdir.\"/\".$ff.\"\\\" \\\"\".$returl.\"/\".$fflast.\"\\\" 2>&1 >/dev/null |\")) {" << endl
            << "      while(<CMD>) {" << endl
            << "        chomp(my $line=$_);" << endl
            << "        $line =~ s/\\r//g;" << endl
            << "        if ($line !~ /^\\s*$/) {" << endl
            << "          push(@copy_lines,$line);" << endl
            << "        }" << endl
            << "      }" << endl
            << "      close(CMD);" << endl
            << "      $ret = $?;" << endl
            << "    } else {" << endl
            << "      $ret = 127<<8;" << endl
            << "    }" << endl
            << "    last if $ret==0;" << endl
            << "    $num_tries++;" << endl
            << "    my $log_message = \"Copy failure ($num_tries of possible \".($max_retries+1).\") to copy output sandbox file $file to $returl\";" << endl
            << "    if (scalar(@copy_lines)) {" << endl
            << "      $log_message .= \":\\n\".join(\"\\n\",@copy_lines).\"\\n\";" << endl
            << "    }" << endl
            << "    log_something(1,$log_message);" << endl
            << "    last if ($num_tries > $max_retries);" << endl
            << endl
            << "    my $time_left = `grid-proxy-info -timeleft 2> /dev/null` || 0;" << endl
            << "    last if ($time_left < $wait_time);" << endl
            << endl
            << "    sleep($wait_time);" << endl
            << "    $wait_time *= 2;" << endl
            << "  }" << endl
            << endl
            << "  if ($ret != 0) {" << endl
            << "    log_something(0,\"Cannot upload $file into $returl\",\"FAILED\",1);" << endl
            << "    # not reached" << endl
            << "  }" << endl
            << "}" << endl
            << endl
            << "if (scalar(@warning_list)) {" << endl
            << "  log_something(0,\"There were some warnings: \".join(\", \",@warning_list),\"OK\",1);" << endl
            << "  # not reached" << endl
            << "} else {" << endl
            << "  doPerlExit(0);" << endl
            << "}" << endl
            << "'" << endl
            << endl
            << "eval 'EDG_WL_SEQUENCE_CODE=`perl -e \"$sboutcmd\"`' 2>&1" << endl
            << "ret=$?" << endl
            << "export EDG_WL_SEQUENCE_CODE" << endl
            << "unset sboutcmd" << endl
            << endl
            << "if [ $ret -ne 0 ]; then" << endl
            << "  doExit $ret" << endl
            << "fi" << endl
            << endl;
}

ostream&
JobWrapper::set_cleanup_globals(ostream& os) const
{
  return os << "uid=`id -u`" << endl
            << "max_days=7" << endl
            << "cutoff=`expr $max_days + 1`" << endl
            << "gnufind=find" << endl
            << "myhost=`hostname | sed -e 's/\\..*//' | tr '[A-Z]' '[a-z]'`" << endl
            << "user=`whoami`" << endl
            << "test \"x$user\" = x && user=`id -nu`" << endl
            << endl;
}

ostream&
JobWrapper::write_cleanup_subr(ostream& os) const
{
  return os << "cleanupDir()" << endl
            << "{" << endl
            << "  (" << endl
            << "    cd \"${1:-.}\" || exit" << endl
            << endl
            << "    oldpat=https_3a_2f_2f*" << endl
            << endl
            << "    for dir in `" << endl
            << "        $gnufind . -noleaf -maxdepth 1 -type d -name \"$oldpat\" \\" << endl
            << "          -uid $uid -ctime +$max_days -print" << endl
            << "      `" << endl
            << "    do" << endl
            << "      case \"$dir\" in" << endl
            << "      ./$oldpat)" << endl
            << "        # safety check on the name..." << endl
            << "        changes=`" << endl
            << "          $gnufind \"$dir\" -noleaf -ctime -$cutoff -print" << endl
            << "        `" << endl
            << "        if [ -z \"$changes\" ]; then" << endl
            << "          rm -rf \"$dir\"" << endl
            << "        fi" << endl
            << "      esac" << endl
            << "    done" << endl
            << endl
            << "    newpat=WMS_*_[0-9]*[0-9]_*" << endl
            << endl
            << "    for dir in `" << endl
            << "        $gnufind . -noleaf -maxdepth 1 -type d -name \"$newpat\" -uid $uid -print" << endl
            << "      `" << endl
            << "    do" << endl
            << "      case \"$dir\" in" << endl
            << "      ./$newpat)" << endl
            << "        # safety check on the name..." << endl
            << "        dirhost=`echo \"$dir\" | sed -e 's|^\\./||' -e 's|^WMS_||' -e 's|_.*||'`" << endl
            << "        dirproc=`echo \"$dir\" | sed -e 's|^\\./||' -e 's|^WMS_||' -e 's|[^_]*_||' -e 's|_.*||'`" << endl
            << endl
            << "        if [ \"x$dirhost\" = \"x$myhost\" ]; then" << endl
            << "          kill -0 $dirproc 2> /dev/null" << endl
            << "          if [ $? -ne 0 ]; then" << endl
            << "            rm -rf \"$dir\"" << endl
            << "          fi" << endl
            << "        fi" << endl
            << "      esac" << endl
            << "    done" << endl
            << "  )" << endl
            << "}" << endl
            << endl;
}

ostream&
JobWrapper::set_scratchdir(ostream& os, bool create_subdir) const
{
  os << "if [ ! -z \"$EDG_WL_SCRATCH\" ]; then" << endl
     << "  cd $EDG_WL_SCRATCH" << endl;

  if (create_subdir)
    os << "  cleanupDir" << endl;
  
  return os << "fi" << endl
            << endl;
}

ostream&
JobWrapper::set_subdir(ostream& os, 
		          const string& dir) const
{
  os << "newdir=WMS_${myhost}_0$$_" << dir << endl
     << endl;

  return os << "initdir=`pwd`" << endl
            << "mkdir ${newdir}" << endl
            << "cd ${newdir}" << endl 
            << endl;	 
}

ostream&
JobWrapper::clean_subdir(ostream& os) const
{
  return os << "cd ${initdir}" << endl
            << "rm -rf ${newdir}" << endl 
	    << endl;
}

ostream&
JobWrapper::check_workdir(ostream& os) const
{
  os << "if [ ! -w . ]; then" << endl
     << "  echo \"Working directory not writable\"" << endl << endl;
  
  os << " "; this->set_lb_sequence_code(os, 
		  "Done", 
		  "Working directory not writable!",
		  "FAILED", 
		  "127");
  
  return os << "  exit 1" << endl
	    << "fi" << endl 
	    << endl;
}

ostream&
JobWrapper::set_workdir(ostream& os) const
{
  return os << "workdir=\"`pwd`\"" << endl
	    << endl;
}

ostream&
JobWrapper::set_environment(ostream& os, 
		            const vector<string>& env) const
{
  for (vector<string>::const_iterator it = m_environment.begin();
      it != m_environment.end(); it++) {
    os << "export " << *it << endl;
  }
  os << endl;
  return os;        	
}

ostream&
JobWrapper::set_environment(ostream& os,
		            const string& variable,
			    const string& value) const
{
  return os << "export " << variable << "=\"" << value << "\"" << endl;
}

ostream&
JobWrapper::check_set_environment(ostream& os,
		                  const string& variable,
		                  const string& value) const
{
  os << "if [ -z \"${" << variable << "}\" ]; then" << endl;
  os << "  "; this->set_environment(os, variable, value); 
  return os << "fi" << endl << endl;
}

ostream&
JobWrapper::set_umask(ostream& os) const
{
  return os << "umask 022" << endl
            << endl;
}

ostream&
JobWrapper::check_file_mod(ostream& os,
		           const string& file) const
{
  os << "if [ -e \"" << file << "\" ]; then" << endl
     << "  chmod +x \"" << m_job << "\" 2> /dev/null" << endl
     << "else" << endl
     << "  echo \"" << file << " not found or unreadable\"" << endl
     << "  echo \"" << file << " not found or unreadable\" >> \"${maradona}\"" << endl << endl;

  string message(file);
  message.append(" not found or unreadable!");
  
  os << "  "; this->set_lb_sequence_code(os, 
		  "Done",
		  message,
                  "FAILED",
                  "127");
  	    
  return os << "  doExit 1" << endl
	    << "fi" << endl 
	    << endl;
}

ostream&
JobWrapper::set_lb_sequence_code(ostream& os,
                                 const string& event,
				 const string& reason,
				 const string& scode,
				 const string& ecode) const
{
  os << "EDG_WL_SEQUENCE_CODE=`$EDG_WL_LOCATION/bin/edg-wl-logev" << " \\" << endl
     << " --jobid=" << "\"" << "$EDG_WL_JOBID" << "\"" << " \\" << endl
     << " --source=LRMS" << " \\" << endl
     << " --sequence=" << "\"" << "$EDG_WL_SEQUENCE_CODE" << "\"" << "\\" << endl
     << " --event=" << "\"" << event << "\"" << "\\" << endl;

     if (!event.compare("Running")) {
       if (m_sync_running_event) {
         os << " --priority";
       }
       os << " --node=$host" << "\\" << endl;
     } else if (!scode.compare("FAILED")) {
       os << " --reason=" << "\"" << reason << "\"" << "\\" << endl
          << " --status_code=" << scode << "\\" << endl
          << " --exit_code=" << ecode << "\\" << endl;
     } else {
       os << " --status_code=" << scode << "\\" << endl
	  << " --exit_code=" << ecode << "\\" << endl;
     }

  os << " || (ret=$?; echo $EDG_WL_SEQUENCE_CODE; exit $ret)`" << endl;

  if (!event.compare("Running") && m_sync_running_event) {
    os << "if [ $? -ne 0 ]; then" << endl
       << "  echo \"Cannot confirm that the logging service recorded that the job is about to run\"" << endl
       << "  echo \"Cannot confirm that the logging service recorded that the job is about to run\" >> \"${maradona}\"" << endl << endl
       << "  doExit 1" << endl
       << "fi" << endl;
  }

  return os << "export EDG_WL_SEQUENCE_CODE" << endl
	    << endl;
}

ostream& 
JobWrapper::check_globus(std::ostream& os) const
{
  os << "if [ -z \"${GLOBUS_LOCATION}\" ]; then" << endl
     << "  echo \"GLOBUS_LOCATION undefined\"" << endl
     << "  echo \"GLOBUS_LOCATION undefined\" >> \"${maradona}\"" << endl << endl;

  os << "  "; this->set_lb_sequence_code(os,
		  "Done",
		  "GLOBUS_LOCATION undefined",
                  "FAILED",
                  "127"); 

  os << "  doExit 1" << endl
     << "elif [ -r \"${GLOBUS_LOCATION}/etc/globus-user-env.sh\" ]; then" << endl
     << "  . ${GLOBUS_LOCATION}/etc/globus-user-env.sh" << endl
     << "else" << endl
     << "  echo \"${GLOBUS_LOCATION}/etc/globus-user-env.sh not found or unreadable\"" << endl
     << "  echo \"${GLOBUS_LOCATION}/etc/globus-user-env.sh not found or unreadable\" >> \"${maradona}\"" << endl << endl;
  
  os << "  "; this->set_lb_sequence_code(os,
		  "Done",
		  "${GLOBUS_LOCATION}/etc/globus-user-env.sh not found or unreadable",
		  "FAILED",
		  "127");
  
  return os << "  doExit 1" << endl
	    << "fi" << endl 
	    << endl;
}

ostream&
JobWrapper::prepare_transfer(ostream& os,
		             const vector<string>& file) const
{
  os << "for f in";
  for (vector<string>::const_iterator it = file.begin();
       it != file.end(); ++it) {
    os << " \"" << *it << "\"";
  }
  return os << "; do" << endl;
}

ostream&
JobWrapper::make_transfer(ostream& os,
		          const common::URL& prefix,
			  const bool& input) const
{
  // source and destination must be '/' terminated
  string value = prefix.as_string();
  if (value[value.size() - 1] != '/') {
    value.append( 1, '/' );
  }

  if (input) {
    os << "  globus-url-copy " << value << "${f}" << " file://${workdir}/${f}" <<
       endl
       << "  if [ $? != 0 ]; then" << endl
       << "    echo \"Cannot download ${f} from " << value << "\"" << endl
       << "    echo \"Cannot download ${f} from " << value << "\" >> \"${maradona}\"" << endl << endl;

    string imessage("Cannot download ${f} from ");
    imessage.append(value);
    
    os << "    "; this->set_lb_sequence_code(os,
		    "Done",
		    imessage,
		    "FAILED",
		    "127");

    os << "    doExit 1" << endl;
  } else {
    os << "  if [ -r \"${f}\" ]; then" << endl;
        
    os << "    output=`dirname $f`" << endl
       << "    if [ \"x${output}\" = \"x.\" ]; then" << endl
       << "      ff=$f" << endl
       << "    else" << endl
       << "      ff=${f##*/}" << endl
       << "    fi" << endl;
	    
    os << "    globus-url-copy file://${workdir}/${f} " << value << "${ff}" << endl
       << "    if [ $? != 0 ]; then" << endl
       << "      echo \"Cannot upload ${f} into " << value << "\"" << endl
       << "      echo \"Cannot upload ${f} into " << value << "\" >> \"${maradona}\"" << endl << endl;

    string omessage("Cannot upload ${f} into ");
    omessage.append(value);
    
    os << "      "; this->set_lb_sequence_code(os,
		       "Done",
		       omessage,
		       "FAILED",
		       "1");

    os << "      doExit 1" << endl
       << "    fi" << endl;
  }
  return os << "  fi" << endl
	    << "done" << endl 
	    << endl;
}

ostream& 
JobWrapper::make_bypass_transfer(ostream& os,
				 const string& prefix) const
{
  return os;
}

ostream&
JobWrapper::execute_job(ostream& os,
		        const string& arguments,
			const string& job,
			const string& stdi,
			const string& stdo,
			const string& stde,
			int   node) const
{
  os << "(" << endl
     << "  perl -e '" << endl
     << "    unless (defined($ENV{\"EDG_WL_NOSETPGRP\"})) {" << endl
     << "      $SIG{\"TTIN\"} = \"IGNORE\";" << endl
     << "      $SIG{\"TTOU\"} = \"IGNORE\";" << endl
     << "      setpgrp(0, 0);" << endl
     << "    }" << endl
     << "    exec(@ARGV);" << endl
     << "    warn \"could not exec $ARGV[0]: $!\\n\";" << endl
     << "    exit(127);" << endl
     << "  ' ";

  if (arguments != "") {
    os << "\"" << job << "\"" << " " << arguments << " $*";
  }
  else {
    os << "\"" << job << "\" $*";
  }
  if (stdi != "") {
    os << " < \"" << stdi << "\"";
  } else {
    os << " < /dev/null ";
  }

  if (stdo != "") {
    os << " > \"" << stdo << "\"";
  } else {
    os << " > /dev/null ";
  }

  if (stdo == stde) {
    os << " 2>&1";
  } else {
    if (stde != "") {
      os << " 2> \"" << stde << "\"";
    } else {
      os << " 2> /dev/null";
    }
  }

  return os << " &" << endl
            << endl
            << "  user_job=$!" << endl
            << endl
            << "  exec 2> /dev/null" << endl
            << endl
            << " ( ulimit -c 0; ulimit -d 52428800; ulimit -m 52428800; ulimit -v 52428800; ulimit -t 600; exec nice perl -e '" << endl
            << "    my @pgroups = (getpgrp(0));" << endl
            << "    if (!defined($ENV{\"EDG_WL_NOSETPGRP\"})) { push(@pgroups,'\"$user_job\"'); }" << endl
            << "    my $start_time = time();" << endl
            << "    my $localid=\"UNAVAILABLE\";" << endl
            << "    if (defined($ENV{\"QSUB_REQID\"})) { $localid=$ENV{\"QSUB_REQID\"} };" << endl
            << "    if (defined($ENV{\"PBS_JOBID\"})) { $localid=$ENV{\"PBS_JOBID\"} };" << endl
            << "    if (defined($ENV{\"LSB_JOBID\"})) { $localid=$ENV{\"LSB_JOBID\"} };" << endl
            << endl
            << "    my @prog = ($ENV{\"LCG_LOCATION\"}.\"/bin/lcg-mon-wn\"," << endl
            << "                \"-j\",$ENV{\"EDG_WL_JOBID\"}," << endl
            << "                \"-p\",$ENV{\"X509_USER_PROXY\"}," << endl
            << "                \"-l\",\"$localid\"," << endl
            << "                \"-v\",\"" << m_vo << "\"," << endl
            << "                \"-r\",\"" << m_rbname << "\"," << endl
            << "                @pgroups);" << endl
            << endl
            << "    if ( defined($ENV{\"LCG_LOCATION\"}) && -x $prog[0] ) {" << endl
            << "      $SIG{INT} = \"IGNORE\";" << endl
            << "      $SIG{QUIT} = \"IGNORE\";" << endl
            << "      open STDOUT, \">>/dev/null\";" << endl
            << "      open STDERR, \">&STDOUT\";" << endl
            << "      open STDIN, \"</dev/null\";" << endl
            << "      exec { $prog[0] } @prog;" << endl
            << "    }" << endl
            << "    exit(127);" << endl
            << "  ' ) &" << endl
            << endl
            << "  reporter=$!" << endl
            << endl
            << "  perl -e '" << endl
            << "    while (1) {" << endl
            << "      $time_left = `grid-proxy-info -timeleft 2> /dev/null` || 0;" << endl
            << "      last if ($time_left <= 0);" << endl
            << "      sleep($time_left);" << endl
            << "    }" << endl
            << "    kill(defined($ENV{\"EDG_WL_NOSETPGRP\"}) ? 9 : -9, '\"$user_job\"');" << endl
            << "    exit(1);" << endl
            << "  ' &" << endl
            << endl
            << "  watchdog=$!" << endl
            << endl
            << "  wait $user_job" << endl
            << "  status=$?" << endl
            << endl
            << "  kill $reporter" << endl
            << endl
            << "  for i in 1 2 3 4 5; do" << endl
            << "    kill -n 0 $reporter || break" << endl
            << "    sleep 1" << endl
            << "  done" << endl
            << endl
            << "  /bin/kill -9 $reporter $watchdog $user_job -$user_job" << endl
            << endl
            << "  exit $status" << endl
            << ")" << endl
            << endl;
}

ostream&
JobWrapper::create_maradona_file(ostream& os, 
		                 const string& jobid_to_filename) const
{
  // create the maradona file
  return os << "maradona=\".maradona." << jobid_to_filename 
	    << ".output\"" << endl
            << "touch \"${maradona}\"" << endl 
            << endl;	
}

ostream&
JobWrapper::doExit(ostream& os, 
		   const string& maradonaprotocol, 
		   bool create_subdir) const
{
  // The doExit function used to copy the maradona file, too...
  os << "doExit()" << endl << "{" << endl
     << "  stat=$1" << endl << endl
     << "  globus-url-copy \"file://${workdir}/${maradona}\" "
     << "\"" << maradonaprotocol << "\"" << endl << endl;

  // create subdirectory and cd into it if requested
  if (create_subdir) {
    clean_subdir(os);
  }

  os << "  exit $stat" << endl
     << "}" << endl 
     << endl;

  return os;  
}

ostream&
JobWrapper::handle_output_data(ostream& os) const
{
  ExprList::const_iterator it;
  bool check;
  
  os << "return_value=0" << endl;
  
  os << "if [ $status -eq 0 ]; then" << endl 
     << "  local=`doDSUploadTmp`" << endl
     << "  status=$?" << endl
     << "  return_value=$status" << endl
     << endl;
  
  for (it = m_outputdata->begin(); it != m_outputdata->end(); ++it) {
    ClassAd* ad = dynamic_cast<ClassAd*>(*it);
    if (ad != 0) {
      string outputfile = requestad::get_output_file(*ad);
      string lfn        = requestad::get_logical_file_name(*ad, check);
      string se         = requestad::get_storage_element(*ad, check);
      
      os << "  local=`doCheckReplicaFile " << outputfile << "`" << endl 
         << "  status=$?" << endl
	 << "  if [ $status -ne 0 ]; then" << endl
	 << "    return_value=1" << endl
	 << "  else" << endl;

      if (lfn.empty() && se.empty()) {
	os << "    local=`doReplicaFile " << outputfile << "`" << endl;
      }
      else if ((!lfn.empty()) && se.empty()) {
	os << "    local=`doReplicaFilewithLFN " << outputfile 
	   << " " << lfn << "`" << endl;
      } 
      else if (lfn.empty() && (!se.empty())) {
        os << "    local=`doReplicaFilewithSE " << outputfile
           << " " << se << "`" << endl;
      }
      else {
	os << "    local=`doReplicaFilewithLFNAndSE " << outputfile
           << " " << lfn << " " << se << "`" << endl;
      }
    
      os << "    status=$?" << endl
         << "  fi" << endl
         << endl;
    }
  }	

  os << "  local=`doDSUpload`" << endl
     << "  status=$?" << endl
     << "fi" << endl 
     << endl;

  return os;
}

ostream&
JobWrapper::doDSUploadTmp(ostream& os) const
{
  os << "doDSUploadTmp()" << endl
     << "{" << endl
     << "  filename=" << m_dsupload << endl
     << "  echo \"#\" >> $filename.tmp" << endl
     << "  echo \"# Autogenerated by JobWrapper!\" >> $filename.tmp" << endl
     << "  echo \"#\" >> $filename.tmp" << endl
     << "  echo \"# The file contains the results of the upload and registration\" >> $filename.tmp" << endl
     << "   echo \"# process in the following format:\" >> $filename.tmp" << endl
     << "   echo \"# <outputfile> <lfn|guid|Error>\" >> $filename.tmp" << endl
     << "   echo \"\" >> $filename.tmp" << endl
     << "}" << endl
     << endl;

  return os;
}

ostream&
JobWrapper::doDSUpload(ostream& os) const
{
  os << "doDSUpload()" << endl
     << "{" << endl
     << "  filename=" << m_dsupload << endl
     << "  mv -fv $filename.tmp $filename" << endl
     << "}" << endl
     << endl;

  return os;
}

ostream&
JobWrapper::doCheckReplicaFile(ostream& os) const
{
  os << "doCheckReplicaFile()" << endl
     << "{" << endl
     << "  sourcefile=$1" << endl
     << "  filename=" << m_dsupload << endl
     << "  exit_status=0" << endl << endl
     << "  if [ ! -f \"${workdir}/$sourcefile\" ]; then" << endl
     << "    echo \"$sourcefile    Error: File $sourcefile has not been found on the WN $host\" >> $filename.tmp" << endl
     << "    exit_status=1" << endl
     << "  fi" << endl << endl
     << "  echo \"\" >> $filename.tmp" << endl
     << "  return $exit_status" << endl
     << "}" << endl
     << endl;

  return os;
}

ostream&
JobWrapper::doReplicaFile(ostream& os) const
{
  os << "doReplicaFile()" << endl
     << "{" << endl
     << "  sourcefile=$1" << endl
     << "  filename=" << m_dsupload << endl
     << "  exit_status=0" << endl 
     << endl;

  os << "  local=`$EDG_WL_LOCATION/bin/edg-rm --vo=" << m_vo 
     << " copyAndRegisterFile file://${workdir}/$sourcefile 2>&1`" << endl
     << "  result=$?" << endl
     << "  if [ $result -eq 0 ]; then" << endl
     << "    echo \"$sourcefile    $local\" >> $filename.tmp" << endl
     << "  else" << endl
     << "    echo \"$sourcefile    Error: $local\" >> $filename.tmp" << endl
     << "    exit_status=1" << endl
     << "  fi" << endl 
     << endl;
  
  os << "  echo \"\" >> $filename.tmp" << endl
     << "  return $exit_status" << endl
     << "}" << endl
     << endl;

  return os;  
}

ostream&
JobWrapper::doReplicaFilewithLFN(ostream& os) const
{
  os << "doReplicaFilewithLFN()" << endl
     << "{" << endl
     << "  sourcefile=$1" << endl
     << "  lfn=$2" << endl
     << "  filename=" << m_dsupload << endl
     << "  exit_status=0" << endl 
     << endl;
  
  os << "  local=`$EDG_WL_LOCATION/bin/edg-rm --vo=" << m_vo
     << " copyAndRegisterFile file://${workdir}/$sourcefile -l $lfn 2>&1`" << endl
     << "  result=$?" << endl
     << "  if [ $result -eq 0 ]; then" << endl
     << "    echo \"$sourcefile    $lfn\" >> $filename.tmp" << endl
     << "  else" << endl
     << "    localnew=`$EDG_WL_LOCATION/bin/edg-rm --vo=" << m_vo
     << " copyAndRegisterFile file://${workdir}/$sourcefile 2>&1`" << endl
     << "    result=$?" << endl
     << "    if [ $result -eq 0 ]; then" << endl
     << "      echo \"$sourcefile    $localnew\" >> $filename.tmp" << endl
     << "    else" << endl
     << "      echo \"$sourcefile    Error: $local; $localnew\" >> $filename.tmp" << endl
     << "      exit_status=1" << endl
     << "    fi" << endl
     << "  fi" << endl 
     << endl;
  
  os << "  echo \"\" >> $filename.tmp" << endl
     << "  return $exit_status" << endl
     << "}" << endl
     << endl;

  return os;
}

ostream&
JobWrapper::doReplicaFilewithSE(ostream& os) const
{
  os << "doReplicaFilewithSE()" << endl
     << "{" << endl
     << "  sourcefile=$1" << endl
     << "  se=$2" << endl
     << "  filename=" << m_dsupload << endl
     << "  exit_status=0" << endl
     << endl;

  os << "  local=`$EDG_WL_LOCATION/bin/edg-rm --vo=" << m_vo
     << " copyAndRegisterFile file://${workdir}/$sourcefile -d $se 2>&1`" << endl
     << "  result=$?" << endl
     << "  if [ $result -eq 0 ]; then" << endl
     << "    echo \"$sourcefile    $local\" >> $filename.tmp" << endl
     << "  else" << endl
     << "    localnew=`$EDG_WL_LOCATION/bin/edg-rm --vo=" << m_vo
     << " copyAndRegisterFile file://${workdir}/$sourcefile 2>&1`" << endl
     << "    result=$?" << endl
     << "    if [ $result -eq 0 ]; then" << endl
     << "      echo \"$sourcefile    $localnew\" >> $filename.tmp" << endl
     << "    else" << endl
     << "      echo \"$sourcefile    Error: $local; $localnew\" >> $filename.tmp" << endl
     << "      exit_status=1" << endl
     << "    fi" << endl
     << "  fi" << endl
     << endl;

  os << "  echo \"\" >> $filename.tmp" << endl
     << "  return $exit_status" << endl
     << "}" << endl
     << endl;

  return os;
}

ostream&
JobWrapper::doReplicaFilewithLFNAndSE(ostream& os) const
{
  os << "doReplicaFilewithLFNAndSE()" << endl
     << "{" << endl
     << "  sourcefile=$1" << endl
     << "  lfn=$2" << endl
     << "  se=$3" << endl
     << "  filename=" << m_dsupload << endl
     << "  exit_status=0" << endl
     << endl;

  os << "  local=`$EDG_WL_LOCATION/bin/edg-rm --vo=" << m_vo
     << " copyAndRegisterFile file://${workdir}/$sourcefile -l $lfn -d $se 2>&1`" << endl
     << "  result=$?" << endl
     << "  if [ $result -eq 0 ]; then" << endl
     << "    echo \"$sourcefile    $lfn\" >> $filename.tmp" << endl
     << "  else" << endl
     << "    localse=`$EDG_WL_LOCATION/bin/edg-rm --vo=" << m_vo
     << " copyAndRegisterFile file://${workdir}/$sourcefile -d $se 2>&1`" << endl
     << "    result=$?" << endl
     << "    if [ $result -eq 0 ]; then" << endl
     << "      echo \"$sourcefile    $localse\" >> $filename.tmp" << endl
     << "    else" << endl
     << "      locallfn=`$EDG_WL_LOCATION/bin/edg-rm --vo=" << m_vo
     << " copyAndRegisterFile file://${workdir}/$sourcefile -l $lfn 2>&1`" << endl
     << "      result=$?" << endl
     << "      if [ $result -eq 0 ]; then" << endl
     << "        echo \"$sourcefile    $locallfn\" >> $filename.tmp" << endl
     << "      else" << endl
     << "        localnew=`$EDG_WL_LOCATION/bin/edg-rm --vo=" << m_vo
     << " copyAndRegisterFile file://${workdir}/$sourcefile 2>&1`" << endl
     << "        result=$?" << endl
     << "        if [ $result -eq 0 ]; then" << endl
     << "          echo \"$sourcefile    $localnew\" >> $filename.tmp" << endl
     << "        else" << endl
     << "          echo \"$sourcefile    Error: $local; $localse; $locallfn; $localnew\" >> $filename.tmp" << endl
     << "          exit_status=1" << endl
     << "        fi" << endl   
     << "      fi" << endl
     << "    fi" << endl
     << "  fi" << endl
     << endl;
	  
  os << "  echo \"\" >> $filename.tmp" << endl
     << "  return $exit_status" << endl
     << "}" << endl
     << endl;

  return os;  
}
		    

ostream&
JobWrapper::print(ostream& os) const
{
  // set shell	
  set_shell(os);

  if (!m_app_dir.empty() && m_app_dir[0] == '/') {
    os << "script=\"" << m_app_dir << "/dteam/lcg-jobwrapper-hook.sh\"" << endl
       << "script2=\"" << m_app_dir << "/lcg-jobwrapper-hook.sh\"" << endl
       << "if [ -r \"$script\" ]; then" << endl
       << "  . \"$script\"" << endl
       << "elif [ -r \"$script2\" ]; then" << endl
       << "  . \"$script2\"" << endl
       << "fi" << endl
       << "unset script script2" << endl << endl;
  }

  // write the doExit function
  doExit(os, m_maradonaprotocol, m_create_subdir);

  // write globals for cleanupDir, and the subroutine itself if we're going to use it
  if (m_create_subdir) {
    set_cleanup_globals(os);
    write_cleanup_subr(os);
  }

  // write the doReplicaFile function
  if (m_outputdata != 0) {
    doDSUploadTmp(os);
    doDSUpload(os);
    doCheckReplicaFile(os);
    doReplicaFile(os);
    doReplicaFilewithLFN(os);
    doReplicaFilewithSE(os);
    doReplicaFilewithLFNAndSE(os);
  }
	
  // check and set the environment variable EDG_WL_LOG_DESTINATION with the 
  // name of the gatekeeper node
  if (m_gatekeeper_hostname != "") {
    check_set_environment(os, "EDG_WL_LOG_DESTINATION", m_gatekeeper_hostname);
  }
  // set the environment variable EDG_WL_JOBID
  if (m_jobid != "") {
    set_environment(os, "EDG_WL_JOBID", m_jobid);
    os << endl;
  }
  // set the environment variable EDG_WL_SEQUENCE_CODE
  set_environment(os, "EDG_WL_SEQUENCE_CODE", "$1");
  os << "shift" << endl << endl;
  // check and set the environment variable EDG_WL_LOCATION
  string edg_location_value("${EDG_LOCATION:-/opt/edg}");
  check_set_environment(os, "EDG_WL_LOCATION", edg_location_value);

  // do the cleanupDir here, if we're using it
  if (m_create_subdir) {
    os << "cleanupDir" << endl
       << "eval cleanupDir ~$user" << endl
       << endl;
  }

  // change to scratch directory if requested, plus optionally clean it
  set_scratchdir(os, m_create_subdir);

  // create subdirectory and cd into it if requested
  if (m_create_subdir) {
    set_subdir(os, m_jobid_to_filename);
  }
  // check if the working directory is writable and set it
  check_workdir(os);
  set_workdir(os);

  // set the environment variable EDG_WL_RB_BROKERINFO
  if (m_brokerinfo != "") {
    string broker_info_value("`pwd`");
    broker_info_value.append("/");
    broker_info_value.append(m_brokerinfo);
    set_environment(os, "EDG_WL_RB_BROKERINFO", broker_info_value);
    os << endl;
  }
  
  // Handling Maradona
  create_maradona_file(os, m_jobid_to_filename);

  // check for failed middleware version selection
  os << "echo ${EDG_MW_VERSION} | grep '^not_found:' >/dev/null 2>&1" << endl
     << "if [ $? -eq 0 ]; then" << endl
     << "  reqver=`echo ${EDG_MW_VERSION} | sed -e 's/^not_found://'`" << endl
     << "  echo \"Requested middleware version not found: $reqver\"" << endl
     << "  echo \"Requested middleware version not found: $reqver\" >> \"${maradona}\"" << endl
     << "  " << endl;
  this->set_lb_sequence_code(os,
		"Done", 
		"Requested middleware version not found: $reqver",
		"FAILED", 
		"127");
  os << "  doExit 1" << endl
     << "fi" << endl << endl;

  // check globus
  check_globus(os);
  
  // set the jdl environment variables
  set_environment(os, m_environment);

  // set the umask
  set_umask(os);
 
  // transfer input sandbox
  prepare_transfer(os, m_input_files);
  make_transfer(os, m_input_base_url, true);
  
  // check the file mod
  check_file_mod(os, m_job);
  
  // It is used for the interactive job
  // Extract host
  string host(m_input_base_url.protocol());
  host.append("://");
  host.append(m_input_base_url.host());
  host.append("/");
  make_bypass_transfer(os, host);
  
  os << "host=`hostname -f`" << endl;
  // update the environment variable EDG_WL_SEQUENCE_CODE
  set_lb_sequence_code(os, "Running", "The job starts running!", "", "");

  // execute job
  execute_job(os, m_arguments, m_job, 
	      m_standard_input, m_standard_output, m_standard_error,
	      m_nodes);

  // log job return code to JSS
  os << "status=$?" << endl;

  // replica the outputdata
  if (m_outputdata != 0) {	  
    handle_output_data(os);
  }
	
  os << "echo \"job exit status = ${status}\"" << endl; 
  os << "echo \"job exit status = ${status}\" >> \"${maradona}\"" << endl
     << endl; 
  
  // transfer output sandbox
  transfer_output_sb(os, m_max_size, m_output_files, m_output_base_url);

  // update the environment variable EDG_WL_SEQUENCE_CODE
  set_lb_sequence_code(os, 
		  "Done", 
		  "The job finished correctly!", 
		  "OK", 
		  "$status");
  
  // Copy back the Maradona file
  os << "doExit 0" << endl;

  return os;
}

std::ostream&
operator<<(std::ostream& os, const JobWrapper& jw)
{
  return jw.print(os);
}

} // namespace jobadapter
} // namespace planning
} // namespace workload
} // namespace edg
