/***************************************************************************
 *  filename  : LDAPSynchConnection.h
 *  authors   : Salvatore Monforte <salvatore.monforte@ct.infn.it>
 *  copyright : (C) 2001, 2002 by INFN
 ***************************************************************************/

// $Id:

/**
 * @file LDAPSynchConnection.h
 * @brief This files contains definitions for LDAP Synchronous connections.
 * @author Salvatore Monforte salvatore.monforte@ct.infn.it
 * @author comments by Marco Pappalardo marco.pappalardo@ct.infn.it and Salvatore Monforte
 */

#ifndef _LDAPSYNCHCONNECTION_
#define _LDAPSYNCHCONNECTION_ 

#include <map>
#include <string>
#include <vector>
#include <sys/time.h>
#include <ldap.h>
#include <lber.h>

#ifdef WITH_LDAP_EXCEPTIONS
#include "edg/workload/common/ldif2classad/exceptions.h"
#endif 

/** Includes the superclass's definitions. */
#include "edg/workload/common/ldif2classad/LDAPConnection.h"
#include "edg/workload/common/ldif2classad/LDIFObject.h"

namespace edg {
namespace workload {
namespace common {
namespace ldif2classad {

typedef	std::pair<std::string,std::string > UnknownAttrPair;
typedef std::map<std::string, std::vector<UnknownAttrPair> > UnknownAttrMap;
typedef std::map<std::string, LDIFObject> LDIFObjectMap;
typedef std::vector<std::string> DNVector;
typedef DNVector::const_iterator DNIterator;

/**
 * Concrete LDAP result entry.
 * This class specialises the standard interface 
 * for concrete LDAPConnectionSynchConnection search result entries.
 */
class ldap_result_entry_t : public generic_result_entry_t
{
public:
  /**
   * Costructor
   * @param ldentry the result  as  returned  by  a  call  to ldap_result or ldap_search_s or ldap_search_st
   * @param ld the andle to ldap connection.
   */
  ldap_result_entry_t(const DNVector& dns,
                      const UnknownAttrMap& undefined_attrs,
                      const LDIFObjectMap& objects)
  {
    m_dns = &dns;
    m_undefined_attrs = &undefined_attrs;
    m_objects = &objects;

    m_itr = m_dns->begin();
  }  

  /**
   * Concrete next entry assignment.
   * @return whether the next entry has been succesfully addressed, or not.
   */
  bool next()
  {
    if( good() ) {
      m_itr++;
      return ( m_itr != m_dns->end() );
    }
    return false;
  }
  /**
   * Concrete validation test
   * @return whether the entry is valid or not.
   */
  bool good () const 
  {
    return ( m_itr != m_dns->end() );
  }
  /**
   * Concrete distinguished name extractor.
   * @return a string representation of entry's distinguished name
   */
  std::string distinguished_name() const 
  {
    if (good())
      return *m_itr;

    return std::string();
  }
  /**
   * Concrete LDIFObject extractor.
   * @return a poiter to a LDIFObject.
   */
  LDIFObject* value() 
  {
    object = LDIFObject();

    if (!good()) return &object;

    UnknownAttrMap::const_iterator it = m_undefined_attrs->find(*m_itr);
    if (it != m_undefined_attrs->end()) {
#ifdef WITH_LDAP_EXCEPTIONS
      const std::vector<UnknownAttrPair>& undefined_pairs = it->second;

      if (undefined_pairs.size() > 0) {
	// report first error message
        const std::string& error_str = undefined_pairs.begin()->first;
        throw UndefinedValueException(std::string("LDIFObject::value()"),
	  	            std::string("ldap_get_values()"), error_str);
      }
#endif
    }

    LDIFObjectMap::const_iterator it2 = m_objects->find(*m_itr);
    if (it2 != m_objects->end()) {
      object = it2->second;
    }

    return &object;
  }

private:

  /* DN vector iterator */
  DNIterator m_itr;

  /* Vector of DN strings */
  const DNVector *m_dns;

  /* Map of DNs to pairs of undefined attributes found in that DN, along with the ldap error */
  const UnknownAttrMap *m_undefined_attrs;

  /* Map of DNs to LDIFObjects */
  const LDIFObjectMap *m_objects;

  /* LDIF object that holds a copy of current object for return to the user */
  LDIFObject object;
};

/**
 * Concrete LDAP result entry.
 * This class specialises the standard interface 
 * for concrete LDAPConnectionSynchConnection search result entries.
 */
class ldap_search_result_t : public generic_search_result_t 
{
public:
  /**
   * Costructor
   */
  ldap_search_result_t(LDAPMessage* ldresult, LDAP* ld) 
  {
    m_good = (ldresult != NULL && ld != NULL);
    if (good()) {
      m_nresults = ldap_count_entries(ld, ldresult);
      if (empty()) {
        ldap_msgfree(ldresult);
        return;
      }
    } else {
      return;
    }
	
    LDAPMessage* ldap_1st_entry = ldap_first_entry( ld, ldresult );
    if( ldap_1st_entry == NULL ) {
      m_good = false;
#ifdef WITH_LDAP_EXCEPTIONS
      std::string error_msg( ldap_err2string( ldap_result2error(ld, ldresult, 0) ) );
      throw QueryException(std::string("ldap_search_result_t"),
                           std::string("ldap_first_entry()"), error_msg);
#endif
      return;
    }

    LDAPMessage *lne = ldap_1st_entry;
    do {
      char *dn_ptr = ldap_get_dn( ld, lne );
      std::string dn(dn_ptr);
      ldap_memfree(dn_ptr);

      BerElement *ber = NULL;
      LDIFObject this_object;
      char *attribute = NULL;

      for(attribute = ldap_first_attribute(ld, lne, &ber); 
	  attribute; attribute = ldap_next_attribute(ld, lne, ber) ) {
      
        char **values = ldap_get_values(ld, lne, attribute);
        if( values ) {
	  for(int i=0; values[i]!=NULL; i++)
	    this_object.add( (std::string)attribute, (std::string)values[i] );
          ldap_value_free(values);
	} else {
          std::string error_msg( ldap_err2string( ldap_result2error(ld, lne, 0) ) );
	  std::string attr_name(attribute);

          m_undefined_attrs[dn].push_back(std::make_pair(error_msg,attr_name));
	}
	ldap_memfree(attribute);
      }
      if( ber != NULL ) ber_free( ber, 0 );
      m_dns.push_back(dn);
      m_objects[dn] = this_object;
    } while(good() && (lne = ldap_next_entry(ld, lne)) != NULL);
    ldap_msgfree(ldresult);
  } 
	  
  ~ldap_search_result_t() 
  {
  }

  /**
   * Concrete Validation test
   */
  bool good() const { return m_good; }
  bool empty() const { return ( good() && (m_nresults == 0) ); }
				/**
   * Concrete Entry Factory Method.
   * @return the first ldap result entry. The memory object should be destroid by explicitally calling delete.
   */
  generic_result_entry_t* make_first_entry() const
  {
    if (!good()) return NULL;

    ldap_result_entry_t *r = new ldap_result_entry_t( m_dns, m_undefined_attrs, m_objects );
    return (r);
  }
  
private:
  /* Indicates whether the ldap_search_result_t object contains a valid result */
  bool m_good;

  /* Is the object is good, contains the number of entries in the result */
  int m_nresults;

  /* Vector of DN strings */
  DNVector m_dns;
  
  /* Map of DNs to pairs of undefined attributes found in that DN, along with the ldap error */
  UnknownAttrMap m_undefined_attrs;

  /* Map of DNs to LDIFObjects */
  LDIFObjectMap m_objects;
};

/**
 * A Concrete Synchronous connection to an LDAP Server.
 * This class implements all virtual superclass methods in order
 * to provide a synchronous concrete connection to the LDAP Server.
 * @see LDAPConnection
 * @author Salvatore Monforte salvatore.monforte@ct.infn.it
 */
class LDAPSynchConnection : public LDAPConnection
{
public:
  /**
   * Constructor.
   */
  LDAPSynchConnection(const std::string&, const std::string&, int, long = 15);
  
  /**
   * Distructor.
   */
  virtual ~LDAPSynchConnection();
  
  /**
   * Opens a connection to the Server. 
   * This function overrides virtual LDAP Connection one.
   * @return true on success, false otherwise.
   */
  bool open();
  /**
   * Closes a connection to the Server. 
   * This function overrides virtual LDAP Connection one.
   * @return true on success, false otherwise.
   */
  bool close();
  /**
   * Performs LDAP search  operations using ldap_search_st() 
   * allowing a timeout to be specified.
   * @param q the query which will be executed.
   * @return a pointer to generic search result. This pointer should be explicitly freed.
   */
  generic_search_result_t* execute(LDAPQuery* q); 
 /**
   * Conenction test wheter established or not.
   * @return true if connection is established, false otherwise.
   */
  bool is_established() const;
  std::string error() const { return (saved_error) ? std::string( ldap_err2string(ldap_last_error) ) : std::string("No error available"); }

private:
  char** make_topics( const std::vector<std::string>& );
  bool l_open(int&, bool&, bool&);
  void l_note_failure();

private:
  /** A connection timeout. */
  struct timeval timeout;
  int source_port;
  std::string source_name;
  std::string base_dn;
  LDAP* handle;

  /** ldap error code from a previous failure */
  int ldap_last_error;

  /** Indicates whether ldap_last_error is set */
  bool saved_error;
  
  /** Maximum number of retries to try */
  int max_retries;

  /** Number of failures encountered as far */
  int failures;
};

} // namespace ldif2classad 
} // namespace common
} // namespace workload
} // namespace edg 

#endif

/*
  Local Variables:
  mode: c++
  End:
*/
