/***************************************************************************
 *  filename  : JobWrapper.h
 *  authors   : Alessio Gianelle <alessio.gianelle@pd.infn.it>
 *              Francesco Giacomini <francesco.giacomini@cnaf.infn.it>
 *              Rosario Peluso <rosario.peluso@pd.infn.it>
 *              Elisabetta Ronchieri <elisabetta.ronchieri@cnaf.infn.it>
 *  Copyright (c) 2002 CERN and INFN on behalf of the EU DataGrid.
 *  For license conditions see LICENSE file or
 *  http://www.edg.org/license.html
 ***************************************************************************/

#ifndef EDG_WORKLOAD_PLANNING_JOBADAPTER_JOB_WRAPPER_H
#define EDG_WORKLOAD_PLANNING_JOBADAPTER_JOB_WRAPPER_H

#ifndef EDG_WORKLOAD_X_STRING
#define EDG_WORKLOAD_X_STRING
#include <string>
#endif

#ifndef EDG_WORKLOAD_X_VECTOR
#define EDG_WORKLOAD_X_VECTOR
#include <vector>
#endif

#ifndef EDG_WORKLOAD_X_IOSTREAM
#define EDG_WORKLOAD_X_IOSTREAM
#include <iostream>
#endif

#ifndef EDG_WORKLOAD_X_UTILITY
#define EDG_WORKLOAD_X_UTILITY
#include <utility>
#endif

#ifndef EDG_WORKLOAD_X_BOOST_SHARED_PTR_HPP
#define EDG_WORKLOAD_X_BOOST_SHARED_PTR_HPP
#include <boost/shared_ptr.hpp>
#endif

#ifndef EDG_WORKLOAD_PLANNING_COMMON_URL_H
#include "edg/workload/planning/common/URL.h"
#endif

namespace classad {
class ExprList;
}

namespace edg {
namespace workload {
namespace planning {
namespace jobadapter {
	
class JobWrapper
{
  friend std::ostream& operator<<(std::ostream& os, const JobWrapper& jw);

public:
  /**
   *  Default constructor.
   *  Initializes the object passing the parameters to the job wrapper.
   *  \param job
   *  \ingroup jobadapter
   */
  JobWrapper(const std::string& job);

  /**
   *  Default destructor.
   *  \ingroup jobadapter
   */
  virtual ~JobWrapper(void);

public:
  /**
   * Declare the standard input file.
   * @param file file to be used as standard input for the user job
   * \ingroup jobadapter
   */
  void standard_input(const std::string& file);

  /**
   * Declare the standard output file.
   * @param file file to be used as standard output of the user job
   * \ingroup jobadapter
   */
  void standard_output(const std::string& file);

  /**
   * Declare the standard error file.
   * @param file file to be used as standard error of the user job
   * \ingroup jobadapter
   */
  void standard_error(const std::string& file);

  /**
   * Declare which files are in the input sandbox and where they are located.
   * @param base_url location of all the files in the input sandbox
   * @param files    list of files in the input sandbox
   * \ingroup jobadapter
   */
  void input_sandbox(const common::URL& base_url, 
                     const std::vector<std::string>& files);
  
  /**
   * Declare which files are in the output sandbox and where they have to be moved.
   * @param base_url location the files in the output sandbox have to be moved to
   * @param files    list of files in the output sandbox
   * \ingroup jobadapter
   */
  void output_sandbox(const common::URL& base_url,
                      const std::vector<std::string>& files);

  /**
   * Set the value of the EDG_WL_RB_BROKERINFO environment variable.
   * If the function is not called the variable is not set.
   * @param file file containing Broker information in the job working
   *             directory. If file is not specified a default value is used.
   * \ingroup jobadapter
   */
  void brokerinfo(const std::string& file = s_brokerinfo_default);

  /**
   * Declare that the job should run in a subdirectory.
   * \ingroup jobadapter
   */
  void create_subdir(void);

  /**
  * Set the value of the EDG_WL_JOBID environment variable.
  * If the function is not called the variable is not set.
  * @param jobid
  * \ingroup jobadapter
  */
  void job_Id(const std::string& jobid);
  
  /**
   * Set the value of the workdir.
   * @param jobid
   * \ingroup jobadapter
   */
  void job_id_to_filename(const std::string& jobid_to_filename);

  /**
   * Set the host name of the submitting resource broker.
   * @param rbname
   * \ingroup jobadapter
   */
  void brokername(const std::string& rbname);

  /**
  * Indicate that sync running event logging is required
  * \ingroup jobadapter
  */
  void set_sync_running_event(void);

  /**
   * Set the maximum total size of files that
   * will be transfered in the output sandbox
   * \ingroup jobadapter
   */
   void set_max_output_size(const int64_t &max_size);
  
  /**
  * Declare arguments
  * @param args
  * \ingroup jobadapter
  */
  void arguments(const std::string& args);

  /*
  * Declare maradonaprotocol
  * @param maradonaprotocol
  * \ingroup maradonaprotocol
  */
  void maradonaprotocol(const std::string& protocol, 
		        const std::string& filename);

  /**
   * Declare gatekeeper_hostname
   * @param gatekeeper
   * \ingroup jobadapter
   */
  void gatekeeper_hostname(const std::string& scode);

  /**
   * Declare which environment variables must be exported.
   * @param env    list of the environment variables
   * \ingroup jobadapter
   */
  void environment(const std::vector<std::string>& env);

  void nodes(int node);

   /**
    * Field the outputdata attribute.
    * @param output_data    list of classads
    * \ingroup jobadapter
    */
   void outputdata(const classad::ExprList* output_data);

   /**
    * Declare virtual organization
    * @param vo 
    * \ingroup jobadapter
    */ 
   void vo(const std::string& vo);

   /**
    * Set the name of the DSUpload output file.
    * @param jobid
    * \ingroup jobadapter
    */
   void dsupload(const common::URL& id);

   /**
    * Declare application directory
    * @param app_dir
    * \ingroup jobadapter
    */ 
   void app_dir(const std::string& app_dir);
   
protected:
  virtual std::ostream& print(std::ostream& os) const;
  
  virtual std::ostream& set_shell(std::ostream& os) const;

  virtual std::ostream& transfer_output_sb(std::ostream& os, const int64_t &max_size,
                   const std::vector<std::string> &file, const common::URL &returl) const;

  virtual std::ostream& set_cleanup_globals(std::ostream& os) const;

  virtual std::ostream& write_cleanup_subr(std::ostream& os) const;

  virtual std::ostream& set_scratchdir(std::ostream& os,
                                       bool create_subdir) const;
  
  virtual std::ostream& set_subdir(std::ostream&      os,
		                   const std::string& dir) const;
  virtual std::ostream& clean_subdir(std::ostream& os) const;
  
  virtual std::ostream& check_workdir(std::ostream& os) const;
  virtual std::ostream& set_workdir(std::ostream& os) const;
  
  virtual std::ostream& set_environment(std::ostream&                   os,
		  		        const std::vector<std::string>& env) const;
  virtual std::ostream& set_environment(std::ostream&      os,
			                const std::string& variable,
					const std::string& value) const;
  virtual std::ostream& check_set_environment(std::ostream& os,
                                        const std::string&  variable,
                                        const std::string&  value) const;

  virtual std::ostream& set_umask(std::ostream& os) const;
  
  virtual std::ostream& check_file_mod(std::ostream&      os,
		                       const std::string& file) const;
  
  virtual std::ostream& set_lb_sequence_code(std::ostream&      os,
  	                                     const std::string& event,
					     const std::string& reason,
					     const std::string& scode,
					     const std::string& ecode) const;
  virtual std::ostream& check_globus(std::ostream& os) const;

  virtual std::ostream& prepare_transfer(std::ostream&                   os,
			                 const std::vector<std::string>& file) const;
  virtual std::ostream& make_transfer(std::ostream&      os,
		  		      const common::URL& prefix,
				      const bool&        input) const;

  virtual std::ostream& execute_job(std::ostream&      os,
			            const std::string& arguments,
			            const std::string& job,
			            const std::string& stdi,
			            const std::string& stdo,
			            const std::string& stde,
				    int                node) const;

  virtual std::ostream& make_bypass_transfer(std::ostream& os,
		  			     const std::string& prefix) const;

  virtual std::ostream& create_maradona_file(std::ostream& os,
		  			     const std::string& jobid_to_filename) const;
  
  virtual std::ostream& doExit(std::ostream& os,
		  	       const std::string& maradonaprotocol,
			       bool create_subdir) const;

  virtual std::ostream& handle_output_data(std::ostream& os) const;

  virtual std::ostream& doDSUploadTmp(std::ostream& os) const;
  virtual std::ostream& doDSUpload(std::ostream& os) const;
  virtual std::ostream& doCheckReplicaFile(std::ostream& os) const;
  virtual std::ostream& doReplicaFile(std::ostream& os) const;
  virtual std::ostream& doReplicaFilewithLFN(std::ostream& os) const;
  virtual std::ostream& doReplicaFilewithSE(std::ostream& os) const;
  virtual std::ostream& doReplicaFilewithLFNAndSE(std::ostream& os) const;
	
protected:
  // executable, stdin, stdout, stderr, and arguments
  std::string              m_job;
  std::string              m_standard_input;
  std::string              m_standard_output;
  std::string              m_standard_error;
  std::string              m_arguments;
  std::string              m_maradonaprotocol;
  
  // input sandbox
  common::URL              m_input_base_url;
  std::vector<std::string> m_input_files;

  // output sandbox
  common::URL              m_output_base_url;
  std::vector<std::string> m_output_files;

  // brokerinfo file
  std::string              m_brokerinfo;

  // create subdirectory?
  bool                     m_create_subdir;

  // require sync running event logging from LRMS?
  bool			   m_sync_running_event;

  // job id
  std::string              m_jobid;

  // job id file name
  std::string              m_jobid_to_filename;
  
  // gatekeeper hostname
  std::string              m_gatekeeper_hostname;

  // environment variable
  std::vector<std::string> m_environment;

  // node number
  int                      m_nodes;

  // List of OutputData
  boost::shared_ptr<classad::ExprList> m_outputdata;
  
  std::string 		   m_vo;

  std::string 		   m_rbname;

  std::string              m_dsupload;

  std::string              m_app_dir;

  int64_t                  m_max_size;
  
protected:
  // default value of the brokerinfo file
  static const std::string s_brokerinfo_default;
};

} // namespace jobadapter
} // namespace planning
} // namespace workload
} // namespace edg

#endif // EDG_WORKLOAD_PLANNING_JOBADAPTER_JOB_WRAPPER_H
