// File: WMReal.tcc
// Author: Francesco Giacomini <Francesco.Giacomini@cnaf.infn.it>
// Copyright (c) 2002 EU DataGrid.
// For license conditions see http://www.eu-datagrid.org/license.html

// $Id: WMReal.tcc,v 1.28 2005/03/19 19:23:05 dhsmith Exp $

#include <boost/scoped_ptr.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/bind.hpp>
#include <boost/tuple/tuple.hpp>
#include <classad_distribution.h>
#include "edg/workload/planning/manager/lb_utils.h"
#include "edg/workload/planning/common/logger_utils.h"
#include "edg/workload/planning/helper/exceptions.h"
#include "edg/workload/common/jobid/JobId.h"
#include "edg/workload/common/jobid/JobIdExceptions.h"
#include "edg/workload/common/utilities/classad_utils.h"
#include "edg/workload/common/utilities/Exceptions.h"
#include "edg/workload/common/requestad/JobAdManipulation.h"
#include "edg/workload/common/requestad/PrivateAdManipulation.h"
#include "edg/workload/common/requestad/ManipulationExceptions.h"
#include "edg/workload/common/configuration/Configuration.h"
#include "edg/workload/common/configuration/WMConfiguration.h"
#include "edg/workload/logging/client/producer.h"
#include "edg/workload/logging/client/consumer.h"
#include "edg/workload/proxyrenewal/renewal.h"
#include "edg/workload/purger/purger.h"
#include "edg/workload/common/utilities/scope_guard.h"

namespace edg {
namespace workload {
namespace planning {
namespace manager {

namespace {

int get_max_retry_count()
{
  common::configuration::Configuration const* const config
    = common::configuration::Configuration::instance();

  if (!config) {
    Fatal("empty or invalid configuration");
  }
  common::configuration::WMConfiguration const* const wm_config = config->wm();
  if (!wm_config) {
    Fatal("empty WM configuration");
  }

  return wm_config->max_retry_count();
}

int get_max_shallow_count()
{
  common::configuration::Configuration const* const config
    = common::configuration::Configuration::instance();

  if (!config) {
    Fatal("empty or invalid configuration");
  }
  common::configuration::WMConfiguration const* const wm_config = config->wm();
  if (!wm_config) {
    Fatal("empty WM configuration");
  }

  return wm_config->max_shallow_count();
}

} // {anonymous}

template<typename PlanningP, typename DeliveryP, typename CancellingP>
WMReal<PlanningP, DeliveryP, CancellingP>::WMReal()
{
}

template<typename PlanningP, typename DeliveryP, typename CancellingP>
WMReal<PlanningP, DeliveryP, CancellingP>::~WMReal()
{
}

template<typename PlanningP, typename DeliveryP, typename CancellingP>
void
WMReal<PlanningP, DeliveryP, CancellingP>::submit(classad::ClassAd const* request_ad_p)
{
  if (!request_ad_p) {
    Error("request ad is null");
    return;
  }

  // make a copy because we change the sequence code, see below
  classad::ClassAd request_ad(*request_ad_p);

  common::jobid::JobId jobid;

  try {

    common::jobid::JobId tmp_jobid(common::requestad::get_edg_jobid(request_ad));
    jobid = tmp_jobid;

  } catch (common::requestad::CannotGetAttribute const& e) {

    Error(e.what() << " for " << common::utilities::unparse_classad(request_ad));
    return;

  } catch (common::jobid::JobIdException const& e) {

    Error(e.what() << " for " << common::utilities::unparse_classad(request_ad));
    return;

  }

  ContextPtr context_ptr = get_context(jobid);
  assert(context_ptr);
  edg_wll_Context context = *context_ptr;

  // keep the storage guard outside the try block, because the lb logs in the
  // catch blocks require the proxy, which lives in the storage
  common::utilities::scope_guard storage_guard(
    boost::bind(purger::purgeStorage, jobid, std::string(""))
  );

  try {

    common::utilities::scope_guard proxy_guard(
      boost::bind(edg_wlpr_UnregisterProxy, jobid, static_cast<char const*>(0))
    );

    common::utilities::scope_guard context_guard(boost::bind(unregister_context, jobid));

    // update the sequence code before doing the planning
    // some helpers may need a more recent sequence code than what appears in
    // the classad originally received

    char* c_sequence_code = edg_wll_GetSequenceCode(context);
    std::string sequence_code(c_sequence_code);
    free(c_sequence_code);

    common::requestad::remove_lb_sequence_code(request_ad);
    common::requestad::set_lb_sequence_code(request_ad, sequence_code);

    boost::scoped_ptr<classad::ClassAd> planned_ad(Plan(request_ad));

    char const* const c_ce_id = common::requestad::get_ce_id(*planned_ad).c_str();
    int lb_error = edg_wll_LogMatch(context, c_ce_id); 
    if (lb_error != 0) {
      Warning("edg_wll_LogMatch failed for " << jobid
              << " (" << get_lb_message(context) << ")");
    }

    Deliver(*planned_ad);

    context_guard.dismiss();
    proxy_guard.dismiss();

  } catch (common::requestad::ManipulationException const& e) {

    std::string reason("Cannot plan: ");
    reason += e.what();
    Error(reason << " for " << jobid);
    int err; ContextPtr ctx;
    boost::tie(err, ctx) = lb_log(
      boost::bind(edg_wll_LogAbort, _1, reason.c_str()),
      context_ptr
    );
    if (err) {
      Warning(get_logger_message("edg_wll_LogAbort", err, context_ptr, ctx));
    }
    return;

  } catch (helper::HelperError& e) {

    std::string reason("Cannot plan: ");
    reason += e.what();
    Error(reason << " for " << jobid);
    int err; ContextPtr ctx;
    boost::tie(err, ctx) = lb_log(
      boost::bind(edg_wll_LogAbort, _1, reason.c_str()),
      context_ptr
    );
    if (err) {
      Warning(get_logger_message("edg_wll_LogAbort", err, context_ptr, ctx));
    }
    return;

  } catch (std::invalid_argument const& e) {

    std::string reason("Cannot plan: ");
    reason += e.what();
    Error(reason << " for " << jobid);
    int err; ContextPtr ctx;
    boost::tie(err, ctx) = lb_log(
      boost::bind(edg_wll_LogAbort, _1, reason.c_str()),
      context_ptr
    );
    if (err) {
      Warning(get_logger_message("edg_wll_LogAbort", err, context_ptr, ctx));
    }
    return;

    // better: keep the request pending for a while in a
    // std::vector<std::pair<time_type, jdl, retry_count>
    // and log:
    //  edg_wll_LogPending(context, reason);
    //  edg_wll_Enqueue(context, queue_name, jdl, "OK", "");
    // or resubmit()?
  }

  storage_guard.dismiss();
}

template<typename PlanningP, typename DeliveryP, typename CancellingP>
void
WMReal<PlanningP, DeliveryP, CancellingP>::resubmit(common::jobid::JobId const& request_id)
{
  ContextPtr context_ptr = get_context(request_id);

  assert(context_ptr);

  edg_wll_Context context = *context_ptr;

  // keep the storage guard outside the try block, because the lb logs in the
  // catch blocks require the proxy, which lives in the storage
  common::utilities::scope_guard storage_guard(
    boost::bind(purger::purgeStorage, request_id, std::string(""))
  );

  int lb_error;

  try {

    common::utilities::scope_guard proxy_guard(
      boost::bind(edg_wlpr_UnregisterProxy, request_id, static_cast<char const*>(0))
    );

    common::utilities::scope_guard context_guard(boost::bind(unregister_context, request_id));

    // flush the lb events since we'll query the server
    // struct timeval* timeout = 0;
    // lb_error = edg_wll_LogFlush(context, timeout);
    // if (lb_error != 0) {
    //   Warning("edg_wll_LogFlush failed for " << request_id
    //           << " (" << get_lb_message(context) << ")");
    // }

    // retrieve previous matches + start of execution status; no resubmission if query failure
    unsigned int n_standard_submissions;
    bool elligeable_for_shallow;
    std::vector<std::pair<std::string,int> > const previous_matches_ex
      = get_previous_matches_ex(context, request_id, n_standard_submissions, elligeable_for_shallow);

    if (previous_matches_ex.empty()) {
      std::ostringstream os;
      os << "cannot retrieve previous matches for " << request_id;
      std::string reason = os.str();
      Warning(reason);

      int err; ContextPtr ctx;
      boost::tie(err, ctx) = lb_log(
        boost::bind(edg_wll_LogAbort, _1, reason.c_str()),
        context_ptr
      );
      if (err) {
        Warning(get_logger_message("edg_wll_LogAbort", err, context_ptr, ctx));
      }
      return;
    }

    std::vector<std::string> previous_matches;
    for (std::vector<std::pair<std::string,int> >::const_iterator it = previous_matches_ex.begin();
         it != previous_matches_ex.end(); ++it) {
      previous_matches.push_back(it->first);
    }

    // retrieve original jdl
    std::string const job_ad_str = get_original_jdl(context, request_id);
    boost::scoped_ptr<classad::ClassAd> job_ad(common::utilities::parse_classad(job_ad_str));

    // sanity check - if sync running event logging was not explictly enabled
    // allow no possibility of shallow resubmission
    bool good;
    if ((!common::requestad::get_edg_sync_running_event(*job_ad,good) && good) || !good) {
      n_standard_submissions = previous_matches.size();
      elligeable_for_shallow = false;
    } else {
      // check limit on shallow resubmissions
      int max_shallow_count = get_max_shallow_count();
      if (max_shallow_count >= 0) {
        unsigned int curr_shallows = previous_matches.size() - n_standard_submissions;
        if (curr_shallows > static_cast<unsigned>(max_shallow_count)) {
          n_standard_submissions = previous_matches.size() - static_cast<unsigned>(max_shallow_count);
          elligeable_for_shallow = false;
        }
      }
    }

    // maximum allowed number of standard job resubmissions
    unsigned int nominal_allowed_retries = 0;

    // check the system max retry count; abort if exceeded
    int max_retry_count = get_max_retry_count();
    if (max_retry_count < 0
        || n_standard_submissions > static_cast<unsigned>(max_retry_count)) {
      std::ostringstream os;
      os << "MaxRetryCount (" << max_retry_count << ") hit for " << request_id;
      std::string reason = os.str();
      Info(reason);
      int err; ContextPtr ctx;
      boost::tie(err, ctx) = lb_log(
        boost::bind(edg_wll_LogAbort, _1, reason.c_str()),
        context_ptr
      );
      if (err) {
        Warning(get_logger_message("edg_wll_LogAbort", err, context_ptr, ctx));
      }

      return;
    }
    nominal_allowed_retries = max_retry_count;

    // check the job max retry count; abort if exceeded
    bool count_valid = false;
    int job_retry_count = common::requestad::get_retry_count(*job_ad, count_valid);
    if (!count_valid) {
      job_retry_count = 0;
    }
    if (job_retry_count < 0
        || n_standard_submissions > static_cast<unsigned>(job_retry_count)) {
      std::ostringstream os;
      os << "Job RetryCount (" << job_retry_count << ") hit";
      std::string reason = os.str();
      Info(reason << " for " << request_id);
      int err; ContextPtr ctx;
      boost::tie(err, ctx) = lb_log(
        boost::bind(edg_wll_LogAbort, _1, reason.c_str()),
        context_ptr
      );
      if (err) {
        Warning(get_logger_message("edg_wll_LogAbort", err, context_ptr, ctx));
      }

      return;
    }
    if (nominal_allowed_retries > static_cast<unsigned>(job_retry_count))
      nominal_allowed_retries = job_retry_count;

    // in practice nominal_allowed_retries (the actual retry limit for standard submissions) is
    // max(0, min(job_retry_count, max_retry_count))

    // if we're elligeable for shallow resubmission but otherwise
    // wouldn't have been elligeable for standard resubmission, indicate
    // that the job should not be sent to any previously matched site
    bool retry_previous_destinations_disallowed = false;
    if (elligeable_for_shallow && (n_standard_submissions >= nominal_allowed_retries))
      retry_previous_destinations_disallowed = true;

    // do the planning/delivery; abort if failure
    common::requestad::set_edg_previous_matches(*job_ad, previous_matches);
    common::requestad::set_edg_previous_matches_ex(*job_ad, previous_matches_ex);
    common::requestad::set_edg_retry_previous_destinations_disallowed(*job_ad, retry_previous_destinations_disallowed);

    // log debug infomation about shallow submission
    if (elligeable_for_shallow) {
      std::ostringstream os;
      os << "Possible shallow resub " << (previous_matches.size()-n_standard_submissions) << " on retry " <<
            previous_matches.size() << " (nom. max " << nominal_allowed_retries << "). " <<
            (retry_previous_destinations_disallowed ? "Will" : "Will not") << " require unique desintation" <<
            " for " << request_id;
      std::string reason = os.str();
      Info(reason);
    } else {
      std::ostringstream os;
      os << "Standard resub " << n_standard_submissions << " on retry " <<
            previous_matches.size() << " (nom. max " << nominal_allowed_retries << ")" <<
            " for " << request_id;
      std::string reason = os.str();
      Info(reason);
    }

    boost::scoped_ptr<classad::ClassAd> planned_ad(Plan(*job_ad));

    char const* const c_ce_id = common::requestad::get_ce_id(*planned_ad).c_str();
    lb_error = edg_wll_LogMatch(context, c_ce_id);
    if (lb_error != 0) {
      Warning("edg_wll_LogMatch failed for " << request_id
              << " (" << get_lb_message(context) << ")");
    }

    Deliver(*planned_ad);

    context_guard.dismiss();
    proxy_guard.dismiss();

  } catch (common::utilities::CannotParseClassAd const& e) {

    std::string reason;
    if (e.str().empty()) {
      reason = "cannot retrieve original JDL";
    } else {
      reason = "original JDL is not a classad";
    }
    Error(reason << " for " << request_id);
    int err; ContextPtr ctx;
    boost::tie(err, ctx) = lb_log(
      boost::bind(edg_wll_LogAbort, _1, reason.c_str()),
      context_ptr
    );
    if (err) {
      Warning(get_logger_message("edg_wll_LogAbort", err, context_ptr, ctx));
    }

    return;

  } catch (common::requestad::ManipulationException const& e) {

    std::string reason(e.what());
    Error(reason << " for " << request_id);
    int err; ContextPtr ctx;
    boost::tie(err, ctx) = lb_log(
      boost::bind(edg_wll_LogAbort, _1, reason.c_str()),
      context_ptr
    );
    if (err) {
      Warning(get_logger_message("edg_wll_LogAbort", err, context_ptr, ctx));
    }

    return;

  } catch (helper::HelperError& e) {

    std::string reason("Cannot plan: ");
    reason += e.what();
    Error(reason << " for " << request_id);
    int err; ContextPtr ctx;
    boost::tie(err, ctx) = lb_log(
      boost::bind(edg_wll_LogAbort, _1, reason.c_str()),
      context_ptr
    );
    if (err) {
      Warning(get_logger_message("edg_wll_LogAbort", err, context_ptr, ctx));
    }

    return;
  }

  storage_guard.dismiss();
}

template<typename PlanningP, typename DeliveryP, typename CancellingP>
void
WMReal<PlanningP, DeliveryP, CancellingP>::cancel(common::jobid::JobId const& request_id)
{
  Cancel(request_id);
}

}}}} // edg::workload::planning::manager

// Local Variables:
// mode: c++
// End:
