// File: WorkloadManager.h
// Author: Francesco Giacomini <Francesco.Giacomini@cnaf.infn.it>
// Copyright (c) 2001 EU DataGrid.
// For license conditions see http://www.eu-datagrid.org/license.html

// $Id: WorkloadManager.h,v 1.6 2003/03/27 14:13:14 giaco Exp $

#ifndef EDG_WORKLOAD_PLANNING_MANAGER_WORKLOADMANAGER_H
#define EDG_WORKLOAD_PLANNING_MANAGER_WORKLOADMANAGER_H

#ifndef EDG_WORKLOAD_X_BOOST_UTILITY_HPP
#define EDG_WORKLOAD_X_BOOST_UTILITY_HPP
#include <boost/utility.hpp>
#endif

#ifndef EDG_WORKLOAD_X_BOOST_SCOPED_PTR_HPP
#define EDG_WORKLOAD_X_BOOST_SCOPED_PTR_HPP
#include <boost/scoped_ptr.hpp>
#endif

namespace classad {
class ClassAd;
}

namespace edg {
namespace workload {

namespace common {
namespace jobid {
class JobId;
}}

namespace planning {
namespace manager {

class WMImpl;

/**
 * The Workload Manager service
 *
 * The WorkloadManager class represents the access point to the workload
 * management service, intended as scheduling plus reliable job management.
 */
class WorkloadManager: boost::noncopyable
{
  boost::scoped_ptr<WMImpl> m_impl;

public:
  /**
   * Constructor: chooses the right implementation for the current environment.
   *
   * The default constructor chooses the right implementation of the
   * WorkloadManager depending on the current environment. Information may be
   * gathered from the configuration, the process environment, etc.
   */
  WorkloadManager();

  // the dtor is necessary even if it doesn't do anything, because otherwise it
  // is generated automatically inline (?) calling the dtor of m_impl, that
  // requires that the definition of WMImpl is available (now it is a forward
  // declaration) 
  ~WorkloadManager();

  /** 
   * Submit a request to the Workload Manager
   * 
   * The request, if valid, is subject to:
   * -# some planning
   * -# delivering to the resource that is able to satisfy it
   *
   * @param request_ad description of the request
   */
  void submit(classad::ClassAd const* request_ad);

  /**
   * Resubmit a request
   *
   * The resubmission is applied only to requests satisfying resubmission
   * constraits, e.g. requests that have failed and are "idle".  Relevant
   * information will be retrieved from the Logging and Bookkeping Service.
   *
   * @param request_id request identifier.
   */
  void resubmit(common::jobid::JobId const& request_id);

  /** 
   * Cancel a request 
   * 
   * @param request_id identifier of the request to be aborted. 
   */
  void cancel(common::jobid::JobId const& request_id);

};

}}}} // edg::workload::planning::manager

#endif // EDG_WORKLOAD_PLANNING_MANAGER_WORKLOADMANAGER_H

