////////////////////////////////////////////////////////////////////////////////
// The Loki Library
// Copyright (c) 2001 by Andrei Alexandrescu
// This code accompanies the book:
// Alexandrescu, Andrei. "Modern C++ Design: Generic Programming and Design
//     Patterns Applied". Copyright (c) 2001. Addison-Wesley.
// Permission to use, copy, modify, distribute and sell this software for any
//     purpose is hereby granted without fee, provided that the above copyright
//     notice appear in all copies and that both that copyright notice and this
//     permission notice appear in supporting documentation.
// The author or Addison-Welsey Longman make no representations about the
//     suitability of this software for any purpose. It is provided "as is"
//     without express or implied warranty.
////////////////////////////////////////////////////////////////////////////////

// $Id: Factory.h,v 1.1 2003/03/27 12:33:11 giaco Exp $

#ifndef LOKI_FACTORY_H
#define LOKI_FACTORY_H

#include <exception>
#include <map>

namespace Loki
{

////////////////////////////////////////////////////////////////////////////////
// class template DefaultFactoryError
// Manages the "Unknown Type" error in an object factory
////////////////////////////////////////////////////////////////////////////////

template <typename IdentifierType, class AbstractProduct>
struct DefaultFactoryError
{
  struct Exception : public std::exception
  {
    const char* what() const throw () { return "Unknown Product Type"; }
  };

  static AbstractProduct* OnUnknownType(IdentifierType)
  {
    throw Exception();
  }
};

////////////////////////////////////////////////////////////////////////////////
// class template Factory
// Implements a generic object factory
////////////////////////////////////////////////////////////////////////////////

template
<
  class AbstractProduct,
  typename IdentifierType,
  typename ProductCreator = AbstractProduct* (*)(),
  template<typename, class>
class FactoryErrorPolicy = DefaultFactoryError
>
class Factory
  : public FactoryErrorPolicy<IdentifierType, AbstractProduct>
{
public:
  typedef std::map<IdentifierType, ProductCreator> IdToProductMap;

private:
  IdToProductMap associations_;

public:
  bool Register(const IdentifierType& id, ProductCreator creator)
  {
    return associations_.insert(
      IdToProductMap::value_type(id, creator)).second;
  }

  bool Unregister(const IdentifierType& id)
  {
    return associations_.erase(id) == 1;
  }

  AbstractProduct* CreateObject(const IdentifierType& id)
  {
    typename IdToProductMap::const_iterator i = associations_.find(id);
    if (i != associations_.end())
    {
      return (i->second)();
    }
    return OnUnknownType(id);
  }

  IdToProductMap const& List() const
  {
    return associations_;
  }

};

} // namespace Loki

#endif

